unit PickDest;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see AdvOpt.pas for license}
interface

uses
  {$IFDEF MSWINDOWS} Windows, Messages,  {$ENDIF}
  SysUtils, Variants, Classes, Graphics, Controls, NGrid,
  StdCtrls, Grids, Buttons, WhatsThis, Dialogs, GridEx,
  MathSup, ExpPlotLoader;

type

  TPickDestForm = class(THelpForm)
    Grid: TNumberGrid;
    OKButton: TButton;
    CancelButton: TButton;
    Label1: TLabel;
    AllButton: TButton;
    NoneButton: TButton;
    procedure OKButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure AllButtonClick(Sender: TObject);
    procedure GridGetCellIsFlag(Sender: TObject; Col, Row: Integer;
      var IsFlag: Boolean);
    procedure GridDrawCellContents(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState; var Drawn: Boolean);
    procedure GridSetCell(Sender: TObject; ACol, ARow: Integer; var v: Double;
      var AValid: Boolean);
    procedure NoneButtonClick(Sender: TObject);
  private
    nIntensity, nWeighted: Integer;
    ChannelDest: TIntegerArray;
    Limits: TDPointArray;
  public
    Xname, Yname, Wname: string;
    constructor Create(AOwner: TComponent); override;
    function Execute(var AChannelDest: TIntegerArray; var AnWeights: Integer; ALimits: TDPointArray; AName: string): Integer;
  end;

var
  PickDestForm: TPickDestForm;

implementation

uses Forms;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function TPickDestForm.Execute(var AChannelDest: TIntegerArray; var AnWeights: Integer; ALimits: TDPointArray; AName: string): Integer;
var
  i: Integer;
begin
  if (Length(AName) > 0) and not (Name[1] in ['<', '.']) then
    Caption := 'Pick Channels for ' + AName
  else
    Caption := 'Pick Channels';
  AllButton.Hint := 'Flag all unused channels as ' + Yname;
  ChannelDest := AChannelDest;
  Limits := ALimits;
  Grid.RowCount := Length(ChannelDest)+1;
  Grid.Title[0] := 'Channel';
  Grid.Title[1] := Xname;
  Grid.Title[2] := Yname;
  Grid.Title[3] := Wname;
  for i := 0 to 3 do
    Grid.AutoAdjustColumn(i);
  Width := Grid.NaturalWidth + 2*Grid.Left;
  RequestAlign;
  for i := 1 to Length(ChannelDest) do begin
    if ChannelDest[i-1] >= 1000 then
      Grid.Valid[i,3] := True
    else if ChannelDest[i-1] > 0 then
      Grid.Valid[i,2] := True
    else if ChannelDest[i-1] = 0 then
      Grid.Valid[i,1] := True
    else begin
      { Force clear of row }
      Grid.Valid[i,1] := False;
      Grid.Valid[i,2] := False;
      Grid.Valid[i,3] := False;
    end;
    Grid.Lbl[i] := IntToStr(i);
    Grid.Value[i,4] := Limits[i-1].X;
    Grid.Value[i,5] := Limits[i-1].Y;
  end;
  AnWeights := 0;
  if Length(ChannelDest) > 2 then
    Label1.Caption := 'The file has more than two data channels; please select the required channels'
  else
    Label1.Caption := 'Please select the required channels';
  if ShowModal = mrOK then begin
    Result := nIntensity;
    AnWeights := nWeighted;
  end else
    Result := 0;
end;

procedure TPickDestForm.GridDrawCellContents(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState; var Drawn: Boolean);
var
  x, y, h: Integer;
  PenC, BrushC: TColor;
  Style: TBrushStyle;
begin
  with Grid do
    if (ARow > 0) and (ACol >=1) and (ACol <=3) then begin
      x := (Rect.Right + Rect.Left) div 2;
      y := (Rect.Bottom + Rect.Top) div 2;
      h := y-Rect.Top - 2;
      Canvas.Ellipse(x-h,y-h,x+h,y+h);
      if Valid[ARow, ACol] then begin
        PenC := Canvas.Pen.Color;
        BrushC := Canvas.Brush.Color;
        Style := Canvas.Brush.Style;
        Canvas.Pen.Color := clBlack;
        Canvas.Brush.Style := bsSolid;
        Canvas.Brush.Color := clBlack;
        Dec(h,3);
        Canvas.Ellipse(x-h,y-h,x+h,y+h);
        Canvas.Pen.Color := PenC;
        Canvas.Brush.Color := BrushC;
        Canvas.Brush.Style := Style;
      end;
      Drawn := True;
    end
end;

procedure TPickDestForm.GridGetCellIsFlag(Sender: TObject; Col, Row: Integer;
  var IsFlag: Boolean);
begin
  IsFlag := (Col >= 1) and (Col <= 3);
end;

procedure TPickDestForm.GridSetCell(Sender: TObject; ACol, ARow: Integer;
  var v: Double; var AValid: Boolean);
var
  i: Integer;
begin
  with Grid do
    if AValid and (ACol >=1) and (ACol <=3) then begin
      for i := 1 to 3 do
        if i <> ACol then
          Valid[ARow,i] := False;
      { Ony allow one frequency channel }
      if ACol = 1 then
        for i := 1 to Length(ChannelDest) do
          if (i <> ARow) and Valid[i,1] then
            Grid.Valid[i,2] := True;
    end;
end;

procedure TPickDestForm.NoneButtonClick(Sender: TObject);
var
  i,c: Integer;
begin
  for i := 1 to Length(ChannelDest) do
    for c := 1 to 3 do
      Grid.Valid[i,c] := False;
end;

procedure TPickDestForm.OKButtonClick(Sender: TObject);
var
  i: Integer;
  HasFrequency: Boolean;
begin
  nIntensity := 0;
  HasFrequency := False;
  nWeighted := 1000;
  for i := 0 to High(ChannelDest) do begin
    if Grid.Valid[i+1,1] then begin
      if HasFrequency then
        raise Exception.Create('More than one frequency channel');
      ChannelDest[i] := 0;
      HasFrequency := True;
    end else if Grid.Valid[i+1,2] then begin
      Inc(nIntensity);
      ChannelDest[i] := nIntensity;
    end else if Grid.Valid[i+1,3] then begin
      ChannelDest[i] := nWeighted;
      Inc(nWeighted);
    end else
      ChannelDest[i] := -1;
  end;
  Dec(nWeighted,1000);
  if HasFrequency and (nIntensity = 0) then
    Inc(nIntensity); { Line list input - Frequency channel only }
  if (nWeighted > 0) and (nWeighted <> nIntensity) then
    raise Exception.Create('Specify same number of weight and intensity channels');
  ModalResult := mrOk;
end;

procedure TPickDestForm.AllButtonClick(Sender: TObject);
var
  i: Integer;
begin
  for i := 1 to Length(ChannelDest) do
    if not (Grid.Valid[i,1] or Grid.Valid[i,2] or Grid.Valid[i,3]) then
      Grid.Valid[i,2] := True;
end;

procedure TPickDestForm.CancelButtonClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

constructor TPickDestForm.Create(AOwner: TComponent);
begin
  inherited;
  Xname := 'x';
  Yname := 'y';
  Wname := 'Std Dev';
  Grid.OnDrawCellContents := {$IFDEF FPC}@{$ENDIF}GridDrawCellContents;
end;

end.
