unit Transfrm;
{$I DFS.inc}{ Copyright (C) 2007 C M Western; see GraphC.pas for license}
interface

uses {$IFNDEF CLX_} Graphics, Forms, Controls, StdCtrls, Buttons, ExtCtrls,
     {$ELSE} QGraphics, QForms, QControls, QStdCtrls, QButtons, QExtCtrls, {$ENDIF}
     SysUtils, Classes;

type

  { TTransformationsDlg }

  TTransformationsDlg = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    Bevel1: TBevel;
    Instructions: TLabel;
    NoneBtn: TButton;
    xLabel: TLabel;
    xFunction: TEdit;
    Check: TButton;
    yLabel: TLabel;
    yFunction: TEdit;
    Label4: TLabel;
    procedure FormResize(Sender: TObject);
    procedure NoneBtnClick(Sender: TObject);
    procedure CheckClick(Sender: TObject);
    procedure OKBtnClick(Sender: TObject);
    procedure FormShow(Sender: TObject);
  private
    FxVarName: string;
    FyVarName: string;
    procedure SetxVarName(const Value: string);
    procedure UpdateInstructions;
    procedure SetyVarName(const Value: string);
  public
    BaseDataChild: TForm;
    property xVarName: string read FxVarName write SetxVarName;
    property yVarName: string read FyVarName write SetyVarName;
  end;

var
  TransformationsDlg: TTransformationsDlg;

implementation

uses BaseDataC, Expr;

{$IFNDEF LCL_}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

const
  InstructionsText =
    'To transform one or both axes, enter the desired formulae in terms of '
    + '% and ! below. For example, to plot the natural log of ! against 1/% enter '
    + 'ln(!) in the "!" box and 1/% in the "%" box';

procedure TTransformationsDlg.NoneBtnClick(Sender: TObject);
begin
  xFunction.Text := xVarName;
  yFunction.Text := yVarName;
end;

procedure TTransformationsDlg.FormResize(Sender: TObject);
begin
  {$IFDEF LCL_}
  OKBtn.Width := CancelBtn.Width;
  NoneBtn.Width := CancelBtn.Width;
  Check.Width := CancelBtn.Width;
  {$ENDIF}
end;

procedure TTransformationsDlg.CheckClick(Sender: TObject);
var
  Lex: TLex;
  Expr: TExpr;
  ColVar: TVariable;
begin
  Lex := nil;
  Expr := nil;
  ColVar := nil;
  try
    ColVar := TVariable.Create('i');
    Lex := TLex.Create;
    Lex.Options := [xGlobalFunctions];
    with BaseDataChild as TBaseDataChild do begin
      DataSource.AddIndependentVariables(Lex);
      Lex.AddVariable(yVar);
    end;
    Lex.AddVariable(ColVar);
    Expr := TExpr.Create(Lex);
    Expr.Text := xFunction.Text;
    Expr.Text := yFunction.Text;
  finally
    Expr.Free;
    Lex.Free;
    ColVar.Free;
  end;
end;

procedure TTransformationsDlg.OKBtnClick(Sender: TObject);
begin
  CheckClick(Sender);
  ModalResult := mrOK;
end;

procedure TTransformationsDlg.SetxVarName(const Value: string);
begin
  FxVarName := Value;
  xLabel.Caption := xVarName;
  UpdateInstructions;
end;

procedure TTransformationsDlg.UpdateInstructions;
begin
  Instructions.Caption := StringReplace(InstructionsText, '%', xVarName, [rfReplaceAll]);
  Instructions.Caption := StringReplace(Instructions.Caption, '!', yVarName, [rfReplaceAll]);
end;

procedure TTransformationsDlg.SetyVarName(const Value: string);
begin
  FyVarName := Value;
  yLabel.Caption := yVarName;
  UpdateInstructions;
end;

procedure TTransformationsDlg.FormShow(Sender: TObject);
begin
  if xFunction.Text = '' then
    xFunction.Text := xVarName;
  if yFunction.Text = '' then
    yFunction.Text := yVarName;
end;

end.
