unit DrLoader;
{ Copyright (C) 2008 C M Western; see NLFit.pas for license}
interface

{ Loader for new drive files }

uses ExpPlotLoader, TextStrm, Channel;

type

  TDrFileLoader = class(TExperimentalPlotLoader)
  public
    function Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean; override;
  end;

implementation

uses SysUtils, DataTypes;

{ TDrFileLoader }

function TDrFileLoader.Load(var Log: TextFile; AskAboutWeights: Boolean): Boolean;
var
  i, c: Integer;
  Channels: TChannels;
  s: String;
begin
  Result := False;
  if (Ext = '') or (Ext = '.csv') or (Ext = '.dat') or (Ext = '.txt') then begin
    if IsDriveStream(Stream) then
      Ext := '.dr'
  end;
  if Ext <> '.dr' then
    Exit;
  Channels := TChannels.Create;
  Result := True;
  try
    Channels.LoadFromDriveStream(Stream,'',False);
    SetnChannels(Channels.nChannels);
    with Channels do
      for c := 0 to nChannels - 1 do
        with Channels[c] do
          for i := 0 to Count - 1 do
            AddChange(c,Value[i]);
    AutoDetectFrequency;
    CheckAvailableChannels(Channels.nChannels, Channels.nCollected, AskAboutWeights);
    with Channels do begin
      for c := 0 to nChannels - 1 do
        if ChannelDest[c] >= 0 then begin
          for i := 0 to Channels[c].Count - 1 do
            SetValue(-1,c,i,Channels[c].Value[i]);
          with Channels[c] do begin
            s := Name;
            Connector.SetNameN(ChannelDest[c],s);
            if DataSourceName <> '' then
              s := s + ' - ' + DataSourceName;
            if DataSourceInfo <> '' then
              s := s + ' - ' + DataSourceInfo;
          end;
          Connector.SetCommentN(ChannelDest[c],s);
        end;
      for c := 0 to nChannels - 1 do
        if ChannelDest[c] > 0 then
          Connector.SetLength(ChannelDest[c]-1,Channels[c].Count,False,True,True);
    end;
  finally
    Channels.Free;
  end;
end;

initialization
  RegisterPlotLoader(TDrFileLoader, 'Drive Files', '*.dr');
end.
