unit CalFrm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

interface

uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, NGrid, StdCtrls, ExtCtrls, ChildForm, MNode, ComCtrls,
  Overlay, Buttons, Gendlg, AlEdit, EditExtr, NewSpin, Menus, Settings, Units;


type
  TCalForm = class(TChildForm)
    Panel: TPanel;
    Grid: TNumberGrid;
    CopyButton: TSpeedButton;
    DeleteButton: TSpeedButton;
    DeleteUnassigned: TButton;
    AssignButton: TSpeedButton;
    NLfitButton: TSpeedButton;
    Label1: TLabel;
    FitButton: TSpeedButton;
    Label2: TLabel;
    UndoFitButton: TSpeedButton;
    OtherButton: TButton;
    OtherPopupMenu: TPopupMenu;
    ApplyItem: TMenuItem;
    ClearItem: TMenuItem;
    RangeItem: TMenuItem;
    ResetItem: TMenuItem;
    CopyFrequencyItem: TMenuItem;
    PolyOrder: TSpinBox;
    ActiveObject: TComboBox;
    AvgErrEdit: TAlignedEdit;
    EtalonItem: TMenuItem;
    SplineFitItem: TMenuItem;
    procedure CopyFrequencyItemClick(Sender: TObject);
    procedure ResetItemClick(Sender: TObject);
    procedure RangeItemClick(Sender: TObject);
    procedure OtherButtonClick(Sender: TObject);
    procedure ApplyButtonClick(Sender: TObject);
    procedure UndoFitButtonClick(Sender: TObject);
    procedure PolyOrderChange(Sender: TObject);
    procedure FitButtonClick(Sender: TObject);
    procedure NLfitButtonClick(Sender: TObject);
    procedure GridMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure GridDblClick(Sender: TObject);
    procedure GridClick(Sender: TObject);
    procedure AssignButtonClick(Sender: TObject);
    procedure GridDrawCell(Sender: TObject; ACol, ARow: Integer; Rect: TRect;
      State: TGridDrawState);
    procedure DeleteUnassignedClick(Sender: TObject);
    procedure ClearButtonClick(Sender: TObject);
    procedure DeleteButtonClick(Sender: TObject);
    procedure CopyButtonClick(Sender: TObject);
    procedure GridChangeCellQuery(Sender: TObject; Col, Row: Integer;
      var CanChange: Boolean);
    procedure GridSetCell(Sender: TObject; Col, Row: Integer; var v: Double;
      var Valid: Boolean);
    procedure EtalonItemClick(Sender: TObject);
    procedure FormHide(Sender: TObject);
    procedure SplineFitItemClick(Sender: TObject);
  private
    FChanging: Boolean;
    FAssigningPeak: Integer;
    FLastNPeaks: Integer;
    LastCell: TGridCoord;
    LastSortCol: Integer;
    EtalonSpacing: Double;
    DisableReplot: Boolean;
    procedure SetAssigningPeak(Value: Integer);
    function DefaultFunctionText(Order: Integer):string;
  protected
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure HandleControl(Control: TControl; var PName, Value, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure AddExtraVariables(AFormSettings: TFormSettings); override;
    procedure WriteExtraVariables(AFormSettings: TFormSettings); override;
  public
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); override;
    procedure ShowPeak(AIndex: Integer);
    procedure StuffSelector(MNode: TMNode); override;
    procedure PeakChanged(Sender:TPlot; PeakNumber:Integer);
    procedure PeakMeasured(Value: Double; AUnits: TResonanceUnits; Source: TMNode);
    procedure UpdateStatus;
    procedure ArrangeForm(Screen: TRect); override;
    procedure ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent = nil); override;
    property AssigningPeak: Integer read FAssigningPeak write SetAssigningPeak;
  end;

var
  CalForm: TCalForm;

function CalibratingPlot:TPlot;

implementation

uses {$IFDEF LCL_} LResources, Types, {$ENDIF}
     Global, Expr, Math, MathSup, Main, ExpPlot, PlotDataSource, ApplyFrm,
     ResidFrm, WCompat, Peaks, DataFitter;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

function CalibratingPlot:TPlot;
begin
  Result := nil;
  with CalForm do
    if Visible and (CurrentMNode is TPlot) then
      Result := CurrentMNode as TPlot;
end;

{ TCalForm }

function TCalForm.BaseNodeClass: TClass;
begin
  Result := TOverlays;
end;

procedure TCalForm.ClearButtonClick(Sender: TObject);
begin
  if CurrentMNode is TPlot then
    (CurrentMNode as TPlot).ClearPeaks;
  MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.CopyButtonClick(Sender: TObject);
begin
  Grid.CopyToClipboard;
end;

procedure TCalForm.CopyFrequencyItemClick(Sender: TObject);
begin
  ApplyButtonClick(nil);
end;

function TCalForm.DefaultFunctionText(Order: Integer): string;
var
  i: Integer;
begin
  case Order of
    0: Result := 'x+c'; { Simple offset }
    1: Result := 'm*x+c';
  else
    Result := 'poly(x,c,m';
    for i := 2 to Order do
      Result := Result + ',a'+IntToStr(i);
    Result := Result + ')';
  end;
end;

procedure TCalForm.DeleteButtonClick(Sender: TObject);
var
  d, i: Integer;
begin
  Grid.BeginUpdate;
  DisableReplot := True;
  try
    d := Grid.Selection.Top;
    for i := d to Grid.Selection.Bottom do
      (CurrentMNode as TPlot).DeletePeak(d-1);
  finally
    DisableReplot := False;
    Grid.EndUpdate;
  end;
  MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.DeleteUnassignedClick(Sender: TObject);
var
  i: Integer;
begin
  Grid.BeginUpdate;
  DisableReplot := True;
  try
    if CurrentMNode is TPlot then
      with CurrentMNode as TPlot do
        for i := nPeaks-1 downto 0 do
        if IsNan(Peaks[i].Actual) then
          DeletePeak(i);
  finally
    DisableReplot := False;
    Grid.EndUpdate;
  end;
  MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.EtalonItemClick(Sender: TObject);
var
  s: string;
begin
  s := FloatToStr(EtalonSpacing);
  if Dialogs.InputQuery('Find equally spaced peaks', 'Enter Spacing', s) then begin
    EtalonSpacing := StrToFloatEx(s);
    try
      Grid.BeginUpdate;
      DisableReplot := True;
      (CurrentMNode as TExperimentalPlot).Etalon(EtalonSpacing);
    finally
      DisableReplot := False;
      Grid.EndUpdate;
    end;
    AssigningPeak := -1;
    MainForm.ConsiderReplot(False,False);
  end;
end;

procedure TCalForm.FitButtonClick(Sender: TObject);
var
  OldLinearFit, OldUseSVD: Boolean;
begin
  if CurrentMNode is TExperimentalPlot then begin
    with (CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource, DataSourceFitter do
      try
        { Higher order polynomials give "sensible" results with these settings }
        OldLinearFit := LinearFit;
        LinearFit := True;
        OldUseSVD := UseSVD;
        UseSvd := False;
        NLFitDlg.FitOnceClick(nil);
      finally
        LinearFit := OldLinearFit;
        UseSVD := OldUseSVD;
      end;
    MainForm.ResidualsItemClick(nil);
  end;
end;

procedure TCalForm.FormHide(Sender: TObject);
begin
  Grid.EditorMode := False;
  MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.GridChangeCellQuery(Sender: TObject; Col, Row: Integer;
  var CanChange: Boolean);
begin
  CanChange := CanChange and (Col <> 3);
end;

procedure TCalForm.GridClick(Sender: TObject);
begin
  AssigningPeak := -1;
end;

procedure TCalForm.GridDblClick(Sender: TObject);
begin
  if LastCell.Y = 0 then begin
    { Double click on column heading to sort on that column }
    if CurrentMNode is TPlot then
      (CurrentMNode as TPlot).SortPeaks(LastCell.X-1, LastCell.X <> LastSortCol);
    if LastSortCol = LastCell.X then
      LastSortCol := 0
    else
      LastSortCol := LastCell.X; 
  end else if LastCell.X = 2 then
    AssigningPeak := LastCell.Y-1
end;

procedure TCalForm.GridDrawCell(Sender: TObject; ACol, ARow: Integer;
  Rect: TRect; State: TGridDrawState);
begin
  if (AssigningPeak >=0) and (ACol = 2) and (AssigningPeak = ARow-1) then
    ColouredTextInGrid(Grid.Canvas, Rect, 'Measure', clRed);
end;

procedure TCalForm.GridMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  LastCell := Grid.MouseCoord(X,Y);
end;

procedure TCalForm.GridSetCell(Sender: TObject; Col, Row: Integer;
  var v: Double; var Valid: Boolean);
var
  P: TPeak;
  function Update(var OldValue: Double; NewValue: Double): Boolean;
  begin
    Result := (IsNanOrInf(OldValue) xor IsNanOrInf(NewValue))
              or (OldValue <> NewValue);
    if Result then
      OldValue := NewValue;
  end;
begin
  if not Valid then
    SetNan(v);
  if CurrentMNode is TPlot then
    with CurrentMNode as TPlot do begin
      if (Row >= nPeaks) and not IsNanOrInf(v) then
        nPeaks := Row;
      P := Peaks[Row-1];
      case Col of
        1: Modified := Update(P.Position,v);
        2: Modified := Update(P.Actual,v);
        4: Modified := Update(P.Weight,v);
      end;
      try
        FChanging := True;
        Peaks[Row-1] := P;
      finally
        FChanging := False;
      end;
    end;
  AssigningPeak := -1;
end;

procedure TCalForm.HandleControl(Control: TControl; var PName, Value,
  PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
begin
  if (Control = ActiveObject) or (Control = PolyOrder) then
    inherited;
end;

procedure TCalForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin

end;

procedure TCalForm.PeakChanged(Sender: TPlot; PeakNumber: Integer);
var
  AutoAssign: Boolean;
begin
  if FChanging or not Visible then
    Exit;
  if Sender = CurrentMNode then
    with Sender do begin
      Modified := True;
      AutoAssign := FLastNPeaks = nPeaks-1;
      if (PeakNumber < 0) or not Grid.Enabled or (Grid.RowCount <> nPeaks+1) then
        ShowItem(Sender)
      else
        ShowPeak(PeakNumber);
      if AutoAssign then
        AssigningPeak := nPeaks-1;
    end;
end;

procedure TCalForm.PeakMeasured(Value: Double; AUnits: TResonanceUnits; Source: TMNode);
var
  Peak: TPeak;
begin
  if FChanging or not Visible then
    Exit;
  if (AssigningPeak >= 0) and (CurrentMNode <> Source) then
    with CurrentMNode as TPlot do begin
      Peak := Peaks[AssigningPeak];
      Peak.Actual := Convert(Value, AUnits, Units);
      Peak.Residual := Peak.Actual-ConvertPeakPosition(Peak.Position,Units);
      Peaks[AssigningPeak] := Peak;
    end;
end;

procedure TCalForm.PolyOrderChange(Sender: TObject);
begin
  if (PolyOrder.Value >= 0) and (CurrentMNode is TExperimentalPlot) then
    with ((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg do begin
      FunctionText := DefaultFunctionText(PolyOrder.Value);
      PlotFunctionClick(nil);
    end;
end;

procedure TCalForm.RangeItemClick(Sender: TObject);
var
  Peak: TPeak;
begin
  if CurrentMNode is TExperimentalPlot then
    with CurrentMNode as TExperimentalPlot do begin
      Peak := EmptyPeak;
      Peak.Weight := 1000;
      if Length(OriginalFrequency) > 0 then begin
        Peak.Position := OriginalFrequency[0];
        AddPeak(Peak);
        Peak.Position := OriginalFrequency[High(OriginalFrequency)];
      end else begin
        Peak.Position := Frequency[0];
        AddPeak(Peak);
        Peak.Position := Frequency[High(Frequency)];
      end;
      AddPeak(Peak);
    end;
  AssigningPeak := -1;
end;

procedure TCalForm.ResetItemClick(Sender: TObject);
begin
  if CurrentMNode is TExperimentalPlot then
    with CurrentMNode as TExperimentalPlot do begin
      if Length(OriginalFrequency) > 0 then
        Frequency := OriginalFrequency;
        OriginalFrequency := nil;
        FrequencyScale := 1;
        FrequencyOffset := 0;
      MainForm.ConsiderReplot(False,False);
    end;
end;

procedure TCalForm.SetAssigningPeak(Value: Integer);
begin
  if not (CurrentMNode is TPlot) then
    Value := -1
  else
    Value := Min(Value,(CurrentMNode as TPlot).nPeaks-1);
  AssignButton.Down := Value >= 0;
  if Value = AssigningPeak then
    Exit;
  FAssigningPeak := Value;
  Grid.Invalidate;
end;

procedure TCalForm.SettingChanged(MNode: TMNode; Index: Integer);
begin

end;

procedure TCalForm.ShowItem(MNode: TMNode);
var
  i, n, poly: Integer;
  ft: string;
begin
  { Update Node display }
  Grid.EditorMode := False;
  SetItemIndex(ActiveObject,MNode);
  CurrentMNode := MNode;
  AssigningPeak := -1;
  FLastNPeaks := 0;
  UpdateStatus;
  Caption := 'Calibrating';
    poly := 1;
  if MNode is TPlot then begin
    Caption := CurrentMNode.Name + ' - ' + Caption;
    n := (MNode as TPlot).nPeaks;
    if MNode is TExperimentalPlot then
      with TExperimentalPlot(MNode) do
        if HasDataSource then begin
          poly := -1;
          ft := LowerCase(Trim((DataSource as TInteractiveDataSource).NLFitDlg.FunctionText));
          for i := PolyOrder.MinValue+1 to PolyOrder.MaxValue do
            if ft = DefaultFunctionText(i) then
              poly := i;
        end;
  end else
    n := 0;
  if n = 0 then begin
    Grid.RowCount := 2;
    Grid.Clear;
    Grid.Enabled := False;
  end else begin
    Grid.Enabled := True;
    Grid.RowCount := n+1;
    for i := 0 to n-1 do
      ShowPeak(i);
  end;
  with PolyOrder do
    if poly <> Value then begin
      OnChange := nil; { Disable function update }
      Value := poly;
      OnChange := {$IFDEF FPC}@{$ENDIF}PolyOrderChange;
    end;
  if not DisableReplot then
    MainForm.ConsiderReplot(False,False);
end;

procedure TCalForm.ShowMenuItemHelp(MenuItem: TMenuItem; TopItem: TComponent);
begin
  if TopItem = OtherPopupMenu then
    TopItem := Self;
  inherited;
end;

procedure TCalForm.ShowPeak(AIndex: Integer);
begin
  AssigningPeak := -1;
  FLastNPeaks := -2;
  if CurrentMNode is TPlot then
    with TPlot(CurrentMNode) do begin
      FLastNPeaks := nPeaks;
      with Peaks[Aindex] do begin
        Grid[AIndex+1,1] := Position;
        Grid[AIndex+1,2] := Actual;
        Grid[AIndex+1,3] := Residual;
        Grid[AIndex+1,4] := Weight;
      end;
    end;
end;

procedure TCalForm.SplineFitItemClick(Sender: TObject);
begin
((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg.SplineItemClick(nil);
end;

procedure TCalForm.AddExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
begin
  if EtalonSpacing = 0 then
    EtalonSpacing := 1;
  inherited;
  Index := AFormSettings.IndexOfParameter('EtalonSpacing');
  if Index < 0 then
    Index := AFormSettings.AddParameter('EtalonSpacing', 1);
  AFormSettings.SetParameter(Index, EtalonSpacing);
end;

procedure TCalForm.ApplyButtonClick(Sender: TObject);
var
  Plot: TMNode;
  NotUsed: Boolean;
  i: Integer;
begin
  Plot := PromptForOverlay('Apply Calibration To',NotUsed,'',True,TExperimentalPlot);
  if Assigned(Plot) then begin
    if Plot is TPlot then
      (CurrentMNode as TExperimentalPlot).ApplyCalibration(Plot as TExperimentalPlot,Sender = nil)
    else begin { Apply to all }
      for i := 0 to GlobalOverlays.ComponentCount - 1 do
        if (GlobalOverlays.Components[i] <> CurrentMNode)
           and (GlobalOverlays.Components[i] is TExperimentalPlot) then
            (CurrentMNode as TExperimentalPlot).ApplyCalibration(GlobalOverlays.Components[i] as TExperimentalPlot,Sender = nil);
      { Doing the current one last }
      (CurrentMNode as TExperimentalPlot).ApplyCalibration(CurrentMNode as TExperimentalPlot,Sender = nil);
    end;
    MainForm.ConsiderReplot(False,False);
  end;
end;

procedure TCalForm.ArrangeForm(Screen: TRect);
var
  R, MainRect: TRect;
begin
  inherited;
  R := ScreenBounds(Self);
  MainRect := ScreenBounds(MainForm);
  OffsetRect(R,MainRect.Right-R.Left,MainRect.Top-R.Top);
  if R.Right > Screen.Right then
    OffsetRect(R,Screen.Right-R.Right,0);
  if R.Bottom > Screen.Bottom then
    OffsetRect(R,0,Screen.Bottom-R.Bottom);
  SetScreenBounds(Self,R);
end;

procedure TCalForm.AssignButtonClick(Sender: TObject);
begin
  if AssigningPeak >= 0 then
    AssigningPeak := -1
  else
    AssigningPeak := Grid.Row-1;
end;

procedure TCalForm.NLfitButtonClick(Sender: TObject);
begin
  if CurrentMNode is TExperimentalPlot then
    ReShow(((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg);
end;

procedure TCalForm.OtherButtonClick(Sender: TObject);
begin
  OtherPopupMenu.PopupComponent := OtherButton;
  OtherPopupMenu.Popup(Mouse.CursorPos.X, Mouse.CursorPos.Y);
end;

function TCalForm.StuffObject(const Name1, Name2: string; MNode: TMNode;
  var ParentNode: TTreeNode): Integer;
begin
  if MNode is TPlot then begin
    ActiveObject.AddItem(Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end else
    Result := 0;
end;

procedure TCalForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TCalForm.UndoFitButtonClick(Sender: TObject);
begin
  if CurrentMNode is TExperimentalPlot then
    ((CurrentMNode as TExperimentalPlot).DataSource as TInteractiveDataSource).NLFitDlg.UndoClick(nil);
end;

procedure TCalForm.UpdateStatus;
begin
  if not Visible or (csDestroying in Application.ComponentState) then
    Exit;
  if CurrentMNode is TExperimentalPlot then
    with CurrentMNode as TExperimentalPlot do
      if HasDataSource then
        with DataSource as TPlotDataSource do begin
          UndoFitButton.Enabled := NLFitDlg.Undo.Enabled;
          if IsNan(AvgErr) then
            AvgErrEdit.Text := ''
          else
            AvgErrEdit.Text := DtosM(AvgErr, 8, -1);
          Exit;
        end;
  UndoFitButton.Enabled := False;
  AvgErrEdit.Text := '';
end;

procedure TCalForm.WriteExtraVariables(AFormSettings: TFormSettings);
var
  Index: Integer;
begin
  inherited;
  Index := AFormSettings.IndexOfParameter('EtalonSpacing');
  if Index > 0 then
    EtalonSpacing := AFormSettings.Parameters[Index];
end;

end.
