unit Global;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses {$IFNDEF CONSOLE} Forms, {$ENDIF}
     Mixture, Overlay, MNode, Units, UCompat;

function DecodePgopherVersion(VersionName: string):Integer;
var
  VersionInfo: string;

{$IFNDEF CONSOLE}
procedure ReShow(Form: TCustomForm);
function GlobalMixture: TMixture;
function GlobalOverlays: TOverlays;
function GlobalLineList: TLineList;

procedure LoadGeneralFile(AFileName: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
procedure LoadAnyObservations;
{$ELSE}
var
  GlobalMixture: TMixture;
  GlobalOverlays: TOverlays;
{$ENDIF}
procedure AddOverlay(AFileName: string; Merge:Boolean; DefaultUnits: TResonanceUnits);

procedure PeakMeasured(Position: Double; Units: TResonanceUnits; Lab: string; Source: TMNode);

type
  TFileType = (fiParameters, fiObservations, fiOverlay);
function FileType(AFileName: string; PreferOverlay: Boolean): TFileType;

procedure ShowInfo(const Msg: string);
procedure ReShowLogForm;

type

  TPlottedPoint = record
    X, Y:  Single;  { Plot Position }
    Index: Integer; { Observation number }
    Extra: {$IFDEF FPC}PtrInt{$ELSE}Integer{$ENDIF}; { Information }
  end;

  TPlottedPoints = array of TPlottedPoint;

  TOnGetPointInfo = function (Index: Integer; Extra: PtrInt):string of object;

implementation

uses {$IFNDEF CONSOLE}
        CalFrm, Huge, Clipbrd, LineFrm, Main, PlotFrm, Dialogs, LogUnit,
        {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$ENDIF}
     SysUtils, Classes, Hitran, Worker;

function DecodePgopherVersion(VersionName: string):Integer;
var
  p: Integer;
  InNumber: Boolean;
  LastNumber: Integer;
  nNumbers: Integer;
begin
  Result := 0;
  InNumber := False;
  LastNumber := 0;
  VersionName := VersionName + ' ';
  nNumbers := 0;
  for p := 1 to Length(VersionName) do
    if VersionName[p] in ['0'..'9'] then begin
      InNumber := True;
        LastNumber := LastNumber*10 + Ord(VersionName[p])-Ord('0')
    end else if InNumber then begin
      InNumber := False;
      Result := Result*1000 + LastNumber;
      LastNumber := 0;
      Inc(nNumbers);
      if nNumbers >=3 then
        Exit;
    end;
  for p := nNumbers+1 to 3 do
    Result := Result*1000;
end;

procedure AddOverlay(AFileName: string; Merge:Boolean; DefaultUnits: TResonanceUnits);
begin
  {$IFNDEF CONSOLE}
    if Assigned(MainForm) then
      MainForm.AddOverlay(AFileName, Merge, DefaultUnits);
  {$ELSE}
    AddToOverlay(GlobalOverlays, AFileName, Merge, DefaultUnits, False, Arbitrary);
  {$ENDIF}
end;

{$IFNDEF CONSOLE}
procedure LoadGeneralFile(AFileName: string; Merge, PreferOverlay: Boolean; DefaultUnits: TResonanceUnits);
begin
  if Assigned(MainForm) then
    MainForm.LoadGeneralFile(AFileName, Merge, PreferOverlay, DefaultUnits);
end;

procedure LoadAnyObservations;
begin
  if Assigned(MainForm) then
    MainForm.LoadAnyObservations;
end;

function GlobalMixture: TMixture;
begin
  if Assigned(MainForm) then
    Result := MainForm.Mixture
  else
    Result := nil;
end;

function GlobalOverlays: TOverlays;
begin
  if Assigned(MainForm) then
    Result := MainForm.Overlays
  else
    Result := nil;
end;

function GlobalLineList: TLineList;
var
  Mixture: TMixture;
begin
  Mixture := GlobalMixture;
  if Assigned(Mixture) then
    Result := Mixture.LineList
  else
    Result := nil;
end;

procedure ReShow(Form: TCustomForm);
begin
  if not Form.Visible then
    Form.Show;
  if Form.WindowState in [wsMinimized, wsMaximized] then
    Form.WindowState := wsNormal;
  Form.BringToFront;
end;

{$ENDIF}

function IsMixtureFile(AFileName: string):Boolean;
var
  Buffer: string[100];
begin
  Result := False;
  try
    with TFileStream.Create(AFileName,fmOpenRead or fmShareDenyWrite) do
      try
        SetLength(Buffer, Read(Buffer[1], 100));
        Result := Pos('<Mixture ',Buffer) > 1;
      finally
        Free;
      end;
  except
  end;
end;

function FileType(AFileName: string; PreferOverlay: Boolean): TFileType;
var
  Ext: string;
begin
  Ext := LowerCase(ExtractFileExt(AFileName));
  if (Ext = '.pgo')
     or ((Ext = '.par') and not IsHITRANfile(AFileName))
     or ((Ext = '.xml') and IsMixtureFile(AFileName))
     or (Ext = '.var') or (Ext = '.svw') or (Ext = '.jbw') or (Ext = '.sta') then
    Result := fiParameters
  else if (Ext = '.lin') or (Ext = '.obs') or (((Ext = '.cat') or (Ext = '.mrg')) and not PreferOverlay) then
    Result := fiObservations
  else
    Result := fiOverlay;
end;

{$IFNDEF CONSOLE}
var
  CF_Measurement: DWORD;
{$ENDIF}

procedure PeakMeasured(Position: Double; Units: TResonanceUnits; Lab: string; Source: TMNode);
{$IFNDEF CONSOLE}
var
  Stream: THugeMemoryStream;
{$ENDIF}
begin
  {$IFNDEF CONSOLE}
  if CF_Measurement = 0 then
    CF_Measurement := RegisterClipboardFormat('Measurement');
  Stream := THugeMemoryStream.Create;
  (* Mimic the following C structure:
    typedef struct {
      double value;
      char signature[4];    /* Normally MEAS */
      char filename[1];     /* Null terminated file name; can be longer */
    } ClipboardMeasurement ;
  *)
  Stream.Write(Position, Sizeof(Double));
  Stream.Write('MEAS', 4);
  if Lab <> '' then
    Stream.Write(Lab[1], Length(Lab)+1)
  else
    Stream.Write(#0#0, 1);
  { Copy to clipboard }
  Clipboard.Open;
  try
    Clipboard.AsText := FloatToStr(Position);
    Stream.CopyToClipboard(CF_Measurement);
  finally
    Clipboard.Close;
    Stream.Free;
  end;
  if Assigned(CalForm) then
    CalForm.PeakMeasured(Position, Units, Source);
  if Assigned(LineForm) then
    LineForm.AssignLines(Position, Units, Lab, Source, False);
  {$ENDIF}
end;

procedure ShowInfo(const Msg: string);
begin
  {$IFDEF CONSOLE}
  WriteLn(Log,Msg);
  {$ELSE}
  ShowMessage(Msg);
  {$ENDIF}
end;

procedure ReShowLogForm;
begin
  {$IFNDEF CONSOLE}
  ReShow(LogForm);
  {$ENDIF}
end;

end.
