unit HundA;

{ Hund's Case (a) basis functions }

interface
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}

uses SysUtils, Classes, MNode, Mixture, QuantumNumber, MatrixElements, Obs, pgDOM, MathSup, QuantumMath;

type

  TLinearMolecule = class;

  { TLinear }

  TLinear = class(TState)
  private
    FMolecule: TLinearMolecule;
    FInverted: Boolean;
  public { Functions and properties specific to TLinear }
    function HRot(J, Lamp, Sigp, Omp, Lam, Sig, Om: TQuantumNumber): Double;
    function NRn(n: Integer; J, Lam, Spin, Omp, Om: TQuantumNumber): Double;
    property Molecule: TLinearMolecule read FMolecule;
    property Inverted: Boolean read FInverted;
  public
    constructor Create(AOwner: TComponent); override;
    class function SpinSpinType: TSpinSpinClass; override;
    function FixupProperties: Exception; override;
    function AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer; override;
    function AllowedSphericalComponent(TransitionMoment: TSphericalTransitionMoment; State: TState; out Component: TQuantumNumber):Exception; override;
    function GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function HyperfineMatrixElement(ABra, AKet: TBasisState; Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex; override;
    function RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex; override;
    function SymbolicRotationalMatrixElement(AIndex: Integer):TMatrixElements; override; { For given parameter }
    function MatrixElementType(const Changes: array of Integer; const H: string):string; override;
    function ConvertParameter(AIndex: Integer; Zero: Boolean):TPerturbation; override;
    function RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber):Complex; override;
    function RotationalPerturbationMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber):Complex; override;
    function StateOrigin: Double; override; { Estimate lowest J=0 energy - Used to stop Boltzman factors blowing up at low temperatures }
    function SettingToStr(AIndex: Integer): string; override;
    function SettingToPrettyStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const Str: string); override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function HelpOnParameter(AIndex: Integer): string; override;
    function ReadBranchFormat(Text: string; var Obs: TObservation; UpperState: TState; Reader: TObservationsReaderBase):Boolean; override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
    procedure AdjustDefaults; override;
  published
    property B:        Double index  LSP   read GetParameter write SetParameter;
    property A:        Double index  LSP+1 read GetParameter write SetParameter;
    property LambdaSS: Double index  LSP+2 read GetParameter write SetParameter;
    property o:        Double index  LSP+3 read GetParameter write SetParameter;
    property theta:    Double index  LSP+4 read GetParameter write SetParameter; { S >= 2 only }
    property gamma:    Double index  LSP+5 read GetParameter write SetParameter;
    property p:        Double index  LSP+6 read GetParameter write SetParameter;
    property q:        Double index  LSP+7 read GetParameter write SetParameter;
    { Centrifugal Distortion }
    property D:        Double index  LSP+8 read GetParameter write SetParameter;
    property H:        Double index  LSP+9 read GetParameter write SetParameter;
    property L:        Double index  LSP+10 read GetParameter write SetParameter;
    property M:        Double index  LSP+11 read GetParameter write SetParameter;
    property PP: { P } Double index  LSP+12 read GetParameter write SetParameter;
    { Centrifugal Distortion of A }
    property AD:       Double index LSP+13 read GetParameter write SetParameter;
    { Centrifugal Distortion of Lambda }
    property LambdaD:  Double index  LSP+14 read GetParameter write SetParameter;
    property LambdaH:  Double index  LSP+15 read GetParameter write SetParameter;
    { Centrifugal Distortion of lambda doubling }
    property oD:       Double index  LSP+16 read GetParameter write SetParameter;
    property oH:       Double index  LSP+17 read GetParameter write SetParameter;
    property oL:       Double index  LSP+18 read GetParameter write SetParameter;
    property pD:       Double index  LSP+19 read GetParameter write SetParameter;
    property pH:       Double index  LSP+20 read GetParameter write SetParameter;
    property pL:       Double index  LSP+21 read GetParameter write SetParameter;
    property qD:       Double index  LSP+22 read GetParameter write SetParameter;
    property qH:       Double index  LSP+23 read GetParameter write SetParameter;
    property qL:       Double index  LSP+24 read GetParameter write SetParameter;
    { Centrifugal distortion of gamma }
    property gammaD:   Double index  LSP+25 read GetParameter write SetParameter;
    property gammaH:   Double index  LSP+26 read GetParameter write SetParameter;
    property gammaL:   Double index  LSP+27 read GetParameter write SetParameter;

    property Lambda:    TQuantumNumber  index LSI read GetSetting write SetSetting; { Use -1 for sigma- }
    property gerade:    Boolean         index LSI+1 read GetFlag write SetFlag;
    property OmegaSelect: TQuantumNumber index LSI+2 read GetSetting write SetSetting;
  end;

  TLinearNucleus = class(TNucleus)
  public
    function HelpOnParameter(AIndex: Integer): string; override;
    function HasIdentical: Boolean; override;
  published
    property a:        Double index  1 read GetParameter write SetParameter;
    property b:        Double index  2 read GetParameter write SetParameter;
    property c:        Double index  3 read GetParameter write SetParameter;
    property d:        Double index  4 read GetParameter write SetParameter;
    property eQq0:     Double index  5 read GetParameter write SetParameter;
    property eQq2:     Double index  6 read GetParameter write SetParameter;
    property cI:       Double index  7 read GetParameter write SetParameter;
    { Centrifugal distortion of the above.
      Note that these only have an outline implementation, and may
      not be the same as other programs }
    property aJ:       Double index  8 read GetParameter write SetParameter;
    property bJ:       Double index  9 read GetParameter write SetParameter;
    property cJ:       Double index 10 read GetParameter write SetParameter;
    property dJ:       Double index 11 read GetParameter write SetParameter;
    property eQq0J:    Double index 12 read GetParameter write SetParameter;
    property eQq2J:    Double index 13 read GetParameter write SetParameter;
  end;

  TLinearSpinSpin = class(TSpinSpin)
    property J:        Double index  1 read GetParameter write SetParameter;
    property S:        Double index  2 read GetParameter write SetParameter;
  end;

  { TLinearBasisState }

  TLinearBasisState = class(TBasisState)
  public
    Omega: TQuantumNumber;
    Fn: Integer; { Indicates F1, F2, F3 .... Not guaranteed for J < S }
    function EstimateN: TQuantumNumber; { Best guess at N - computed from Fn }
    function Parity: Integer;
    function St: TLinear;
    function efLabel: string;

    function TrackIndex: Integer; override;
    function BasisLabel: string; override; { Basis State Label }
    function BasisLabelKey: string; override; { Explanation of Basis State Label }
    function StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string; override; { State name that goes with the given basis state
                                            (assuming match between basis state and final state is OK) }
    function StateLabelKey: string; override; { Explanation of State Name }
    function Qno(QnoType: TQnoType): TQuantumNumber; override; { extra quantum numbers - Omega, Fn }
    function MakeSignature(SignA:Integer; Ket: TBasisState; SignB: Integer): TIntegerArray; override;
    procedure SetSymbolicVariables(SignA:Integer; Ket: TBasisState; SignB: Integer); override;
    function TxName(Upper, Lower: TBasisState): string; override;    { Make transition label such as R(3) }
    function TxNameKey(Upper, Lower: TBasisState): string; override; { String explaining transition label }
  end;

  TLinearManifold = class(TManifold)
  public
    function FixupProperties: Exception; override;
    function nSym: Integer; override; { Number of different symmetries }
  end;

  TLinearMolecule = class(TMolecule)
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties: Exception; override;
    function IsValidChildClass(MNodeClass: TMNodeClass):Boolean; override;
    function Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean; override;
    function JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer; override;
    function LastSym: Integer; override;
    function SymName(Sym: Integer):string; override;
    function JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string; override;
    function QnoInfo(QnoType: TQnoType): TQnoInfo; override;
    function IsShowingN: Boolean;
    function HelpOnSetting(AIndex: Integer): string; override;
    procedure OmegaRange(out Smallest, Largest: TQuantumNumber);
  published
    property Symmetric: Boolean index LMI   read GetFlag write SetFlag; { True for D infinity h }
    { Statistical Weights for symmetric molecules }
    property SymWt:     Integer index LMI+1 read GetSetting write SetSetting;
    property AsymWt:    Integer index LMI+2 read GetSetting write SetSetting;
    { If true, used the R^2 Hamiltonian rather than the default N^2 form }
    property RSquaredH: Boolean index LMI+3 read GetFlag write SetFlag;
    { What quantum numbers to (try and) show }
    property ShowJ:       Boolean index LMI+4 read GetFlag write SetFlag;
    property ShowOmega:   Boolean index LMI+5 read GetFlag write SetFlag;
    property ShowN:       Boolean index LMI+6 read GetFlag write SetFlag;
    property ShowFNumber: Boolean index LMI+7 read GetFlag write SetFlag;
    property Showef:      Boolean index LMI+8 read GetFlag write SetFlag;
  end;

  TLinearOperator = (LS, Luncouple, Suncouple, Homog, NS, RS, LNuncouple, aL, e, eQq1);

  TLinearPerturbation = class (TPerturbation)
  private
    procedure SetOperator(const Index: Integer; const AValue: TLinearOperator);
    function GetOperator(const Index: Integer): TLinearOperator;
  public
    constructor Create(AOwner: TComponent); override;
    function FixupProperties: Exception; override;
    function PairName: string; override;
    function SymbolicMatrixElements:TMatrixElements; override;
    function HelpOnSetting(AIndex: Integer): string; override;
    function IsSettingQuantum(AIndex: Integer):Boolean; override;
    function SettingToStr(AIndex: Integer): string; override;
    procedure SetSettingFromStr(AIndex: Integer; const s: string); override;
    function ReadProperties(Node: TDOMNode; CheckUnread: Boolean = True):Integer; override;
  published
    property Op:TLinearOperator    index LPI read GetOperator write SetOperator;
    property n:Integer             index LPI+1 read GetSetting write SetSetting;
    property Npower:Integer        index LPI+2 read GetSetting write SetSetting;
    property OmegaSelect: TQuantumNumber index LPI+3 read GetSetting write SetSetting;
    property Srank: TQuantumNumber index LPI+4 read GetSetting write SetSetting;
    property Scomp: TQuantumNumber index LPI+5 read GetSetting write SetSetting;
  end;

implementation

uses Math, nj, Expr, njExpr, Profile, Worker, StaticField;

var
  eJ: TSortedVariableNode; { Note that these are not thread safe }

{ TLinearBasisState }

function TLinearBasisState.BasisLabel: string;
begin
  Result := Format('%s %2s', [inherited BasisLabel, State.Manifold.Mixture.QuantumNumberToStr(Omega)])
            + HyperfineStateLabel;
end;

function TLinearBasisState.BasisLabelKey: string;
begin
  Result := inherited BasisLabelKey + ' Omega' + HyperfineStateLabelKey;
end;

function TLinearBasisState.EstimateN: TQuantumNumber;
var
  OmLow, nSkip: TQuantumNumber;
begin
  if S = 0 then
    Result := J
  else with St do begin
    if Lambda > 0 then begin
      if Inverted then
        OmLow := S+Lambda
      else
        OmLow := S-Lambda;
      nSkip := OmLow - Min(OmLow, J);
      Result := Max(Abs(J-S),Lambda) + 2*Fn - 2 - nSkip;
    end else
      { Sigma States: Note that these two expressions are equivalent, unless J < S }
      if Inverted then
        Result := J + S + 2*(Fn - (S+1))
      else
        Result := Abs(J-S) + 2*(Fn-1);
        // Result := Max(Abs(J-S),Lambda) + 2*(Fn-1)
    {$IFOPT R+}
    if not IsTriangle(J, S, Result) then
      raise Exception.CreateFmt('TLinearBasisState.N: %d %d %d', [J, S, Result]);
    {$ENDIF}
  end;
end;

function TLinearBasisState.MakeSignature(SignA: Integer; Ket: TBasisState; SignB: Integer): TIntegerArray;
var
  KetK: TLinearBasisState;
begin
  SetLength(Result,4);
  Result[0] := Max(St.Lambda,0)*SignA;
  Result[1] := Omega*SignA;
  KetK := Ket as TLinearBasisState;
  Result[2] := Max(KetK.St.Lambda,0)*SignB;
  Result[3] := KetK.Omega*SignB;
end;

procedure TLinearBasisState.SetSymbolicVariables(SignA: Integer; Ket: TBasisState; SignB: Integer);
begin
  eJ.Value := Ket.J / 2;
end;

function TLinearBasisState.Parity: Integer;
begin
  Result := 1 - Sym*2;
end;

function TLinearBasisState.St: TLinear;
begin
  Result := State as TLinear;
end;

function TLinearBasisState.efLabel: string;
begin
  if ((J and 2) <> 0) xor (Sym = 1) then
    Result := 'f'
  else
    Result := 'e';
end;

function TLinearBasisState.StateLabel(AQuantumNumberFormat: TQuantumNumberFormat): string;
begin
  Result := inherited StateLabel(AQuantumNumberFormat);
  with State.Manifold.Molecule as TLinearMolecule do begin
    if ShowJ then
      Result := Result + Format(' %2s', [QuantumNumber.QuantumNumberToStr(J,AQuantumNumberFormat)]);
    if IsShowingN then
      Result := Result + Format(' %2s', [QuantumNumber.QuantumNumberToStr(N,AQuantumNumberFormat)]);
    if not AllSinglet then begin
      if ShowOmega then
        Result := Result + Format(' %2s', [QuantumNumber.QuantumNumberToStr(Omega,AQuantumNumberFormat)]);
      if ShowFNumber then
        Result := Result + Format(' F%d', [Fn]);
    end;
    if Showef then begin
      if not (ShowFNumber and not AllSinglet) then
        Result := Result + ' ';
      Result := Result + efLabel;
    end;
  end;
  Result := Result + HyperfineStateLabel;
end;

function TLinearBasisState.StateLabelKey: string;
begin
  Result := inherited StateLabelKey;
  with  State.Manifold.Molecule as TLinearMolecule do begin
    if ShowJ then
      Result := Result + '  J';
    if IsShowingN then
      Result := Result + '  N';
    if not AllSinglet then begin
      if ShowOmega then
        Result := Result + ' Om';
      if ShowFNumber then
        Result := Result + ' Fn';
    end;
    if Showef then
      Result := Result + ' p';
  end;
  Result := Result + HyperfineStateLabelKey;
end;

function TLinear.ReadBranchFormat(Text: string; var Obs: TObservation;
  UpperState: TState; Reader: TObservationsReaderBase): Boolean;
var
  DeltaN, DeltaJ: TQuantumNumber;
  FU, FL, JU, JL, sp: Integer;
  HyperfineL, HyperfineU: THyperfineBasis;
  OrgName, ErrMsg: string;
  AltObs: TObservation;
  HasFields: Boolean;
  function TrySym(var AObs: TObservation; ASym: Integer):Boolean;
  var
    Lname: string;
  begin
    AObs.Name := OrgName;
    Lname := Manifold.FindStateLabel(AObs.IndexL, AObs.SymL, AObs.JL,
                                     [Low(TQuantumNumber),FL], AObs.QuantumNumberFormat, AObs.ExField,
                                     ASym, JL, -1, Self, Reader.RemoveSpins,HyperfineL);
    if AObs.SymL >= 0 then begin
      if Odd(Reader.BranchRank div 2) then
        AObs.SymU := 1-AObs.SymL
      else
        AObs.SymU := AObs.SymL;
      AObs.Name := AObs.Name + ' : '
                   + UpperState.Manifold.FindStateLabel(AObs.IndexU, AObs.SymU, AObs.JU,
                                                       [Low(TQuantumNumber),FU], AObs.QuantumNumberFormat, AObs.ExField,
                                                       AObs.SymU, JU, -1, UpperState, Reader.RemoveSpins, HyperfineU)
                   + ' - ' + Lname;
    end else
      Aobs.Name := AObs.Name + LName;
    Result := (AObs.SymL >=0) and (AObs.SymU >= 0);
  end;
begin
  Result := False;
  Obs.Name := ' ' + Text;
  if not IsDeltaJName(Text[1],DeltaN) then
    Exit
  else
    Delete(Text,1,1);
  if not IsDeltaJName(Text[1],DeltaJ) then
    DeltaJ := DeltaN
  else
    Delete(Text,1,1);
  if Text[1] in ['1'..'9'] then begin
    FU := StrToInt(Text[1]);
    Delete(Text,1,1);
    if Text[1] in ['1'..'9'] then begin
      FL := StrToInt(Text[1]);
      Delete(Text,1,1);
    end else
      FL := FU;
  end else begin
    FL := 1;
    FU := 1;
  end;
  if Text[1] in ['e','f'] then begin
    Obs.SymL := Ord(Text[1])-Ord('e');
    Delete(Text,1,1);
  end else
    Obs.SymL := -1;
  sp := Pos(')',Text);
  if (Text[1] <> '(') or (sp <= 2) then
    Exit;
  if Reader.ReadQuantumNumber(JL,Copy(Text,2,sp-2)) then
    Exit;
  Delete(Text,1,sp);
  if (Obs.SymL >= 0) and Odd(JL div 2) then
    Obs.SymL := 1-Obs.SymL;
  JU := JL+DeltaJ;
  HasFields := IsNonZero(Obs.ExField);
  if HasFields or (Molecule.nNuclei > Reader.RemoveSpins) then begin
    { Some hyperfine structure to be found }
    if ReadHyperfineBranch(Text, Obs, HyperfineU, HyperfineL, Reader, HasFields) then
      Exit;
  end else if Text <> '' then
    Exit
  else begin
    { Set total angular momentum }
    Obs.JL := JL;
    Obs.JU := JU;
  end;
  OrgName := Obs.Name;
  Result := TrySym(Obs, Obs.SymL);
  if not Result and (Obs.IndexL = -3) and (Obs.SymL < 0) then begin
    { Two matches - try both }
    ErrMsg := Obs.Name;
    Result := TrySym(Obs, 0);
    if Result then begin
      AltObs := Obs;
      if TrySym(AltObs,1) then
        if Reader.DefaultSymmetry >= 0 then begin
          { Two matches - use default symmetry }
          { TODO: Should offer taking both as a possibility }
          if Obs.SymL = Molecule.JAdjustedSym(Obs.JL,Reader.DefaultSymmetry) then
            Obs := AltObs;
        end else
          Obs.Name := ErrMsg; { Default symmetry not set - record initial message }
    end else
      Result := TrySym(Obs,1);
  end;
  if Result then begin
    Obs.IndexU := 0;
    Obs.IndexL := 0;
  end;
end;

function TLinear.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  ParSel: string;
begin
  Result := inherited ReadProperties(Node, False);
  { For backwards copatibility }
  ParSel := ReadValue(Node, 'ParitySelect', '');
  if (ParSel <> '') and (ParSel <> 'all') then
    if ParSel = 'e_only' then
      RveSelect := 0
    else if ParSel = 'f_only' then
      RveSelect := 1
    else
      raise EReportException.Create('Bad ParitySelect in '+Name+': '+ParSel);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

function TLinearBasisState.TxName(Upper, Lower: TBasisState): string;
var
  U, L: TLinearBasisState;
  US, LS: TLinear;
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  U := Upper as TLinearBasisState;
  US := U.State as TLinear;
  L := Lower as TLinearBasisState;
  LS := L.State as TLinear;
  Result := DeltaJName(U.J-L.J);
  if Molecule.IsShowingN then
    Result := LowerCase(DeltaJName(U.N-L.N)) + Result;
  if (US.S > 0) or (LS.S > 0) then begin
    if U.Fn <> L.Fn then
      Result := Result + IntToStr(U.Fn);
    Result := Result + IntToStr(L.Fn);
  end;
  if (US.Lambda > 0) and (LS.Lambda > 0) then
    Result := Result + L.efLabel;
  Result := Result + Format('(%s)',[Molecule.QuantumNumberToStr(L.J)]);
  if (Length(Lower.HyperfineState) > 0) or (Owner as TEigenvalues).HasFields then
    Result := Result + Upper.HyperfineStateLabel(False) + ',' + Lower.HyperfineStateLabel(False);
  if Length(Result) < 8 then
    Result := StringOfChar(' ',8-Length(Result)) + Result;
end;

function TLinearBasisState.TxNameKey(Upper, Lower: TBasisState): string;
var
  U, L: TLinearBasisState;
  US, LS: TLinear;
  Molecule: TLinearMolecule;
begin
  Molecule := State.Manifold.Molecule as TLinearMolecule;
  U := Upper as TLinearBasisState;
  US := U.State as TLinear;
  L := Lower as TLinearBasisState;
  LS := L.State as TLinear;
  Result := 'deltaJ';
  if Molecule.IsShowingN then
    Result := 'deltaN ' + Result;
  if (US.S > 0) or (LS.S > 0) then begin
    Result := Result + ' (F'') F"';
  end;
  if (US.Lambda > 0) and (LS.Lambda > 0) then
    Result := Result + ' e/f"';
  Result := Result + '(J")' + HyperfineTxNameKey(False);
end;

function TLinearBasisState.TrackIndex: Integer;
begin
  Result := Fn-1;
end;

function TLinearBasisState.Qno(QnoType: TQnoType): TQuantumNumber;
begin
  case QnoType of
    qtA: Result := Omega;
    qtB: Result := Fn;
  else
    Result := inherited Qno(QnoType);
  end;
end;

{ TLinear }

constructor TLinear.Create(AOwner: TComponent);
begin
  if Assigned(AOwner) then
    FMolecule := AOwner.Owner as TLinearMolecule;
  SavedProperties := SavedProperties + ',gerade=True,OmegaSelect=all';
  inherited;
end;

function TLinear.AddOneJ(Eigenvalues: TEigenvalues; J: TQuantumNumber; Sym: Integer; HyperfineState:THyperfineBasis; Seffective: TQuantumNumber; out StatWtGroup: Integer): Integer;
var
  Sigma, Dir, Nmin: TQuantumNumber;
  FNumber, FNumberInc, StartIndex, nAdded, i, PauliPhase: Integer;
  procedure Add(Omega: TQuantumNumber);
  var
    BasisState: TLinearBasisState;
    NewPhase: Integer;
  begin
    if (J < Abs(Omega)) or ((OmegaSelect > -9999) and (Omega <> OmegaSelect)) then Exit;
    if (Lambda <= 0) and (Omega = 0) then
      NewPhase := 0
    else begin
      NewPhase := Minus(J-Seffective+Sym*2);
      if Lambda < 0 then
        NewPhase := - NewPhase;
    end;
    BasisState := TLinearBasisState.Create(Eigenvalues,J,Sym);
    BasisState.State := Self;
    BasisState.J := J;
    BasisState.S := Seffective;
    BasisState.Omega := Omega;
    BasisState.HyperfineState := HyperfineState;
    BasisState.StateIndex := Eigenvalues.LastStateIndex;
    BasisState.Phase := NewPhase;
  end;
begin
  { Statistical Weight }
  if Molecule.Symmetric then begin
    if gerade xor (Sym > 0) then begin
      Result := Molecule.SymWt;
      StatWtGroup := 1;
    end else begin
      Result := Molecule.AsymWt;
      StatWtGroup := 2;
    end;
    if Length(HyperfineState) > 0 then begin
      PauliPhase := 0;
      { Look for equivalent nuclei }
      for i := 0 to High(HyperfineState) do
        if Molecule.Nuclei[i].AsNext then begin
          if PauliPhase = 0 then
            if gerade xor (Sym > 0) then
              PauliPhase := 1
            else
              PauliPhase := -1;
          PauliPhase := PauliPhase * Minus(HyperfineState[i]);
        end;
      if PauliPhase < 0 then
        Exit;
    end;
  end else begin
    Result := 1;
    StatWtGroup := 1;
  end;
  if Result = 0 then
    Exit;
  MatchOdd(Seffective, J);
  StartIndex := Eigenvalues.Size;
  if Lambda <= 0 then begin
    { Sigma states }
    { In general have N = J-S, J-S+1, ... J+S.
                 and Fn = F1,  F2,    ... F(2S+1) }
    Nmin := Abs(J-Seffective);
    { However, for any parity, only Even or Odd N will be present }
    if (Minus(Nmin) < 0) xor (Sym = 1) xor (Lambda < 0) then
      Inc(Nmin,2);
    Sigma := SmallestProjection(Seffective);
    if not Odd(Seffective) and ((Lambda < 0) xor (Minus(J-Seffective) < 0) xor (Sym = 1)) then
      { Even multiplicity - may be missing Sigma = 0 }
      Sigma := 2;
    nAdded := (Seffective-Sigma) div 2 + 1; { The number added for all but the lowest J }
    while Sigma <= Seffective do begin
      Add(Sigma);
      Inc(Sigma, 2);
    end;
    { In general, lowest F number will be 1 or 2 - work it out from lowest N }
    FNumber := (Nmin-J-Seffective) div 2 + Seffective + 1;
    { For the case where not all 2S+1 N values are present,
      the above will omit the lowest F numbers; this is correct if we
      want to strictly derive the F number from N. This gives an odd
      correlation with energy though; for regular states the missing levels
      are the higher energy levels, which correspond to the higher F numbers. }
    if not Inverted and (nAdded > Eigenvalues.Size - StartIndex) { Some missing } then
       { Want F1 or F2 as the lowest F number }
       FNumber := 2 - FNumber mod 2;
    FNumberInc := 2;
  end else begin
    { As a speed up, try adding in rising energy order }
    if Inverted then
      Dir := -1
    else
      Dir := 1;
    Sigma := -Dir*Seffective - 2*Dir;
    FNumber := 1;
    FNumberInc := 1;
    { Note special case here - normally have
      F1, F2, ... Fn
      and loose either the lowest or the highest F numbers for J < Lambda + S.
      However, For S > Lambda actually loose some at both ends }
    repeat
      Inc(Sigma, 2*Dir);
      Add(Lambda+Sigma);
      nAdded := Eigenvalues.Size - StartIndex;
      if nAdded < 1 then
        Inc(FNumber);
    until Sigma = Dir*Seffective;
  end;
  Eigenvalues.SortFrom(StartIndex,1e100);
  for i := StartIndex to Eigenvalues.Size-1 do begin
    with Eigenvalues.Basis[i+1] as TLinearBasisState do begin
      { TODO: Logic needs adjusting for OmegaSelect > -9999 }
      Fn := FNumber;
      N := EstimateN;
    end;
    Inc(FNumber, FNumberInc);
  end;
  Inc(Eigenvalues.LastStateIndex);
end;

{ SetupRmat calculates the matrices of even powers of R or N in rn.
  To allow efficient use of these, the most recent result is cached }

const
  MaxRmatPower = 12; { Maximum power of R/N to calculate }
type
  TRMat = array of array of Double;
threadvar
  JCurrent, LambdaCurrent, SCurrent:TQuantumNumber;
  RSquaredHCurrent, RMatInitialised: Boolean;
  OmegaMin: TQuantumNumber; { Omega range covered by current matrix }
  nOmegas: Integer;
  rn: array [1..MaxRmatPower div 2] of TRMat;

procedure MatSymMul(a, b, c: TRMat);
{ Matrix product a = b*c of SYMMETRIC matrices }
var
  i, j, k: Integer;
  Sum: Double;
begin
  for i := 0 to High(a) do
    for j := 0 to i do begin
      Sum := 0;
      for k := 0 to High(a) do
        Sum := Sum + b[i,k]*c[j,k];
      a[j,i] := Sum;
      a[i,j] := Sum;
    end;
end;

procedure SetupRmat(J, Lambda, S: TQuantumNumber; RSquaredH: Boolean);
var
  Omega, OmegaMax: TQuantumNumber;
  i, ii: Integer;
  sjl: Double;
begin
  if not RMatInitialised or (RSquaredHCurrent <> RSquaredH)
     or (JCurrent <> J) or (LambdaCurrent <> Lambda) or (SCurrent <> S) then begin
    { Recalculation needed }
    RMatInitialised := True;
    RSquaredHCurrent := RSquaredH;
    JCurrent := J;
    LambdaCurrent :=Lambda;
    SCurrent := S;
    OmegaMin := Lambda - S;
    if OmegaMin < - J then
      OmegaMin := -J;
    OmegaMax := Lambda + S;
    if OmegaMax > J then
      OmegaMax := J;
    nOmegas := (OmegaMax - OmegaMin) div 2 + 1;
    if { nOmegas > Length(rn[1]) } True { TODO: use expression }then
      for i := 1 to High(rn) do
        SetLength(rn[i], nOmegas, nOmegas);
    { Form matrix of R**2 }
    { Diagonal Elements }
    Omega := OmegaMin;
    sjl := j*(j+2) + s*(s+2);
    if not RSquaredH then
      sjl := sjl + lambda*lambda;
    for i := 0 to nOmegas-1 do begin
      rn[1,i,i] := (sjl - omega*omega - Sqr(omega-lambda)) / 4;
      Inc(Omega,2);
    end;
    { Off diagonal by 1 }
    Omega := OmegaMin + 2;
    for i := 1 to nOmegas-1 do begin
      rn[1,i,i-1] :=
        - Sqrt( (J*(J+Two) - Omega * (Omega-2))
                *(S*(S+2) - (Omega-Lambda) * (Omega-Lambda-2)) )/4;
      rn[1,i-1,i] := rn[1,i,i-1];
      Inc(Omega,2);
    end;
    { All other elements zero }
    for i := 2 to nOmegas-1 do
      for ii := 0 to i-2 do begin
        rn[1,i,ii] := 0;
        rn[1,ii,i] := 0;
      end;
    { Now generate higher powers }
    for i := 2 to MaxRmatPower div 2 do
      MatSymMul(rn[i],rn[i-1],rn[1]);
  end;
end;

function TLinear.NRn(n: Integer; J, Lam, Spin, Omp, Om: TQuantumNumber): Double;
{ Matrix Elements of N^n or R^n, depending on state of RSquaredH }
var
  Ompp, OmMax: TQuantumNumber;
begin
  if Odd(n) then
    raise EReportException.Create('Linear.Nrn:NPower must be even:'+IntToStr(n));
  if Abs(Om-Omp) = n then
    Result := Minus(n) * rl(J, Omp, Om) * rl(Spin, Lam-Omp, Lam-Om)
  else if (Om = Omp) and (n = 2) then begin
    Result := J*(J+2) + Spin*(Spin+2) - Om*Om - Sqr(Lam-Om);
    if not Molecule.RSquaredH then
      Result := Result + Lam*Lam;
    Result := Result / 4;
  end else if Abs(Omp-Om) < n then begin
    { Handle using recursion. There may be scope for caching results, as is done
      for SetupRmat above }
    Result := 0;
    Ompp := Max(Omp - 2, Lam-Spin);
    OmMax := Min(Omp + 2, Lam+Spin);
    while Ompp <= OmMax do begin
      if Abs(Ompp) <= J then
        Result := Result + NRn(2,J,Lam,Spin,Omp,Ompp)*NRn(n-2,J,Lam,Spin,Ompp,Om);
      Inc(Ompp,2);
    end;
  end else
    Result := 0;
end;

function TLinear.RotationalMatrixElement(ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMnode; Flag: Integer): Complex;
var
  Bra, Ket: TLinearBasisState;
  Lam, Lamp, Om, Omp: TQuantumNumber;
begin
  Bra := ABra as TLinearBasisState;
  Ket := AKet as TLinearBasisState;
  Result := z0;
  if (Bra.Sym <> Ket.Sym) or (Bra.J <> Ket.J) then
    raise Exception.Create('Bad TLinear.RotationalMatrixElement call');
  if SignA < 0 then begin
    if SignB > 0 then begin
      Bra := TLinearBasisState(AKet);
      Ket := TLinearBasisState(ABra);
    end;
    SignA := -SignA;
    SignB := -SignB;
  end;
  if Lambda < 0 then begin
    Lamp := 0;
    Lam := 0;
  end else begin
    Lamp := Lambda*SignA;
    Lam := Lambda*SignB;
  end;
  Omp := Bra.Omega*SignA;
  Om := Ket.Omega*SignB;
  case TMatElemType(Flag) of
    meHrot: Result.re := HRot(Bra.J, Lamp, Omp-Lamp, Omp, Lam, Om-Lam, Om);
    meWidth: begin { TODO: Add more operators? Make a perturbation type?
                     If adding more operators, adjust HasWidth }
               if (Bra = Ket) and (SignA = SignB) then
                 Result.re := Width
             end;
  else
    raise NotYet;
  end;
end;

function TLinear.HRot(J, Lamp, Sigp, Omp, Lam, Sig, Om: TQuantumNumber): Double;
{ Matrix elements in an unsymmetrised case (a) basis. }
var
  i, ii, ir, iir, irmax: Integer;
begin
  Result := 0;
  if Lamp < 0 then
    raise Exception.Create('HRot: states out of order');
  if Lamp = Lam then begin
    { Prepare R^2n and the indices into it }
    SetupRmat(J, Lamp, S, Molecule.RSquaredH);
    i := (Omp-OmegaMin) div 2;
    ii := (Om-OmegaMin) div 2;
    if Sig = Sigp then begin
      Result := Origin
                + A*Lamp*sig/4 { ALzSz }
                { 2/3 lambda (3Sz^2-S^2) }
      	        + 2/3 * LambdaSS*(3*Sig*Sig - S*(S+2)) / 4
                { 1/12 theta (35Sz^4-30S^2Sz^2+25Sz^2-6S^2+3S^4) }
                + theta *(35*Sig*Sig*Sig*Sig/16
                          -30*S*(S+2)*Sig*Sig/16
                          +25*Sig*Sig/4 - 6*S*(S+2)/4
                          + 3*S*S*(S+2)*(S+2)/16)/12;
    end;
    { BR^2 - DR^4 + HR^6 + LR^8 + MR^10 + PR^12 }
    Result := Result + B*rn[1,i,ii] - D*rn[2,i,ii]
                     + H*rn[3,i,ii] + L*rn[4,i,ii]
                     + M*rn[5,i,ii] + PP*rn[6,i,ii];
    { 1/2 AD (LzSzR^2 + R^2LzSz) }
    Result := Result + AD*rn[1,i,ii]*(Lam*Sig + Lamp*Sigp)/8;
    { gamma N . S }
    if Sig = Sigp then
      Result := Result + gamma * (Om*Sig - S*(S+2))/4
    else if Abs(Sig-Sigp) = 2 then
      Result := Result + gamma * rl(J,Omp,Om)*rl(S,Sigp,Sig)/2;
    { 1/2 [NzS,R**2 + ...]+ }
    Result := Result + (gammaD*rn[1,i,ii] + gammaH*rn[2,i,ii] + gammaL*rn[3,i,ii])
                       * ((Om*Sig+Omp*Sigp)/2 - S*(S+2))/4;
    { 1/2 N+-R**2 + ... }
    if i > 0 then
      Result := Result + (gammaD*rn[1,i-1,ii] + gammaH*rn[2,i-1,ii] + gammaL*rn[3,i-1,ii])
                         * rl(J,Omp,Omp-2) * rl(S,Sigp,Sigp-2) / 4;
    if i < nOmegas-1 then
      Result := Result + (gammaD*rn[1,i+1,ii] + gammaH*rn[2,i+1,ii] + gammaL*rn[3,i+1,ii])
                         * rl(J,Omp,Omp+2) * rl(S,Sigp,Sigp+2) / 4;
    { 1/2 (R**2 + ...) N+- }
    if ii > 0 then
      Result := Result + (gammaD*rn[1,i,ii-1] + gammaH*rn[2,i,ii-1] + gammaL*rn[3,i,ii-1])
                         * rl(J,Om-2,Om) * rl(S,Sig-2,Sig) / 4;
    if ii < nOmegas-1 then
      Result := Result + (gammaD*rn[1,i,ii+1] + gammaH*rn[2,i,ii+1] + gammaL*rn[3,i,ii+1])
                         * rl(J,Om+2,Om) * rl(S,Sig+2,Sig) / 4;
    { 2/3 lambdaD 1/2 (3Sz^2-s^2) + similar higher terms }
    Result := Result + 2/3 * (LambdaD * rn[1,i,ii] + LambdaH * rn[2,i,ii])
                       * (3*(Sig*Sig+Sigp*Sigp)/2- S*(S+2)) / 4;
  end else if Abs(Lamp-Lam) = 4 then begin
    { Prepare R^2n and the indices into it }
    SetupRmat(J, Lamp, S, Molecule.RSquaredH);
    { To be able to use rn for lambda < 0, must look up opposite
      sign of omega. This assumes lambdap >= 0 }
    { indices to use with lambdap }
    i := (Omp-OmegaMin) div 2;
    { for -lambda=lambdap, sigma, omega+4 }
    ii := (Lamp+Sig-OmegaMin) div 2;
    { indices to use with lambda }
    ir := (-(Lam+Sigp)-OmegaMin) div 2;
    iir := (-Om-OmegaMin) div 2;
    { imax := (OmegaMax-OmegaMin) div 2; = nOmegas-1 }
    irmax := (-Max(-J,Min(J,Lam-S))-OmegaMin) div 2;
    { 1/2 (o+p+q) (S+**2 + S-**2) }
    if Om = Omp then
      Result := Result + (o+p+q)*rl(S,Sigp,Sig) / 2;
    { 1/4 (oD+pD+qD) [R**2, S+**2 + S-**2] }
    { (S+**2 does not contribute as this requires lambdap=lambda-2) }
    { R**2 S+**2
      if ii < nOmegas then
        Result := Result + alphaD*rn[1,i,ii+2]*rl(S, Sig+4, Sig)/4; }
    { (R**2 + R**4 + R**6) S-**2 }
    if (ii > 1) and (ii-1 <= nOmegas) then
      Result := Result + ((oD+pD+qD)*rn[1,i,ii-2] + (oH+pH+qH)*rn[2,i,ii-2] + (oL+pL+qL)*rn[3,i,ii-2])
                         *rl(S, Sig-4, Sig)/4;
    { S-**2 (R**2 + R**4 + R**6) }
    if (ir > 1) and (ir-2 <= irmax) then
      Result := Result + ((oD+pD+qD)*rn[1,ir-2,iir] + (oH+pH+qH)*rn[2,ir-2,iir] + (oL+pL+qL)*rn[3,ir-2,iir])
                         * rl(S,Sigp,Sigp+4) / 4;
    { S+**2 R**2
      if ir+2 < nOmegas then
        Result := Result + alphaD*rl(S,Sigp,Sigp-4)*rn[1,ir+2,iir]/4; }
    { -1/2 (p+2q) (J-S- + J-S-) }
    { Only S- contributes }
    if Sigp = Sig-2 then
      Result := Result - rl(S,Sigp,Sig)*(p+2*q)*rl(J,Omp,Om) / 2;
    { -1/2 (pD+2qD) (R**2 + R**4 + R**6) J-S- }
    if (ii > 0) and (ii <= nOmegas) then
      Result := Result - rl(J,Om+2,Om) * rl(S,Sig-2,Sig) / 4
                         * ( (pD+2*qD)*rn[1,i,ii-1] + (pH+2*qH)*rn[2,i,ii-1] + (pL+2*qL)*rn[3,i,ii-1] );
    { -1/2 (pD+2qD) J-S- (R**2 + R**4 + R**6) }
    if (ir > 0) and (ir <= nOmegas) then
      Result := Result - rl(J,Omp,Omp-2) * rl(S,Sigp,Sigp+2) / 4
                         * ( (pD+2*qD)*rn[1,ir-1,iir] + (pH+2*qH)*rn[2,ir-1,iir] + (pL+2*qL)*rn[3,ir-1,iir] );
    { 1/2 q (J+**2 + J-**2) }
    { Only J- contributes }
    if Sig = Sigp then
      Result := Result + q * rl(J, Omp, Om) / 2;
    { 1/4 (R**2 + R**4 + R**8) J-**2 }
    if (ii >= 0) and (ii < nOmegas) then
      Result := Result + rl(J,Om+4,Om) * ( qD*rn[1,i,ii] + qH*rn[2,i,ii] + qL*rn[3,i,ii] ) / 4;
    { 1/4 J-**2 (R**2 + R**4 + R**8) }
    if (ir >= 0) and (ir <= irmax) then
      Result := Result + rl(J,Omp,Omp-4) * ( qD*rn[1,ir,iir] + qH*rn[2,ir,iir] + qL*rn[3,ir,iir] ) / 4;
  end else if Abs(Lam-Lamp) = 8 then begin
    { Lambda Doubling in a Delta state
      As this rountine is always called with lambdap positive, only need exp(4i phi) term }
    if Sigp = Sig-8 then
      { 1/2 m(S+**4 + S-**4) }
      Result :=  o { = m } * rl(S,Sig,SigP) / 2
    else if Sigp = Sig-6 then
      { -1/2 n(S+**3J+ + S-**3J-) }
      Result := - oD { = n } * rl(S,Sig,SigP) * rl(j,Om+2,Om) / 2
    else if Sigp = Sig-4 then
      { 1/2 o(S+**2J+**2 + S-**2J-^**2) }
      Result := oH { = o } * rl(S,Sig,SigP) * rl(j,Om+4,Om) / 2
    else if Sigp = Sig-2 then
      { -1/2 p(S+J+**3 + S-J-**3) }
      Result := - p * rl(S,Sig,SigP) * rl(j,Om+6,Om) / 2
    else if Sigp = Sig then
      { 1/2 q(J+**4 + J-**4) }
      Result := q * rl(j,Omp,Om) / 2
  end;
end;

function TLinear.StateOrigin: Double;
var
  SigmaStep, Sigma: TQuantumNumber;
  E: Double;
begin
  { TODO: Call Hrot, rather than a partial evaluation;
          would ensure it still works if all parameters converted to perturbations }
  Result := Origin;
  { Estimate lowest spin orbit + spin-spin energy;
    ( Hrot(J=Lambda+S) may also be suitable, but more complicated )
    Just consider sigma = -s, 0 and s states }
  SigmaStep := Max(S,2);
  Sigma := -S;
  while Sigma <= S do begin
    E := Origin + A*Lambda*Sigma/4 { ALzSz }
                { 2/3 lambda (3Sz^2-S^2) }
      	        + 2/3 * LambdaSS*(3*Sigma*Sigma - S*(S+2)) / 4;
    if E < Result then
      Result := E;
    Inc(Sigma, SigmaStep);
  end;
end;

function TLinear.GetSettingsNames(AIndex: Integer; out LowestValue:Integer): PPChar;
const
  LambdaNames: array [0..14] of PChar = ('Sigma-', 'Sigma+', '', 'Pi', '', 'Delta', '', 'Phi', '', 'Gamma', '', 'H', '', 'I', nil);
begin
  Result := inherited GetSettingsNames(AIndex,LowestValue);
  if AIndex = LSI then begin
    Result := @LambdaNames[0];
    LowestValue := -1;
  end;
end;

function TLinear.HyperfineMatrixElement(ABra, AKet: TBasisState;
  Signp, Sign: Integer; MNode: TMNode; RankWanted: TQuantumNumber): Complex;
var
  Bra, Ket: TLinearBasisState;
  JP, OmegaP, LamP, SigmaP, J, Omega, Lam, Sigma, dOm: TQuantumNumber;
  Nuc: TLinearNucleus;
  jjp1: Double;
begin
  Bra := ABra as TLinearBasisState;
  Ket := AKet as TLinearBasisState;
  Nuc := MNode as TLinearNucleus;
  { This routine works for any sign of Lambda.
    Note that swapping trick used for Hrot will not work here, as it would
    require swapping the F quantum numbers which are evaluated elsewhere.
    Can't use sigma(xz) either, as we only have a part of the overall
    Hamiltonian operator here, so we would need to reflect that also }
{ This code will do the swap and adjust the phase for some cases,
  but may not be completely general
  Phase := 1;
  if Signp < 0 then begin
    if Sign > 0 then begin
      Bra := TLinearBasisState(B);
      Ket := TLinearBasisState(A);
    end;
    Phase := Minus(Bra.J-Ket.J);
    SignP := -SignP;
    Sign  := -Sign;
  end; }
  JP := Bra.J;
  OmegaP := SignP*Bra.Omega;
  LamP := SignP*Max(Lambda,0);
  SigmaP := OmegaP - LamP;
  J := Ket.J;
  Omega := Sign*Ket.Omega;
  Lam := Sign*Max(Lambda,0);
  Sigma := Omega - Lam;
  Result := z0;
  if RankWanted = 2 then begin
    { Magnetic Hyperfine }
    { Work out average J for use in centrifugal distortion }
    { TODO: Use [,N^2]+ instead }
    jjp1 := (J + Jp)*(J + JP + 2) / 16;
    if LamP = Lam then begin
       dOm := OmegaP - Omega;
       if SigmaP = Sigma then
         Result.re := (Nuc.a + Nuc.aJ*jjp1)*Lam/2;
       Result.re := Result.re + (Nuc.b + Nuc.c/3 + (Nuc.bJ+Nuc.cJ/3)*jjp1)
                                        *Minus(S - SigmaP)
                                        *threej(S, -SigmaP, 2, dOm, S, Sigma)
                                        *Sqrt(S*(S+2)*(S+One)/4);
       Result.re := Result.re + (Nuc.c + Nuc.cJ*jjp1)
                                *Minus(dOm + S - SigmaP)
                                *threej(S, -SigmaP, 2, dOm, S, Sigma)
                                *threej(2, -dOm, 4, 0, 2, dOm)
                                *Sqrt(30*S*(S+2)*(S+One)/4)/3;
       Result.re := Result.re*Minus(Jp - OmegaP) * threej(Jp, -OmegaP, 2, dOm, J, Omega);
       if (Sigma = SigmaP) and (J = JP) then
         Result.re := Result.re + Nuc.cI*Sqrt( J * (J+Two) / (Jp+1) / 4 );
    end else if Abs(LamP-Lam) = 4 then begin
       dOm := SigmaP - Sigma;
       Result.re := (Nuc.d + Nuc.dJ*jjp1)
                 *Minus(JP - OmegaP + S - SigmaP)
                 *threej(JP, -OmegaP, 2, -dOm, J, Omega)
                 *threej(S, -SigmaP, 2, dOm, S, Sigma)
                 *Sqrt(S*(S+2)*(S+One)/4);
    end else
      Exit;
    Result.re := Result.re * Sqrt(Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1)*(JP+One)/4);
  end else if RankWanted = 4 then begin
    { Quadrupole }
    if Sigma <> SigmaP then
      Exit;
    { Work out average J for use in centrifugal distortion }
    jjp1 := (J + Jp)*(J + JP + 2) / 16;
    dOm := OmegaP - Omega;
    if dOm = 0 then
      Result.re := Nuc.eQq0/2 + Nuc.eQq0J*jjp1
    else if Abs(dOm) = 4 then
      Result.re := Nuc.eQq2*Sqrt(Six)/12 + Nuc.eQq2J*jjp1;
    if Result.re <> 0 then
      Result.re := Result.re { * minus(j + i + f) * sixj(f, i, j, 4, j p, i) }
                      * Minus(J + Omega + dOm) * Sqrt((J + 1) * (JP + One))
                      * threej(4, -dOm, J, -Omega, JP, OmegaP)
                      /( 2 * threej(Nuc.Spin, -Nuc.Spin, 4, 0, Nuc.Spin, Nuc.Spin) );
  end else
    raise NotYet;
end;

function TLinear.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [LSI, LSI+2];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TLinear.RotationalTransitionMatrixElement(ABra, AKet: TBasisState; SignA,
  SignB: Integer; TransitionMoment: TMNode; F: TQuantumNumber): Complex;
var
  Comp: TQuantumNumber;
  Bra, Ket: TLinearBasisState;
  JP, OmegaP, LamP, SigmaP, J, Omega, Lam, Sigma: TQuantumNumber;
  Switch: Boolean;
  mm: Double;
begin
  Bra := ABra as TLinearBasisState;
  Ket := AKet as TLinearBasisState;
  JP := Bra.J;
  OmegaP := SignA*Bra.Omega;
  LamP := SignA*Max(Bra.St.Lambda,0);
  SigmaP := OmegaP - LamP;
  J := Ket.J;
  Omega := SignB*Ket.Omega;
  Lam := SignB*Max(Ket.St.Lambda,0);
  Sigma := Omega - Lam;
  Result := z0;

  { TODO: More spin forbidden transitions? }
  if TransitionMoment is TRotationalTransitionMoment then begin
    if (J = Jp) and (OmegaP = Omega) and (SigmaP = Sigma) then  { TODO: Document that this is gJ rather than gN; component never used }
      { <J||J||J> }
      Result.re := TRotationalTransitionMoment(TransitionMoment).Strength * Sqrt(J*(J+2)*(J+One)/4);
    Exit;
  end else if TransitionMoment is TSphericalTransitionMoment then begin
    Comp := TSphericalTransitionMoment(TransitionMoment).ActualComponent;
    if TransitionMoment is TSpinTransitionMoment then begin
      if (Abs(SigmaP-Sigma) <> Abs(Comp)) or (Lam <> Lamp) then
        Exit;
      Switch := (Comp <> OmegaP-Omega) or ((Comp = 0) and (Sigma < 0));
      case Abs(Comp) of
        0: Result.re := Abs(Sigma)/2; { Sign of simga is taken care of in switching }
        2: if (Bra.S = Ket.S) or (Bra.S <= 2) then
             Result.re := -rl(Bra.S,SigmaP,Sigma)/Sqrt(Two) { S+-; }
           else
             raise NotYet; { TODO: Use Wigner-Eckart to get right scaling Sigma dependence for off diagonal in Spin }
      else
        raise NotYet;
      end;
    end else begin
      if Abs(Omega-OmegaP) <> Comp then
        Exit;
      if (Bra.S = Ket.S) and (Sigma <> Sigmap) then
        Exit; { Only enforce Delta Sigma = 0 rule for spin allowed transitions }
      Switch := (Comp <> OmegaP-Omega) or ((Comp = 0) and (Lam < 0));
      Result.re := 1;
      { Note that Lambda dependence is not included in matrix element }
    end;
    { If the other sign of Comp = q is required, apply the following symmetry
      relationship to relate the T(k,q) and T(k,-q) matrix elements:
      E* <S' SIGMA' LAMBDA' sym'| T(k,q) | S SIGMA LAMBDA sym>
       = (-1)^(S'-OMEGA'+sym') <S' -SIGMA' -LAMBDA' sym'|
             (-1)^q T(k,-q)
             (-1)^(S-OMEGA+sym) | S -SIGMA -LAMBDA sym>
       = (-1)^(S'-S+sym'-sym) <S' -SIGMA' -LAMBDA' sym'| T(k,-q) | S -SIGMA -LAMBDA sym>
    }
    if Switch then begin
      Result.re := Result.re*Minus(Bra.S-Ket.S);
      if (Bra.St.Lambda < 0) xor (Ket.St.Lambda < 0) then
        Result.re := -Result.re;
      if TTransitionMoment(TransitionMoment).Magnetic then
        Result.re := Result.re * Minus(TTransitionMoment(TransitionMoment).Rank); { Different rule for reflecting magnetic operators }
    end;
    with TransitionMoment as TBaseRotationalTransitionMoment do begin
      mm := (Jp*(Jp+2)+ J*(J+2))/8;
      Result.re := Result.re*Strength*Sqrt(1 + C*mm + D*mm*mm);
    end;
  end else if TransitionMoment is TNuclearMagneticMoment then begin
    if (JP=J) and (OmegaP=Omega) and (SigmaP=Sigma) then
      Result.re :=1;
    Exit;
  end else
    raise NotYet;
  Result.re := Result.re*Sqrt((J+1)*(Jp+1))
                        * Minus(Jp-Omegap) * threej(Jp, -Omegap, TTransitionMoment(TransitionMoment).Rank, OmegaP-Omega, J, Omega);
  //Was:          *Minus(J-Omega+Comp)
  //Was:          *threej(Jp, OmegaP, TTransitionMoment(TransitionMoment).Rank, Omega-OmegaP, J, -Omega);
  { Note that eq 11 in Dixon, Bayley and Ashfold paper is incomplete, as
    it assumes the M and K are projections of the same quantum number, which is
    not the case if hyperfine is included. }
end;

function DeterminePhase(out ChangePhase, Swap:Boolean;
                        OmP, LamP, Om, Lam: TQuantumNumber;
                        Bra, Ket: TLinear):Integer;
begin
  { For off diagonal matrix elements define matrix element to be that
    LamP > Lam and LamP > 0 (or SigmaP > Sigma and SigmaP > 0 if Lambda=LambdaP = 0)
    and use symmetry and hermiticity to relate to other elements. }
  if LamP > Lam then begin
    ChangePhase := LamP <= 0;
    Swap := ChangePhase;
  end else if LamP < Lam then begin
    ChangePhase := Lam <= 0;
    Swap := not ChangePhase;
  end else if LamP <> 0 then begin
    { Lam = Lamp }
    ChangePhase := LamP < 0;
    if ChangePhase then
      Swap := OmP < Om
    else
      Swap := OmP > Om;
    if Om = Omp then
      Swap := Bra.S < Ket.S;
  end else if OmP > Om then begin { LamP = Lam = 0 }
    ChangePhase := OmP <= 0;
    Swap := ChangePhase;
  end else if OmP < Om then begin
    ChangePhase := Om <= 0;
    Swap := not ChangePhase;
  end else begin
    { LamP = Lam = 0 and SigmaP = Sigma }
    ChangePhase := OmP < 0;
    Swap := Bra.S < Ket.S;
  end;
  if ChangePhase then begin
    Result := Minus(Ket.S-Bra.S);
    if (Bra.Lambda < 0) xor (Ket.Lambda < 0) then
      Result := -Result;
  end else
    Result := 1;
end;

function TLinear.RotationalPerturbationMatrixElement(
   ABra, AKet: TBasisState; SignA, SignB: Integer; MNode: TMNode; F: TQuantumNumber): Complex;
var
  Phase, dOm, SignS: Integer;
  Bra, Ket: TLinearBasisState;
  JP, OmP, LamP, J, Om, Lam: TQuantumNumber;
  PE: TLinearPerturbation;
  ChangePhase, Swap: Boolean;
  Nuc: TLinearNucleus;

  MaxOmega, OmPP: TQuantumNumber;

  function BasePerturbation(OmegaP, Omega: TQuantumNumber): Double;
  { Perturbation without any centrifugal distortion }
  var
    SigmaP, Sigma: TQuantumNumber;
    function SJHelper(n:Integer):Double;
    var
      Sproj: TQuantumNumber;
    begin
      if PE.Srank < 0 then
        Result := rl(J,OmegaP,Omega) * rl(Ket.S,SigmaP,Sigma)* PE.Value
      else begin
        if n > 0 then
          SProj := PE.Scomp
        else if n < 0 then
          SProj := PE.Scomp
        else
          SProj := SigmaP - Sigma;
        if 2*Ket.S >= PE.Srank then begin
          Result := cg(PE.Srank, SProj, Ket.S, Sigma, Ket.S, Sigma+SProj)
                    * ReducedMatrixElement(PE.Srank,Ket.S)
                    / sqrt(Ket.S+One)
                    * phase * Minus(PE.Srank)
                    * rl(J,OmegaP,Omega) * rl(Ket.S,SigmaP,Sigma+Sproj)
                    * PE.Value
        end else
          Result := 0;
      end;
    end;
  begin
    SigmaP := OmegaP - LamP;
    Sigma := Omega - Lam;
    Result := 0;
    case PE.Op of
      LS: { Spin orbit coupling, with omega dependence from Wigner-Eckart theorem;
            See Lefebvre-Brion and Field, 2nd Edition Equation 3.4.36 }
        if (Omega = OmegaP) and (Abs(SigmaP-Sigma) <= 2) then begin
          if ChangePhase then
            SignS := -1
          else
            SignS := 1;
          if Swap then
            Result := phase * PE.Value * Minus(Ket.S-SignS*SigmaP)
                      * threej(Ket.S, -SignS*Sigma, 2, SignS*(Sigma-SigmaP), Bra.S, SignS*SigmaP)
          else
            Result := phase * PE.Value * Minus(Bra.S-SignS*Sigma)
                      * threej(Bra.S, -SignS*Sigmap, 2, SignS*(SigmaP-Sigma), Ket.S, SignS*Sigma);
        end;

      Luncouple: { J+^n L-^n + J-^n L+^n }
        if ( (LamP - Lam = 2*PE.n) and (SigmaP-Sigma = PE.Scomp)
             or (LamP - Lam = -2*PE.n) and (SigmaP-Sigma = -PE.Scomp) ) then
         { Test is done like this as effective operator is:
           J+^(-n-PE.Scomp/2) L+^n S+^PE.Scomp/2 + J-^(n+PE.Scomp/2) L-^-n S-^-PE.Scomp/2
           where negative powers are taken to indicate switching bewteen raising and lowering operators }
         Result := SJHelper(LamP-Lam);

      Suncouple:  { J+^n S-^n + J-^n S+^n }
        if (LamP = Lam) and (Abs(SigmaP-Sigma) = Abs(2*PE.n + PE.Scomp)) then
          Result := SJHelper((Sigmap-Sigma)*(2*PE.n + PE.Scomp));

      Homog: { Simple homogeneous Perturbation with delta omega = 0; deprectated }
        if (SignA = SignB) and (OmegaP = Omega) then begin
          Result := Phase*PE.Value;
          if PE.n > 1 then
            Result := Result * IntPower(J*(J+2)/4,PE.n div 2);
        end;

      NS,RS:
        if LamP = Lam then begin
          if Sigma = SigmaP then begin
            Result := Omega*Sigma;
            if PE.Op = RS then
              Result := Result - Lam*Sigma;
            Result := phase * PE.Value * (Result - Ket.S*(Ket.S+2))/4
          end else if Abs(SigmaP-Sigma)=2 then
            Result := phase * PE.Value * rl(J,OmegaP,Omega) * rl(Ket.S,SigmaP,Sigma) / 2;
        end;

      LNuncouple: { N-+^(n+m) L+-^n S+-^m = (J-+ - S-+)^(n+m) L+-^n S+-^m where m = PE.Scomp/2 }
        if ( (LamP - Lam = 2*PE.n) and (SigmaP-Sigma >= -2*PE.n) and (SigmaP-Sigma <= PE.Scomp)
             or (LamP - Lam = -2*PE.n) and (SigmaP-Sigma <= 2*PE.n) and (SigmaP-Sigma >= -PE.Scomp) ) then
          Result := SJHelper(LamP-Lam)
                    * Minus(2*PE.n+PE.Scomp-OmegaP+Omega)
                    * BinomialCoefficient(Abs(PE.n+PE.Scomp div 2), Abs(OmegaP-Omega) div 2);

      aL: if (SigmaP = Sigma) and (Abs(LamP-Lam) = 2) then
            Result := Phase * Minus(JP-OmegaP) * threej(JP, -OmegaP, 2, OmegaP-Omega, J, Omega)
                     * (Omega-OmegaP) * PE.Value
                     * Sqrt(Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1)*(JP+One)/32);

      e:  if Abs(LamP-Lam) = 2 then begin
            dOm := OmegaP - Omega;
            Result := -Phase * PE.Value * Minus(JP - OmegaP + S - SigmaP + dOm)
                      * threej(JP, -OmegaP, 2, -dOm, J, Omega)
                      * threej(S, -SigmaP, 2, SigmaP-Sigma, S, Sigma)
                      * threej(2, SigmaP-Sigma, 4, LamP-Lam, 2, -dOm)
                      * Sqrt(20*S*(S+2)*(S+One)/4
                             * Nuc.Spin*(Nuc.Spin+2)*(Nuc.Spin+1)*(J+1)*(JP+One)/4)
          end;

      eQq1: if (SigmaP = Sigma) and (Abs(LamP-Lam) = 2) then begin
              Result := -Phase * PE.Value
                        * Minus(J + OmegaP) * Sqrt((J + 1) * (JP + One) / 6)
                        * threej(4, Omega-OmegaP, J, -Omega, JP, OmegaP)
                        /( 2 * threej(Nuc.Spin, -Nuc.Spin, 4, 0, Nuc.Spin, Nuc.Spin));
              if LamP > Lam then begin
                if Lam < 0 then
                  Result := - Result;
              end else if Lam > 0 then
                Result := - Result;
            end;
    else
      raise NotYet;
    end;
  end;

begin
  Result := z0;
  PE := MNode as TLinearPerturbation;
  Bra := ABra as TLinearBasisState;
  Ket := AKet as TLinearBasisState;
  { Omega selection }
  if PE.OmegaSelect > -9999 then begin
    { For the moment, require omega selection to be diagonal in omega;
      Use Omega from basis to avoid confusion with signs }
    if (Bra.Omega <> PE.OmegaSelect) or (Ket.Omega <> PE.OmegaSelect) then
      Exit;
  end;
  JP := Bra.J;
  J := Ket.J;
  if (J <> JP) and (PE.Op < aL) then
    Exit; { Only hyperfine perturbations can be off diagonal in J }

  if PE.Nucleus > 0 then
    Nuc := Nucleus(PE.Nucleus-1) as TLinearNucleus
  else
    Nuc := nil;
  OmP := SignA*Bra.Omega;
  LamP := SignA*Max(Bra.St.Lambda,0);
  Om := SignB*Ket.Omega;
  Lam := SignB*Max(Ket.St.Lambda,0);
  Phase := DeterminePhase(ChangePhase, Swap, Omp, LamP, Om, Lam, Bra.St, Ket.St);

  if PE.Npower <= 0 then begin
    Result.re := BasePerturbation(OmP, Om);
    if PE.Npower < 0 then
      Result.re := Result.re*IntPower(J*(J+2)/4,-PE.NPower div 2);
  end else begin
    OmPP := Max(Om-PE.Npower, Lam-Ket.S);
    MaxOmega := Min(Om+PE.Npower, Lam+Ket.S);
    while OmPP <= MaxOmega do begin
      if Abs(OmPP) <= J then
        Result.re := Result.re + BasePerturbation(OmP,OmPP)
                                 * NRn(PE.Npower, J, Lam, Ket.S, OmPP, Om);
      Inc(Ompp,2);
    end;
    OmPP := Max(Omp-PE.Npower, LamP-Bra.S);
    MaxOmega := Min(Omp+PE.Npower, LamP+Bra.S);
    while OmPP <= MaxOmega do begin
      if Abs(OmPP) <= J then
        Result.re := Result.re + NRn(PE.Npower, JP, LamP, Bra.S, OmP, OmPP)
                                 * BasePerturbation(OmPP,Om);
      Inc(Ompp,2);
    end;
    Result.re := Result.re / 2;
  end;
end;

const
  LinearSettingNames: array[LSI..LSI+2] of string = (
'Electronic angular momentum - use Sigma-, Sigma+, Pi, Delta, Phi, Gamma, H, I or a numerical value',
'true for gerade states; ignored if molecule is not symmetric.',
'Omega Selection in basis.'
);

function TLinear.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearSettingNames))
     and (AIndex <= High(LinearSettingNames)) then
    Result := LinearSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

const
 LinearParameterNames: array [LSP..LSP+27] of string = (
'Rotational constant.',
'Spin-orbit coupling constant.',
'Spin - spin coupling constant (=1.5*epsilon).',
'Spin-spin coupling constant giving lambda doubling for Omega = 0 states.',
'Higher Order Spin-Spin interaction (S>=2 only).',
'Spin rotation coupling constant.',
'Lambda doubling constant.',
'Lambda doubling constant.',
'Quartic centrifugal distortion.',
'Sextic centrifugal distortion.',
'J^8  centrifugal distortion.',
'J^10  centrifugal distortion.',
'J^12 centrifugal distortion. (Not P to avoid confusion with p.)',
'Centrifugal distortion of A.',
'Centrifugal Distortion of Lambda.',
'J^4 Centrifugal Distortion of Lambda.',
'Centrifugal distortion of o.',
'J^4 Centrifugal distortion of o.',
'J^6 Centrifugal distortion of o.',
'Centrifugal distortion of p.',
'J^6 Centrifugal distortion of p.',
'J^8 Centrifugal distortion of p.',
'Centrifugal distortion of q.',
'J^6 Centrifugal distortion of q.',
'J^8 Centrifugal distortion of q.',
'Centrifugal distortion of gamma.',
'J^6 Centrifugal distortion of gamma.',
'J^8 Centrifugal distortion of gamma.'
);

function TLinear.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearParameterNames))
     and (AIndex <= High(LinearParameterNames)) then
    Result := LinearParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

type
  TPertEquiv = record
    Name:  string;
    Op:    TLinearOperator;
    n:     Integer;
    NPower:Integer;
    Srank: TQuantumNumber;
    Scomp: TQuantumNumber;
    Num, Den:    Integer; { Scale factor = Sqrt(Num/Den)*Sign(Num) }
    SymSelect:   Integer;
    OmegaSelect: TQuantumNumber;
  end;

const
  PertEquivs: array [0..28] of TPertEquiv = (
   (Name:'Origin'; Op:Luncouple; n:0; NPower: 0; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),

   (Name:'B';  Op:Luncouple; n:0; NPower: 2; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'D';  Op:Luncouple; n:0; NPower: 4; Srank: 0; Scomp: 0; Num:-1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'H';  Op:Luncouple; n:0; NPower: 6; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'L';  Op:Luncouple; n:0; NPower: 8; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'M';  Op:Luncouple; n:0; NPower:10; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'PP'; Op:Luncouple; n:0; NPower:12; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),

   { Next two need special scaling factor }
   (Name:'A';  Op:LS;        n:1; NPower:0;  Srank: 0; Scomp: 0; Num:0; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'AD'; Op:LS;        n:1; NPower:2;  Srank: 0; Scomp: 0; Num:0; Den:1; SymSelect:-1; OmegaSelect:-9999),

   (Name:'o';  Op:LNuncouple; n:2; NPower: 0; Srank: -1; Scomp: -4; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'oD'; Op:LNuncouple; n:2; NPower: 2; Srank: -1; Scomp: -4; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'oH'; Op:LNuncouple; n:2; NPower: 4; Srank: -1; Scomp: -4; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'oL'; Op:LNuncouple; n:2; NPower: 6; Srank: -1; Scomp: -4; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),

   (Name:'p';  Op:LNuncouple; n:2; NPower: 0; Srank: -1; Scomp: -2; Num:-1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'pD'; Op:LNuncouple; n:2; NPower: 2; Srank: -1; Scomp: -2; Num:-1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'pH'; Op:LNuncouple; n:2; NPower: 4; Srank: -1; Scomp: -2; Num:-1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'pL'; Op:LNuncouple; n:2; NPower: 6; Srank: -1; Scomp: -2; Num:-1; Den:4; SymSelect:-1; OmegaSelect:-9999),

   (Name:'q';  Op:LNuncouple; n:2; NPower: 0; Srank: 0; Scomp: 0; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'qD'; Op:LNuncouple; n:2; NPower: 2; Srank: 0; Scomp: 0; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'qH'; Op:LNuncouple; n:2; NPower: 4; Srank: 0; Scomp: 0; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),
   (Name:'qL'; Op:LNuncouple; n:2; NPower: 6; Srank: 0; Scomp: 0; Num:1; Den:4; SymSelect:-1; OmegaSelect:-9999),

   (Name:'gamma';  Op:NS; n:1; NPower: 0; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'gammaD'; Op:NS; n:1; NPower: 2; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'gammaH'; Op:NS; n:1; NPower: 4; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),
   (Name:'gammaL'; Op:NS; n:1; NPower: 6; Srank: 0; Scomp: 0; Num:1; Den:1; SymSelect:-1; OmegaSelect:-9999),

   (Name:'LambdaSS'; Op:Luncouple; n:0; NPower: 0; Srank: 4; Scomp: 0; Num:8; Den:3; SymSelect:-1; OmegaSelect:-9999),
   (Name:'LambdaD';  Op:Luncouple; n:0; NPower: 2; Srank: 4; Scomp: 0; Num:8; Den:3; SymSelect:-1; OmegaSelect:-9999),
   (Name:'LambdaH';  Op:Luncouple; n:0; NPower: 4; Srank: 4; Scomp: 0; Num:8; Den:3; SymSelect:-1; OmegaSelect:-9999),

   { Brown and Carrington 10.194 }
   (Name:'theta';    Op:Luncouple; n:0; NPower: 0; Srank: 8; Scomp: 0; Num:70; Den:36; SymSelect:-1; OmegaSelect:-9999)
);

function FindPertEquiv(const AName: string):TPertEquiv;
  function FindParameter(const Equivs: array of TPertEquiv):TPertEquiv;
  var
    i: Integer;
  begin
    Result.Name := '';
    for i := Low(Equivs) to High(Equivs) do
      if CompareText(Equivs[i].Name,AName) = 0 then begin
        Result := Equivs[i];
        Exit;
      end;
  end;
begin
  Result := FindParameter(PertEquivs)
end;

function PEScaleFactor(PE:TPertEquiv; Ket:TLinear):TExprNode;
begin
 if PE.Num = 0 then
   { Special case for spin orbit coupling }
   if (Ket.Lambda > 0) and (Ket.S > 0) then
     Result := EN([Ket.Lambda div 2, '*', Halve(Ket.S), '/', EN(['threej',Halve(Ket.S),Halve(-Ket.S),1,0,Halve(Ket.S),Halve(Ket.S)])])
   else
     Result := EN([0])
 else
   Result := EN([Abs(PE.Num), '/', PE.Den, 'sqrt', '*', Sign(PE.Num)]);
end;

type
  TSymbolicNrnEvaluator = class
    MatrixElements: TMatrixElements;
    destructor Destroy; override;
    function NRn(n: Integer; Lam, Spin, Omp, Om: TQuantumNumber; RSquaredH:Boolean): TExprNode;
  end;
var
  SymbolicNrnEvaluator: TSymbolicNrnEvaluator = nil;

{ TSymbolicNrnEvaluator }

destructor TSymbolicNrnEvaluator.Destroy;
begin
  DiscardMatrixElements(MatrixElements);
  inherited;
end;

function TSymbolicNrnEvaluator.NRn(n: Integer; Lam, Spin, Omp,
  Om: TQuantumNumber; RSquaredH: Boolean): TExprNode;
{ Matrix Elements of N^n or R^n, depending on state of RSquaredH }
var
  i,j:Integer;
  Sig: array [0..5] of Integer;
  Match: Boolean;
  Ompp, OmMax: TQuantumNumber;
  eSpin: TExprNode;
  iv: Integer;
begin
  { Cache results, as otherwise recusive evaluation is time consuming }
  Sig[0] := n;
  Sig[1] := Lam;
  Sig[2] := Spin;
  Sig[3] := Omp;
  Sig[4] := Om;
  Sig[5] := Ord(RSquaredH);
  for i := 0 to High(MatrixElements) do begin
    Match := True;
    with MatrixElements[i] do
      for j := 0 to 5 do
        if Signature[j] <> Sig[j] then begin
          Match := False;
          Break;
        end;
    if Match then begin
      Result := MatrixElements[i].Expr.Duplicate;
      Exit;
    end;
  end;
  if Odd(n) then
    raise EReportException.Create('Nrn:NPower must be even:'+IntToStr(n));
  eSpin := Halve(Spin);
  if Abs(Om-Omp) = n then begin
    Result := EN([Minus(n), '*', rlE(eJ, Omp, Om), '*', rlE(eSpin, Lam-Omp, Lam-Om)])
  end else if (Om = Omp) and (n = 2) then begin
    Result := EB([eJ, '+', 1, '*', eJ]);
    iv := Spin*(Spin+2) - Om*Om - Sqr(Lam-Om);
    if not RSquaredH then
      iv := iv  + Lam*Lam;
    Result := EN([iv, '/', 4, '+', Result]);
  end else if Abs(Omp-Om) < n then begin
    Result := TIntegerNode.Create(0);
    Ompp := Max(Omp - 2, Lam-Spin);
    OmMax := Min(Omp + 2, Lam+Spin);
    while Ompp <= OmMax do begin
      Result := EN([Result, '+', EN([NRn(2,Lam,Spin,Omp,Ompp,RSquaredH),
                                    '*', NRn(n-2,Lam,Spin,Ompp,Om,RSquaredH)])]);
      Inc(Ompp,2);
    end;
  end else
    Result := TIntegerNode.Create(0);
  eSpin.Free;
  { Cache result }
  SetLength(MatrixElements, Length(MatrixElements)+1);
  with MatrixElements[High(MatrixElements)] do begin
    SetLength(Signature, 6);
    for j := 0 to 5 do
      Signature[j] := Sig[j];
    Result := Result.Simplify([siExpand]);
    Result := Result.Simplify;
    Expr := Result.Duplicate;
  end;
end;

function SymbolicRotationalMatrixElement(PE: TPertEquiv; Molecule: TLinearMolecule; Bra, Ket: TLinear): TMatrixElements;
var
  LamP, Lam: TQuantumNumber;
  SigP, Sig: TQuantumNumber;
  OmegaPP, MaxOmega, OmegaP, Omega: TQuantumNumber;
  Node: TExprNode;
  nMatrixElements, Phase: Integer;
  ChangePhase, Swap: Boolean;
  
  { The following pair of routines cycle through all possible (signed) values
    of Lambda and Sigma for a given |Lambda| and S }
  procedure Start(State: TLinear; out Omega, Lambda, Sigma: TQuantumNumber);
  begin
    Lambda := Max(State.Lambda,0);
    if Lambda > 0 then
      Lambda := -Lambda;
    Sigma := - State.S;
    Omega := Lambda+Sigma;
  end;
  function Cycle(State: TLinear; var Omega, Lambda, Sigma: TQuantumNumber):Boolean;
  begin
    Result := False;
    Inc(Sigma,2);
    if Sigma > State.S then begin
      if Lambda < 0 then begin
        Lambda := -Lambda;
        Sigma := -State.S;
      end else
        Result := True; { Force stop }
    end;
    Omega := Lambda+Sigma;
  end;
  function CheckOmega(Lambda, Sigma: TQuantumNumber):Boolean;
  { Honor Omega selection for matrix element.
    For the moment, require omega selection to be diagonal in omega.
    To be compatible with numerical version the omega compared must be
    the value used to specify the basis state. This is normally the positive value,
    but formally has Lambda > 0 or failing that Sigma > 0 }
  begin
    if (Lambda > 0) or ((Lambda =0) and (Sigma >=0)) then
      Result := PE.OmegaSelect = (Lambda+Sigma)
    else
      Result := PE.OmegaSelect = -(Lambda+Sigma)
  end;

  function BasePerturbation(OmegaP, Omega: TQuantumNumber; Expand: Boolean = False):TExprNode;
  var
    SigmaP, Sigma: TQuantumNumber;
    SignS, EPhase: Integer;
    function SJHelper(n:Integer):TExprNode;
    var
      Sproj: TQuantumNumber;
      eSpin: TExprNode;
    begin
      if PE.Srank < 0 then begin
        eSpin := Halve(Ket.S);
        Result :=EN([rlE(eJ,OmegaP,Omega), '*', rlE(eSpin,SigmaP,Sigma)]);
        eSpin.Free;
      end else begin
        if n > 0 then
          SProj := PE.Scomp
        else if n < 0 then
          SProj := PE.Scomp
        else
          SProj := SigmaP - Sigma;
        if 2*Ket.S >= PE.Srank then begin
          eSpin := Halve(Ket.S);
          Result := EN(['CG', PE.Srank div 2, SProj div 2, Halve(Ket.S), Halve(Sigma), Halve(Ket.S), Halve(Sigma+SProj)]);
          Result := EN([Result, '/', EN(['sqrt', EB([2, '*', eSpin, '+', 1])]), '*', ReducedMatrixElementE(PE.Srank,eSpin)]);
          if Odd(PE.Srank div 2) then
            Result := EN([-1, '*', Result]);
          Result :=EN([rlE(eJ,OmegaP,Omega), '*', rlE(eSpin,SigmaP,Sigma+SProj), '*', Result]);
          eSpin.Free;
          if phase < 0 then
            Result := EN([-1, '*', Result]);
        end else
          Result := nil;
      end;
    end;
    function SignBK(Lambda, Sigma:TQuantumNumber):Integer;
    begin
      { Helper for Homog }
      if Lambda <> 0 then
        Result := Sign(Lambda)
      else if Sigma <> 0 then
        Result := Sign(Sigma)
      else
        Result := 1;
    end;
  var
    eSpin: TExprNode;
  begin
    SigmaP := OmegaP - LamP;
    Sigma := Omega - Lam;
    Result := nil;
    case PE.Op of
      LS: { Spin orbit coupling, with omega dependence from Wigner-Eckart theorem;
            See Lefebvre-Brion and Field, 2nd Edition Equation 3.4.36 }
        if (Omega = OmegaP) and (Abs(SigmaP-Sigma) <= 2) then begin
          if (PE.Srank <> 0) or (PE.Scomp <> 0) then
            raise NotYet;
          if ChangePhase then
            SignS := -1
          else
            SignS := 1;
          if Swap then begin
            EPhase := Phase * Minus(Ket.S-SignS*SigmaP);
            Result := EN(['threej', Halve(Ket.S), Halve(-SignS*Sigma), 1, Halve(SignS*(Sigma-SigmaP)), Halve(Bra.S), Halve(SignS*SigmaP)]);
          end else begin
            EPhase := Phase * Minus(Bra.S-SignS*Sigma);
            Result := EN(['threej', Halve(Bra.S), Halve(-SignS*Sigmap), 1, Halve(SignS*(SigmaP-Sigma)), Halve(Ket.S), Halve(SignS*Sigma)]);
          end;
          if EPhase < 0 then
            Result := EN([-1, '*', Result]);
        end;

      Luncouple:
        if ( (LamP - Lam = 2*PE.n) and (SigmaP-Sigma = PE.Scomp)
             or (LamP - Lam = -2*PE.n) and (SigmaP-Sigma = -PE.Scomp) ) then
         { Test is done like this as effective operator is:
           J+^(-n-PE.Scomp/2) L+^n S+^PE.Scomp/2 + J-^(n+PE.Scomp/2) L-^-n S-^-PE.Scomp/2
           where negative powers are taken to indicate switching bewteen raising and lowering operators }
         Result := SJHelper(LamP-Lam);

      Suncouple:
        if (LamP = Lam) and (Abs(SigmaP-Sigma) = Abs(2*PE.n + PE.Scomp)) then
          Result := SJHelper((Sigmap-Sigma)*(2*PE.n + PE.Scomp));

      Homog:
        if (OmegaP = Omega) and (SignBK(LamP,SigmaP) = SignBK(Lam,Sigma)) then begin
          if (PE.Srank <> 0) or (PE.Scomp <> 0) then
            raise NotYet;
          Result := EN([Phase]);
          if PE.n > 1 then
            Result := EN([EB([eJ, '+', 1, '*', eJ]), '^', PE.n div 2, '*', Result]);
        end;

      NS, RS:
        if LamP = Lam then begin
          if (PE.Srank <> 0) or (PE.Scomp <> 0) then
            raise NotYet;
          if Sigma = SigmaP then
            if PE.Op = NS then
              Result := EN([(Omega*Sigma - Ket.S*(Ket.S+2)), '/', 4])
            else
              Result := EN([((Omega-Lam)*Sigma - Ket.S*(Ket.S+2)), '/', 4])
          else if Abs(SigmaP-Sigma)=2 then begin
            eSpin := Halve(Ket.S);
            Result := EN([rlE(eJ,OmegaP,Omega), '*', rlE(eSpin,SigmaP,Sigma), '/', 2]);
            eSpin.Free;
          end;
          if Assigned(Result) and (phase < 0) then
            Result := MonadicMinus.CreateExprNode(Result).Simplify;
        end;

      LNuncouple: { N-+^(n+m) L+-^n S+-^m = (J-+ - S-+)^(n+m) L+-^n S+-^m where m = PE.Scomp/2 }
        if ( (LamP - Lam = 2*PE.n) and (SigmaP-Sigma >= -2*PE.n) and (SigmaP-Sigma <= PE.Scomp)
             or (LamP - Lam = -2*PE.n) and (SigmaP-Sigma <= 2*PE.n) and (SigmaP-Sigma >= -PE.Scomp) ) then
          Result := EN([SJHelper(LamP-Lam),
                       '*', Minus(2*PE.n+PE.Scomp-OmegaP+Omega)*BinomialCoefficient(Abs(PE.n+PE.Scomp div 2), Abs(OmegaP-Omega) div 2)]);
    else
      raise NotYet; // aL: ; e: ; eQq1: ;
    end;
    if Assigned(Result) and Expand then
      Result := Result.Simplify([siExpand]);
  end;
var
  NodeA: TExprNode;
begin
  SetLength(Result,0);
  nMatrixElements := 0;
  Start(Bra, OmegaP, LamP, SigP);
  repeat
    Start(Ket, Omega, Lam, Sig);
    repeat
      { Omega selection }
      if PE.OmegaSelect > -9999 then
        if not CheckOmega(LamP, SigP) or not CheckOmega(Lam, Sig) then
          Continue;
      Phase := DeterminePhase(ChangePhase, Swap, OmegaP, LamP, Omega, Lam, Bra, Ket);
      if PE.Npower <= 0 then begin
        Node := BasePerturbation(OmegaP, Omega);
        if PE.NPower < 0 then
          Node := EN([Node, '*', EB([eJ, '+', 1, '*', eJ, '^', -PE.NPower div 2])]);
      end else begin
        OmegaPP := Max(Omega-PE.Npower, Lam-Ket.S);
        MaxOmega := Min(Omega+PE.Npower, Lam+Ket.S);
        Node := nil;
        while OmegaPP <= MaxOmega do begin
          NodeA := BasePerturbation(OmegaP,OmegaPP,True);
          if NodeA.EqualTo(0) then
            NodeA.Free
          else
            Node := EN([ NodeA,
                       '*', SymbolicNrnEvaluator.NRn(PE.Npower, Lam, Ket.S, OmegaPP, Omega, Molecule.RSquaredH),
                       '+', Node]);
          Inc(OmegaPP,2);
        end;
        OmegaPP := Max(OmegaP-PE.Npower, LamP-Bra.S);
        MaxOmega := Min(OmegaP+PE.Npower, LamP+Bra.S);
        while OmegaPP <= MaxOmega do begin
          NodeA := BasePerturbation(OmegaPP,Omega,True);
          { TODO: This should use JP }
          if NodeA.EqualTo(0) then
            NodeA.Free
          else
            Node := EN([SymbolicNrnEvaluator.NRn(PE.Npower, LamP, Bra.S, OmegaP, OmegaPP,Molecule.RSquaredH),
                        '*', NodeA,
                        '+', Node]);
          Inc(OmegaPP,2);
        end;
        Node := EN([Node, '/', 2]);
      end;
      if Node.EqualTo(0) then
        Node.Free
      else begin
        Node := EN([Node, '*', PEScaleFactor(PE,Ket)]);
        if nMatrixElements >= Length(Result) then
          SetLength(Result, Length(Result)+10);
        Result[nMatrixElements].Expr := Node;
        Result[nMatrixElements].Evaluator := nil;
        SetLength(Result[nMatrixElements].Signature,4);
        Result[nMatrixElements].Signature[0] := LamP;
        Result[nMatrixElements].Signature[1] := OmegaP;
        Result[nMatrixElements].Signature[2] := Lam;
        Result[nMatrixElements].Signature[3] := Omega;
        Inc(nMatrixElements);
      end;
    until Cycle(Ket, Omega, Lam, Sig);
  until Cycle(Bra, OmegaP, LamP, SigP);
  SetLength(Result, nMatrixElements);
end;

function TLinear.SymbolicRotationalMatrixElement(AIndex: Integer): TMatrixElements;
var
  PE: TPertEquiv;
begin
  PE := FindPertEquiv(ParameterNames[AIndex]);
  if PE.Name = '' then
    Result := inherited SymbolicRotationalMatrixElement(AIndex)
  else
    Result := HundA.SymbolicRotationalMatrixElement(PE, Molecule, Self, Self);
end;

function TLinear.MatrixElementType(const Changes: array of Integer; const H: string): string;
begin
  Result := '<J,Lambda=' + Manifold.QuantumNumberToStr(Changes[0]) + ',Omega=' + Manifold.QuantumNumberToStr(Changes[1]) + '|' + H
             + '|J,Lambda=' + Manifold.QuantumNumberToStr(Changes[2]) + ',Omega=' + Manifold.QuantumNumberToStr(Changes[3]) + '>';
end;

procedure TLinear.AdjustDefaults;
var
  i: Integer;
  MySelf: TLinear;
begin
  inherited;
  MySelf := Self;
  { States within a single manifold must be all u or all g }
  with Owner as TManifold do
    if nStates > 1 then
      for i := 0 to nStates - 1 do
        if State[i] <> MySelf then begin
          MySelf.gerade := (State[i] as TLinear).gerade;
          Exit;
        end;
end;

function TLinear.AllowedSphericalComponent(
  TransitionMoment: TSphericalTransitionMoment; State: TState;
  out Component: TQuantumNumber): Exception;
var
  Ket: TLinear;
  OtherComponent: TQuantumNumber;
begin
  Ket := State as TLinear;
  if S <> Ket.S then
    Result := EReportException.Create(TransitionMoment.UniqueName+': Spin forbidden transition - component must be specified')
  else if TransitionMoment is TMagneticTransitionMoment then
    Result := EReportException.Create(TransitionMoment.UniqueName+': auto not implemented for magnetic dipole transitions')
  else if (TransitionMoment.Component <= 2) and (Min(Lambda,Ket.Lambda) < 0) and (Max(Lambda,Ket.Lambda) = 0) then
    Result := EReportException.Create(TransitionMoment.UniqueName+': Sigma+ to Sigma- transitions are forbidden')
  else begin
    Component := Abs(Max(Lambda,0)-Max(Ket.Lambda,0));
    if Component > TransitionMoment.Rank then
      Result := EReportException.Create(TransitionMoment.UniqueName+': Transition is forbidden - too large a change in lambda')
    else begin
      OtherComponent := Abs(Max(Lambda,0))+Abs(Max(Ket.Lambda,0));
      if (OtherComponent <> Component) and (OtherComponent <= TransitionMoment.Rank) then
        Result := EReportException.Create(TransitionMoment.UniqueName+': More than one possible component for an allowed transition - please set component')
      else
        Result := nil;
    end;
  end;
end;

function TLinear.ConvertParameter(AIndex: Integer; Zero: Boolean): TPerturbation;
var
  Pert: TLinearPerturbation;
  PE: TPertEquiv;
  e: TExprNode;
begin
  Result := nil;
  PE := FindPertEquiv(ParameterNames[AIndex]);
  if PE.Name <> '' then begin
    Pert := TLinearPerturbation.Create(Owner);
    with PE do begin
      e := PEScaleFactor(PE,Self);
      if e.EqualTo(-1) then
        Pert.Comment := '-'+Name
      else if not e.EqualTo(1) then
        Pert.Comment := e.Text(True)+'*'+Name
      else
        Pert.Comment := Name;
      Pert.Value := e.RealValue*Parameters[AIndex];
      e.Free;
      Pert.Op := Op;
      Pert.n := n;
      Pert.Npower := NPower;
      Pert.Srank := Srank;
      Pert.Scomp := Scomp;
    end;
    Pert.SetBraKetName(Name,Name);
    Result := Pert;
    if Zero then
      Parameters[AIndex] := 0;
  end;
end;

function TLinear.FixupProperties: Exception;
begin
  if not Assigned(SymbolicNrnEvaluator) then
    SymbolicNrnEvaluator := TSymbolicNrnEvaluator.Create;
  Result := inherited FixupProperties;
  if not Assigned(Result) then
    { TODO: Roll into lowest state origin logic }
    if (OmegaSelect > -9999) and Odd(OmegaSelect+S) then
       Result := EReportException.Create('OmegaSelect invalid for ' + UniqueName)
    else
      try
        if S = 0 then
          FInverted := False
        else if Lambda > 0 then
          { TODO: Call DiagonalMatrixlemnts somehow, so that everything as perturbations is still correct }
          FInverted := HRot(Lambda+S, Lambda, S, Lambda+S, Lambda, S, Lambda+S)
                       < HRot(Lambda+S, Lambda, -S, Lambda-S, Lambda, -S, Lambda-S)
        else
          FInverted := HRot(S, 0, 0, 0, 0, 0, 0) > HRot(S, 0, S, S, 0, S, S)
      except
        on E: Exception do
          { Copy Exception }
          Result := Exception.Create(E.Message);
      end;
end;

function TLinear.SettingToPrettyStr(AIndex: Integer): string;
begin
  Result := '';
  case AIndex of
    LSI:   { Lambda } begin Result := SettingToStr(AIndex);  Exit; end;
    LSI+1: if not Molecule.Symmetric then Exit; { g/u }
    LSI+2: if OmegaSelect <= -9999 then Exit;
  end;
  Result := inherited SettingToPrettyStr(AIndex);
end;

function TLinear.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = LSI+2) and (OmegaSelect <= -9999) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

class function TLinear.SpinSpinType: TSpinSpinClass;
begin
  Result := TLinearSpinSpin;
end;

procedure TLinear.SetSettingFromStr(AIndex: Integer; const Str: string);
begin
  if (AIndex = LSI+2) and (Str = 'all') then
    OmegaSelect := -9999
  else
    inherited;
end;

{ TLinearManifold }

function TLinearManifold.FixupProperties: Exception;
begin
  Result:=inherited FixupProperties;
  { TODO: Consider this test (and possibly others) as a check for possible problem state? (Do not enforce rigdly, as confusing to users }
//  if not Assigned(Result) and not (Molecule as TLinearMolecule).Symmetric then
//    for i := 0 to nStates-1 do
//      if not (State[i] as TLinear).gerade then begin
//        Result := EReportException.Create('ungerade state (' + State[i].Name + ') in unsymmetric molecule (' + Molecule.Name + ')');
//        Exit;
//      end;
end;

function TLinearManifold.nSym: Integer;
begin
  Result := 2;
end;

{ TLinearMolecule }

constructor TLinearMolecule.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',ShowJ=True,ShowN=True,ShowFNumber=True,Showef=True,SymWt=1,AsymWt=1';
  inherited Create(AOwner);
end;

function TLinearMolecule.FixupProperties: Exception;
begin
  if Symmetric then
    FnSpinSpecies := 2
  else
    FnSpinSpecies := 1;
  Result := inherited FixupProperties;
end;

function TLinearMolecule.Allowed(TransitionMoment: TTransitionMoment; Sym1, Sym2: Integer): Boolean;
begin
  if TransitionMoment.Magnetic then
    Result := Sym1 = Sym2
  else
    Result := ((Sym1 xor Sym2) xor (TransitionMoment.Rank div 2) and 1) = 0; { Parity selection rule }
end;

function TLinearMolecule.JAdjustedSym(J: TQuantumNumber; Sym: Integer): Integer;
begin
  if (Sym >= 0) and JAdjustSym and ((J and 2) <> 0) then
    Result := 1 - Sym
  else
    Result := Sym;
end;

function TLinearMolecule.LastSym: Integer;
begin
  Result := 1;
end;

function TLinearMolecule.SymName(Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else if Sym = 0 then
    Result :='+'
  else
    Result := '-';
end;

function TLinearMolecule.JAdjustedSymNameBase(J: TQuantumNumber; Sym: Integer): string;
begin
  if Sym < 0 then
    Result := '-'
  else if ((J and 2) <> 0) xor (Sym = 1) then
    Result := 'f'
  else
    Result := 'e'
end;

const
  LinearMoleculeSettingNames: array[LMI..LMI+8] of string = (
'True for D infinity h, False for C infinity v.',
'Statistical weight of symmetric rotational levels. Ignored if no centre of symmetry.',
'Statistical weight of asymmetric rotational levels. Ignored if no centre of symmetry.',
'Set to use the R^2 Hamiltonian rather than the default N^2 form.',
'Show J in state label.',
'Show Omega in state label.',
'Show N in state label.',
'Show F Number in state label.',
'Show e/f symmetry in state label.'
);

function TLinearMolecule.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearMoleculeSettingNames))
     and (AIndex <= High(LinearMoleculeSettingNames)) then
    Result := LinearMoleculeSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TLinearMolecule.IsShowingN: Boolean;
begin
  Result := ShowN and (not AllSinglet or not ShowN);
end;

function TLinearMolecule.IsValidChildClass(MNodeClass: TMNodeClass): Boolean;
begin
  Result := inherited IsValidChildClass(MNodeClass);
  if Result and MNodeClass.InheritsFrom(TTransitionMoment) then
    Result := (MNodeClass = TSpinTransitionMoment)
              or (MNodeClass = TOrbitalTransitionMoment)
              or (MNodeClass = TRotationalTransitionMoment)
              or (MNodeClass = TSphericalTransitionMoment)
              or (MNodeClass = TNuclearMagneticMoment);
end;

procedure TLinearMolecule.OmegaRange(out Smallest, Largest: TQuantumNumber);
var
  I, J: Integer;
  Lam: TQuantumNumber;
begin
  Smallest := High(Integer);
  Largest := Low(Integer);
  for I := 0 to nManifolds - 1 do
    with Manifold[I] do
      for J := 0 to nStates - 1 do
        with State[J] as TLinear do begin
          Lam := Max(Lambda, 0);
          Smallest := Min(Lam-S,Smallest);
          Largest := Max(Lam+S,Largest);
        end;
end;

function TLinearMolecule.QnoInfo(QnoType: TQnoType): TQnoInfo;
var
  Smallest, Largest: TQuantumNumber;
begin
  Result := inherited QnoInfo(QnoType);
  with Result do
    case QnoType of
      qtA: begin OmegaRange(Smallest,Largest); Name := 'Omega'; MinValue := Smallest; MaxValue := Largest; end;
      qtB: begin OmegaRange(Smallest,Largest); Name := 'Fn'; MinValue := 1; MaxValue := (Largest-Smallest) div 2 + 1; StepSize := 1; end;
    end;
end;

{ TLinearPerturbation }

function TLinearPerturbation.FixupProperties: Exception;
begin
  Result := inherited FixupProperties;
  if not Assigned(Result)
     and ( ((Srank <> 0) or (Scomp <> 0))
           and not (Op in [Luncouple, Suncouple, LNuncouple])) then
    Result := Exception.Create('Srank/Scomp can''t be used with '+ SettingToStr(LPI));
end;

function TLinearPerturbation.GetOperator(const Index: Integer): TLinearOperator;
begin
  Result := TLinearOperator(GetSetting(Index));
end;

procedure TLinearPerturbation.SetOperator(const Index: Integer; const AValue: TLinearOperator);
begin
  SetSetting(Index, Ord(AValue));
  case AValue of
    aL, e: NuclearRank := 2;
    eQq1:  NuclearRank := 4;
  else
    NuclearRank := 0;
  end;
end;

function TLinearPerturbation.PairName: string;
  function Show(Op: string; Power: Integer): string;
  begin
    case Power of
      0: Result := '';
      1: Result := Op;
    else
      Result := Op + IntToStr(Power)
    end;
  end;
begin
  case Op of
    LS: Result := 'LS';
    Luncouple: Result := Show('J+',n) + Show('L-',n);
    Suncouple: Result := Show('J+',n) + Show('S-',n);
    NS: Result := 'N.S';
    RS: Result := 'R.S';
    LNuncouple: Result := Show('N+',n) + Show('L-',n);
    Homog: Result := Show('J(J+1)',n div 2);
    aL: Result := 'I-L+';
    e:  Result := 'e';
    eQq1: Result := 'eQq1';
  end;
  if Srank < 0 then
    Result := Result + Show('S+',Scomp div 2)
  else if Srank > 0 then
    Result := Result + Format('T^%s_%s(S)', [(RootNode as TMixture).QuantumNumberToStr(Srank), (RootNode as TMixture).QuantumNumberToStr(Scomp)]);
  if Npower > 0 then
    Result := Result + 'N^' + IntToStr(Npower)
  else if Npower < 0 then
    Result := Result + 'J^' + IntToStr(-Npower);
  if SymSelect >= 0 then
    Result := Result + TMolecule(AncestorWithClass(TMolecule)).JAdjustedSymName(0,SymSelect);
  if OmegaSelect > -9999 then
    Result := Result + ' Om=' + (RootNode as TMixture).QuantumNumberToStr(OmegaSelect);
end;

function TLinearPerturbation.ReadProperties(Node: TDOMNode; CheckUnread: Boolean):Integer;
var
  ParSel: string;
begin
  Result := inherited ReadProperties(Node, False);
  { For backwards copatibility }
  ParSel := ReadValue(Node, 'ParitySelect', '');
  if (ParSel <> '') and (ParSel <> 'all') then
    if ParSel = 'e_only' then
      SymSelect := 0
    else if ParSel = 'f_only' then
      SymSelect := 1
    else
      raise EReportException.Create('Bad ParitySelect in '+Name+': '+ParSel);
  if CheckUnread then
    Result := Result + nUnreadProperties(Node);
end;

constructor TLinearPerturbation.Create(AOwner: TComponent);
begin
  SavedProperties := SavedProperties + ',OmegaSelect=all,n=1';
  inherited;
end;

const
 LinearPerturbationSettingNames: array [LPI..LPI+3] of string = (
'Perturbation type.',
'Power of operator in perturbation.',
'Power of N (for centrifugal distrtion of perturbation).',
'Omega Selection.'
 { TODO: Srank, Scomp }
);

function TLinearPerturbation.HelpOnSetting(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearPerturbationSettingNames))
     and (AIndex <= High(LinearPerturbationSettingNames)) then
    Result := LinearPerturbationSettingNames[AIndex]
  else
    Result := inherited HelpOnSetting(AIndex);
end;

function TLinearPerturbation.IsSettingQuantum(AIndex: Integer): Boolean;
begin
  Result := AIndex in [LPI+3..LPI+5];
  if not Result then
    Result := inherited IsSettingQuantum(AIndex);
end;

function TLinearPerturbation.SettingToStr(AIndex: Integer): string;
begin
  if (AIndex = LPI+3) and (OmegaSelect <= -9999) then
      Result := 'all'
  else
    Result := inherited SettingToStr(AIndex);
end;

function TLinearPerturbation.SymbolicMatrixElements: TMatrixElements;
var
  PE: TPertEquiv;
begin
  PE.Op := Op;
  PE.n := n;
  PE.NPower := Npower;
  PE.Srank := Srank;
  PE.Scomp := Scomp;
  PE.Num := 1;
  PE.Den := 1;
  PE.Name := PairName;
  PE.SymSelect := SymSelect;
  PE.OmegaSelect := OmegaSelect;
  Result := HundA.SymbolicRotationalMatrixElement(PE, Owner.Owner as TLinearMolecule, Bra as TLinear, Ket as TLinear);
end;

procedure TLinearPerturbation.SetSettingFromStr(AIndex: Integer; const s: string);
begin
  if (AIndex = LPI+3) and (s = 'all') then
    OmegaSelect := -9999
  else
    inherited;
end;

{ TLinearNucleus }

const
 LinearNucleusParameterNames: array [1..13] of string = (
'Nuclear spin - orbit interaction (I.L).',
'Nuclear spin - electron spin (I.S). Note that b = bF - c/3.',
'Nuclear spin - electron spin dipole dipole interaction diagonal in Lambda.',
'Nuclear spin - electron spin dipole dipole interaction off diagonal in Lambda by 2.',
'Nuclear quadrupole coupling constant diagonal in Lambda.',
'Nuclear quadrupole coupling constant off diagonal in Lambda by 2.',
'Nuclear spin rotation (I.J).',
'Centrifugal distortion of nuclear spin - orbit interaction (I.L). (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin (I.S). (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin dipole dipole interaction diagonal in Lambda. (Outline implementation)',
'Centrifugal distortion of nuclear spin - electron spin dipole dipole interaction off diagonal in Lambda by 2. (Outline implementation)',
'Centrifugal distortion of nuclear quadrupole coupling constant diagonal in Lambda. (Outline implementation)',
'Centrifugal distortion of nuclear quadrupole coupling constant off diagonal in Lambda by 2. (Outline implementation)'
);

function TLinearNucleus.HasIdentical: Boolean;
begin
  Result := (Molecule as TLinearMolecule).Symmetric;
end;

function TLinearNucleus.HelpOnParameter(AIndex: Integer): string;
begin
  if (AIndex >= Low(LinearNucleusParameterNames))
     and (AIndex <= High(LinearNucleusParameterNames)) then
    Result := LinearNucleusParameterNames[AIndex]
  else
    Result := inherited HelpOnParameter(AIndex);
end;

procedure CleanRn;
var
  i: Integer;
begin
  for i := 1 to High(rn) do
    SetLength(rn[i],0);
end;

var
  VList: TVariableList;

initialization
  TLinearMolecule.RegisterClass(TSpecies);
  TLinearManifold.RegisterClass(TLinearMolecule);
  TLinear.RegisterClass(TLinearManifold);
  TLinearNucleus.RegisterClass(TLinear);
  TLinearPerturbation.RegisterClass(TLinearManifold);
  TLinearSpinSpin.RegisterClass(TLinear);
  VList := MakeSortedVariables(['J'], [@eJ]);
  eJ.Value := 4.5;
  AtThreadEnd(@CleanRn);
finalization
  CleanRn;
  FreeAndNil(VList);
  FreeAndNil(SymbolicNrnEvaluator);
end.

