unit LevelFrm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, ComCtrls, ScrlCtrl, Graphs, StdCtrls, ExtCtrls,
  WhatsThis, ChildForm, MNode, Mixture, EditExtr, NewSpin,
  AlEdit, Buttons, FormUtils, Menus, PlotOC, Settings, QuantumSpin, GraphI,
  GraphM, Global, UCompat;

type

  TLevelForm = class(TChildForm)
    Panel: TPanel;
    Label1: TLabel;
    ActiveObject: TComboBox;
    ExcludePerturbations: TCheckBox;
    LG: TGraphD;
    StatusBar: TStatusBar;
    Ymax: TFixedSpin;
    Label2: TLabel;
    Label3: TLabel;
    Ymin: TFixedSpin;
    XMin: TQuantumSpin;
    Label4: TLabel;
    Label5: TLabel;
    XMax: TQuantumSpin;
    Label6: TLabel;
    Reduction: TAlignedEdit;
    Label7: TLabel;
    AllButton: TSpeedButton;
    ZoomButton: TSpeedButton;
    ZoomOutButton: TSpeedButton;
    CalcPoints: TSpeedButton;
    CalcLine: TSpeedButton;
    CalcBoth: TSpeedButton;
    JoinHighest: TCheckBox;
    ObsUpper: TSpeedButton;
    ObsLower: TSpeedButton;
    ReloadButton: TSpeedButton;
    JoinObs: TSpeedButton;
    AnotherButton: TButton;
    ListButton: TButton;
    TrackState: TCheckBox;
    TrackQuanta: TCheckBox;
    LabelsStyle: TComboBox;
    Symmetry: TSpinBox;
    Label10: TLabel;
    Label11: TLabel;
    ErrorMultiplier: TFixedSpin;
    XQuantumLabel: TLabel;
    XQuantumNumber: TSpinBox;
    Label13: TLabel;
    IgnoreSpins: TSpinBox;
    GPopupMenu: TPopupMenu;
    EditPointItem: TMenuItem;
    ASelect: TQuantumSpin;
    ALabel: TLabel;
    BSelect: TQuantumSpin;
    BLabel: TLabel;
    ShowLineItem: TMenuItem;
    ExternalField: TCheckBox;
    FieldPanel: TPanel;
    FieldLabel: TLabel;
    MinField: TFixedSpin;
    Label15: TLabel;
    MaxField: TFixedSpin;
    FieldSamples: TSpinBox;
    Label16: TLabel;
    Label17: TLabel;
    ShiftMult: TFixedSpin;
    PlotTypeGroup: TGroupBox;
    EScaleGroup: TGroupBox;
    SelectGroup: TGroupBox;
    QuantumNumberGroup: TGroupBox;
    ObservedGroup: TGroupBox;
    ExcludeGroup: TGroupBox;
    LabelsGroup: TGroupBox;
    Label8: TLabel;
    ESweep: TRadioButton;
    BSweep: TRadioButton;
    AltPlot: TCheckBox;
    RefreshPlot: TButton;
    procedure YminChange(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure GPopupMenuPopup(Sender: TObject);
    procedure EditPointItemClick(Sender: TObject);
    procedure IgnoreSpinsSetIntegerValue(Sender: TObject; NewValue: Integer;
      var NewText: string);
    procedure IgnoreSpinsGetIntegerValue(Sender: TObject; var NewValue: Integer;
      TheText: string; var StringValid: Boolean);
    procedure XQuantumNumberSetIntegerValue(Sender: TObject; NewValue: Integer;
      var NewText: string);
    procedure XQuantumNumberGetIntegerValue(Sender: TObject;
      var NewValue: Integer; TheText: string; var StringValid: Boolean);
    procedure ErrorMultiplierIncrement(Sender: TObject; var NewValue: Double;
      LastValue:Double; n: Integer);
    procedure ListButtonClick(Sender: TObject);
    procedure ExcludePerturbationsClick(Sender: TObject);
    procedure LGMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure AllButtonClick(Sender: TObject);
    procedure ZoomButtonClick(Sender: TObject);
    procedure ZoomOutButtonClick(Sender: TObject);
    procedure ReloadButtonClick(Sender: TObject);
    procedure AnotherButtonClick(Sender: TObject);
    procedure FormClose(Sender: TObject; var CloseAction: TCloseAction);
    procedure ASelectSetIntegerValue(Sender: TObject; NewValue: Integer;
      var NewText: string);
    procedure ASelectGetIntegerValue(Sender: TObject; var NewValue: Integer;
      TheText: string; var StringValid: Boolean);
    procedure ShowLineItemClick(Sender: TObject);
    procedure MaxFieldIncrement(Sender: TObject; var NewValue: Double;
      LastValue: Double; n: Integer);
    procedure MinFieldIncrement(Sender: TObject; var NewValue: Double;
      LastValue: Double; n: Integer);
    procedure YminIncrement(Sender: TObject; var NewValue: Double;
      LastValue: Double; n: Integer);
    procedure RefreshPlotClick(Sender: TObject);
  private
    EigenvaluesValid: Boolean;
    { For plotted points save memory by encoding Manifold, Symmetry and J in Index }
    { Bit pack: Manifold index (8 bits); Symmetry (8 bits); F (16 bits) }
    ResidualPlotter: TObsNumberPlotter;
    YmaxIndex, YminIndex: Integer;
    function OnGetPointInfo(Index: Integer; Extra: PtrInt):string;
    procedure MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer); override;
    function ReadValues(ShowError: Boolean): Boolean;
//    function ReadVal(ValText, ValName: string; var Value: Double; ShowError: Boolean):Boolean;
    procedure CheckValue(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
  protected
    PlottedPoints: TPlottedPoints;
    nPlottedPoints: Integer;
    procedure ParameterChanged(MNode: TMNode; Index: Integer); override;
    procedure SettingChanged(MNode: TMNode; Index: Integer); override;
    function StuffObject(const Name1, Name2: string; MNode: TMNode; var ParentNode: TTreeNode): Integer; override;
    procedure HandleControl(Control: TControl; var PName, AValue, PHint: string; var Save: Boolean; var SettingType: TSettingType; Write: Boolean); override;
    procedure SetParameter(AIndex: Integer; const Value: Double); override;
    function FormSettingsName: string; override;
  public
    procedure Loaded; override;
    destructor Destroy; override;
    function BaseNodeClass: TClass; override;
    procedure ShowItem(MNode: TMNode); overload; override;
    procedure ShowItem(MNode: TMNode; Log, ShowErrors: Boolean); overload;
    procedure StuffSelector(MNode: TMNode); override;
    property ETop: Double index 1 read GetParameter write SetParameter;
    property EBottom: Double index 2 read GetParameter write SetParameter;
  end;

var
  LevelForm: TLevelForm;

implementation

uses Obs, Gmarks, GStyle, MathSup, LogUnit, SlaveEdit,
     QuantumNumber, QuantumMath, MNodeColour, Math, Worker, Progress,
     SettingsForm, LevelPlotter, Main;

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TLevelForm }

function TLevelForm.BaseNodeClass: TClass;
begin
  Result := TMixture;
end;

procedure TLevelForm.YminChange(Sender: TObject);
begin
  if MainForm.CalculationInProgress then
    Exit;
  if Lockout > 0 then
    ReStuff
  else if Assigned(BaseNode) and (BaseNode as TMixture).AutoReplot then
    RefreshPlotClick(Sender);
end;

procedure TLevelForm.YminIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := LastValue + n*(Ymax.Value-Ymin.Value)/10;
end;

destructor TLevelForm.Destroy;
begin
  FreeAndNil(ResidualPlotter);
  inherited;
end;

procedure TLevelForm.ParameterChanged(MNode: TMNode; Index: Integer);
begin
  YminChange(nil);
end;

procedure TLevelForm.SetParameter(AIndex: Integer; const Value: Double);
begin
  inherited;
  if AIndex = YminIndex then
    YMin.Value := Value
  else if AIndex = YmaxIndex then
    YMax.Value := Value
end;

procedure TLevelForm.SettingChanged(MNode: TMNode; Index: Integer);
begin
  YminChange(nil);
end;

procedure TLevelForm.ShowItem(MNode: TMNode; Log, ShowErrors: Boolean);

  procedure UpdateQnoBox(Lbl: TLabel; SpinBox: TQuantumSpin; QnoType: TQnoType);
  var
    QnoInfo: TQnoInfo;
    Node: TMNode;
    V: Integer;
  begin
    if not Assigned(MNode) then begin
      SpinBox.Value := SpinBox.Value;
      Exit;
    end else if MNode is TMixture then
      QnoInfo := TMixture(MNode).QnoInfo(QnoType)
    else if MNode is TSpecies then
      QnoInfo := TSpecies(MNode).QnoInfo(QnoType)
    else begin
      Node := MNode;
      while not (Node is TMolecule) do
        Node := Node.Owner as TMNode;
      QnoInfo := (Node as TMolecule).QnoInfo(QnoType)
    end;
    Lbl.Caption := QnoInfo.Name;
    V := SpinBox.Value - SpinBox.MinValue;
    if QnoInfo.MinValue > Low(QnoInfo.MinValue) then
      Dec(QnoInfo.MinValue,QnoInfo.StepSize);
    if QnoInfo.StepSize > 1 then begin
      V := (V div QnoInfo.StepSize) * QnoInfo.StepSize;
      SpinBox.UseGlobalFormat := True;
    end else begin
      SpinBox.UseGlobalFormat := False;
      SpinBox.Format := qft2J;
    end;
    SpinBox.Increment := QnoInfo.StepSize;
    SpinBox.MinValue := QnoInfo.MinValue;
    SpinBox.MaxValue := QnoInfo.MaxValue;
    SpinBox.Value := Min(V+QnoInfo.MinValue,QnoInfo.MaxValue);
  end;

begin
  if (csLoading in ComponentState) or not Visible then
    Exit;
  FieldPanel.Visible := ExternalField.Checked;
  { Force display of settings as text }
  try
    if Assigned(BaseNode) then
      BaseNode.CheckProperties;
    UpdateSymmetrySpin(Symmetry, MNode);
    UpdateQnoBox(ALabel, ASelect, qtA);
    UpdateQnoBox(BLabel, BSelect, qtB);
    IgnoreSpins.Value := IgnoreSpins.Value;
    XMin.Value;
    XMax.Value;
    if FieldPanel.Visible then begin
      MinField.Value;
      MaxField.Value;
      XQuantumNumber.Visible := False;
      XQuantumLabel.Caption := 'M range:';
      if ESweep.Checked then
        FieldLabel.Caption := 'field V/m:'
      else
        FieldLabel.Caption := 'field T:'
    end else begin
      XQuantumLabel.Caption := 'x Axis Type';
      XQuantumNumber.Value := XQuantumNumber.Value;
      XQuantumNumber.Visible := True;
    end;
    Reduction.Enabled := not FieldPanel.Visible;
    JoinHighest.Enabled := not FieldPanel.Visible;
  except
    on E: Exception do begin
      StatusBar.SimpleText := E.Message;
      if ShowErrors then
        MessageDlg(E.Message, mtError, [mbOk], 0);
      Exit;
    end;
  end;
  if not ReadValues(ShowErrors) then
    Exit;
  { Update Node display }
  if ExternalField.Checked then
    if ESweep.Checked then
      Caption := 'Electric Field Plot'
    else
      Caption := 'Magnetic Field Plot'
  else
    Caption := 'Energy Level Plot';
  if not Assigned(MNode) then Exit;
  Caption := MNode.UniqueName + ' - ' + Caption;
  SetItemIndex(ActiveObject,MNode);
  ResidualPlotter.SetValid(True);
  BeginUpdate;
  try
    (Self as TLevelPlotter).PlotItem(LG,MNode,Log);
    CurrentMNode := MNode;
    EigenvaluesValid := True;
  finally
    EndUpdate(False);
  end;
end;

procedure TLevelForm.ShowItem(MNode: TMNode);
begin
  ShowItem(MNode, False, False);
end;

function TLevelForm.StuffObject(const Name1, Name2: string; MNode: TMNode;
  var ParentNode: TTreeNode): Integer;
begin
  Result := 0;
  if    ( (MNode is TMixture) and (TMixture(MNode).Count >1) )
     or ( (MNode is TSpecies) and (TSpecies(MNode).Count >1) )
     or ( (MNode is TMolecule) and (TMolecule(MNode).nManifolds >1) )
     or ( (MNode is TManifold) and (TManifold(MNode).nStates <>1) )
     or ( MNode is TState ) then begin
    ActiveObject.AddItem(Name1 + Name2, MNode);
    Result := ActiveObject.Items.Count-1;
  end;
end;

procedure TLevelForm.StuffSelector(MNode: TMNode);
begin
  StuffSelectorHelper(MNode,ActiveObject);
  FComboBoxUpdated := False;
end;

procedure TLevelForm.ExcludePerturbationsClick(Sender: TObject);
begin
  if Lockout > 0 then
    ReStuff
  else
    ShowItem(CurrentMNode,False,True);
end;

procedure TLevelForm.LGMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  ResidualPlotter.LastLineNumber := -1;
  ShowGraphInfo(Sender as TGraphD, StatusBar, PlottedPoints, {$IFDEF FPC}@{$ENDIF}OnGetPointInfo);
end;

procedure TLevelForm.GPopupMenuPopup(Sender: TObject);
begin
  EditPointItem.Enabled := ResidualPlotter.LastLineNumber >= 0;
  ShowLineItem.Enabled := ResidualPlotter.LastLineNumber >= 0;
end;

procedure TLevelForm.HandleControl(Control: TControl; var PName, AValue,
  PHint: string; var Save: Boolean; var SettingType: TSettingType;
  Write: Boolean);
begin
  inherited;
  if (Control = Reduction) or (Control = ErrorMultiplier)
     or (Control = Ymax) or (Control = Ymin)
     or (Control = ShiftMult) or (Control = MinField) or (Control = MaxField) then
    SettingType := stDouble;
end;

procedure TLevelForm.AllButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginUpdate;
  with Self as TLevelPlotter do begin
    d := (EMax-Emin)/20;
    if d = 0 then d := 1;
    ETop := EMax+d;
    EBottom := EMin-d;
  end;
  EndUpdate(False);
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.ZoomButtonClick(Sender: TObject);
var
  d, b, t: Double;
begin
  BeginUpdate;
  if LG.AreaSelected then begin
    d := Abs(LG.Ymax-LG.Ymin)/LG.Height;
    t := RoundToIncrement(LG.SelectedMathArea.Top,d);
    b := RoundToIncrement(LG.SelectedMathArea.Bottom,d);
    InOrder(b, t);
    EBottom := b;
    ETop := t;
  end else if ReadValues(True) then begin
    d := (ETop - EBottom)/4;
    Etop := Etop - d;
    EBottom := EBottom + d;
  end;
  EndUpdate(False);
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.CheckValue(Control: TControl; PName, Value, PHint: string; SettingType: TSettingType; Data: TObject);
var
  T: Double;
begin
  if SettingType = stDouble then
    if not TryStrToFloatEx(Value, T) then begin
      Value := 'Bad value for ' + PName + ':' + Value;
      StatusBar.SimpleText := Value;
      if Assigned(Data) then
        MessageDlg(Value, mtError, [mbOK], 0);
      Abort;
    end;
end;

function TLevelForm.ReadValues(ShowError: Boolean): Boolean;
var
  i: Integer;
  ValText: string;
begin
  Result := False;
  try
    ForAllControls({$IFDEF FPC}@{$ENDIF}CheckValue,{$IFDEF FPC}TObject(PtrUInt{$ELSE}(Pointer{$ENDIF}(ShowError)));
  except
    on EAbort do
      Exit;
  end;
  WriteSettings(FormSettings,False,True);
  if (FormSettings.IndexOfParameter('Ymax') <> 1)
     or (FormSettings.IndexOfParameter('Ymin') <> 2) then
    raise NotYet;
  Result := True;
end;

procedure TLevelForm.RefreshPlotClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, False, False);
end;

procedure TLevelForm.ZoomOutButtonClick(Sender: TObject);
var
  d: Double;
begin
  BeginUpdate;
  if ReadValues(True) then begin
    d := (ETop - EBottom)/2;
    Etop := Etop + d;
    EBottom := EBottom - d;
  end;
  EndUpdate(False);
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.ReloadButtonClick(Sender: TObject);
begin
  if not SaveAllWindows then Exit;
  LoadAnyObservations;
  ExcludePerturbationsClick(nil);
end;

procedure TLevelForm.AnotherButtonClick(Sender: TObject);
var
  NewLevelForm: TLevelForm;
begin
  NewLevelForm := TLevelForm.Create(Application);
  NewLevelForm.Show;
  NewLevelForm.BaseNode := BaseNode;
end;

procedure TLevelForm.ASelectGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if SameText(TheText, 'all') or (TheText = '') then begin
    NewValue := (Sender as TSpinBox).MinValue;
    StringValid := True;
  end;
end;

procedure TLevelForm.ASelectSetIntegerValue(Sender: TObject; NewValue: Integer;
  var NewText: string);
begin
  if NewValue = (Sender as TSpinBox).MinValue then
    NewText := 'all';
end;

procedure TLevelForm.FormClose(Sender: TObject; var CloseAction: TCloseAction);
begin
  if Self <> LevelForm then
    CloseAction := caFree;
end;

procedure TLevelForm.FormCreate(Sender: TObject);
begin
  LG.StuffPopupMenu(GPopupMenu);
end;

function TLevelForm.FormSettingsName: string;
begin
  Result := 'LevelWindow'
end;

procedure TLevelForm.ListButtonClick(Sender: TObject);
begin
  ShowItem(CurrentMNode, True, True);
  ReShow(LogForm)
end;

procedure TLevelForm.Loaded;
begin
  inherited;
  ResidualPlotter := TObsNumberPlotter.Create('');
  UpdateSymmetrySpin(Symmetry, nil);
  YmaxIndex := FormSettings.IndexOfParameter('Ymax');
  YminIndex := FormSettings.IndexOfParameter('Ymin');
end;

procedure TLevelForm.MaxFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := SpinScaleBy(n,2,LastValue);
end;

procedure TLevelForm.MinFieldIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := LastValue + n*(MaxField.Value-LastValue)/10;
end;

procedure TLevelForm.MNodeChange(Sender: TMNode; Change: TMNodeChange; Index: Integer);
begin
  { TODO: Respond to mnParameterChange, mnSettingChange only if Autoreplot set, and not during fitting }
  inherited;
  if Change = mnCacheCleared then
    EigenvaluesValid := False;
end;

function TLevelForm.OnGetPointInfo(Index: Integer; Extra: PtrInt): string;
var
  Ev: TEigenvalues;
begin
  if Index >= 0 then
    Result := ResidualPlotter.OnGetPointInfo(Index, Extra)
  else if EigenvaluesValid then begin
    Ev := TEigenvalues(Extra);
    Result := ' ' + Ev.DominantLabel(-Index,GlobalMixture.QuantumNumberFormat)
                  + ':' + IntToStr(-Index);
  end else
    Result := ' ' + IntToStr(-Index);
end;

procedure TLevelForm.EditPointItemClick(Sender: TObject);
begin
  ResidualPlotter.EditLastPoint;
end;

procedure TLevelForm.ShowLineItemClick(Sender: TObject);
begin
  ResidualPlotter.ShowLastPoint;
end;

procedure TLevelForm.ErrorMultiplierIncrement(Sender: TObject; var NewValue: Double; LastValue: Double; n: Integer);
begin
  NewValue := Max(1, SpinScaleBy(n,10,LastValue));
end;

procedure TLevelForm.XQuantumNumberGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
var
  i: Integer;
begin
  for i := 0 to High(QnoTypeNames)-1 do
    if QnoTypeNames[i] = TheText then begin
      NewValue := i;
      StringValid := True;
      Exit;
    end;
end;

procedure TLevelForm.XQuantumNumberSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  NewText := QnoTypeNames[NewValue];
end;

procedure TLevelForm.IgnoreSpinsGetIntegerValue(Sender: TObject;
  var NewValue: Integer; TheText: string; var StringValid: Boolean);
begin
  if TheText = 'all' then begin
    StringValid := True;
    NewValue := 0;
  end;
end;

procedure TLevelForm.IgnoreSpinsSetIntegerValue(Sender: TObject;
  NewValue: Integer; var NewText: string);
begin
  if NewValue = 0 then
    NewText := 'all';
end;

end.
