unit Life;

interface

type
  { How to deal with (state dependent) width }
  TLifeModel = (lmNone=0,       { ignore - discard width }
                lmWidth,        { Simply take width - peak intensity scales as 1/width }
                lmProduct,      { Scale Peak intensity only as width/(width + RefWidth); discard width }
                lmProductWidth, { Use width, and additionally scale intensity as width/(width + RefWidth) }
                lmParent,       { Scale Peak intensity only as 1/(width + RefWidth); discard width }
                lmParentWidth,  { Use width, and additionally scale intensity as 1/(width + RefWidth) }
                lmGate,         { Scale Peak intensity only as exp(-width * RefWidth); discard width }
                lmGateWidth     { Use width, and additionally scale intensity as exp(-width * RefWidth) }
               );
  { Notes:

    RefWidth is a property of the spectrum, rather than the molecule. It is only essential for
    the lmGate/lmGateWidth cases; it can otherwise be left at the default of 0.

    lmProduct and lmProductWidth are designed to model the case where the result of a predissociation
    or other process is being detected, with a rate proprotional to the given width, i.e. zero width
    corresponds to zero signal. In this case RefWidth is a measure of the strength of any competing
    process. As a special case, A RefWidth of 0 is treated as infinity

    lmParent and lmParentWidth are designed to model the case where predissociation or other process
    results in loss of the species being detected, so the larger the width (=rate) the smaller the signal.
    In this case RefWidth is the linewidth in the absence of the loss process.

    lmGate and lmGateWidth are designed to model detecting fluorescence excited by a short pulse, where
    the sample (=gate) time is less than the duration of the fluorescence. Only a fraction of the florescence
    is detected from states with a long lifetime (and thus small width), so transitions involving these
    states appear relatively weakly. In this case RefWidth is proportional to the gate width.

    lmNone, lmProduct, lmParent and lmGate modes are strictly physically unreasonable,
    but are much faster and give the same results if the state dependent width is much less than the instrumental width.
  }

implementation

end.
 
