unit ObsSelect;
{$I DFS.inc}{Copyright (C) 2007 C M Western; see commandline.pas for license}
interface
{ Mechanism for selecting observations by quantum number }

uses Mixture, Expr, QuantumNumber, Contnrs;

function AddSelection(var Selections: TObjectList; Selection: string; QuantumNumberFormat: TQuantumNumberFormat):string;
function CheckSelections(Obs: TObservation; Selections: TObjectList):Boolean;
function CheckSelection(var Obs: TObservation; Expr: TExpr):Boolean;

implementation

uses SysUtils, Fitter;

type
  PObservation = ^TObservation;

var
  CurrentObs: PObservation; { GLobal variable - not threadsafe }

type
  TComputedVariable = class (TArray)
  private
    ObsType: TObsType;
  public
    constructor Create(AName: String; AObsType: TObsType);
    function Value(var a:array of Double):Double; override;
  end;

var
  Parser: TLex;
  Ju, Jl,
  Frequency, Weight, StdDev, SymU, SymL, IndexU, IndexL: TVariable;
  EU, EL, Intensity: TFunction;


function AddSelection(var Selections: TObjectList; Selection: string; QuantumNumberFormat: TQuantumNumberFormat):string;
var
  Expr: TExpr;
begin
  Expr := TExpr.Create(Parser);
  Result := '';
  try
    Expr.Text := Selection;
    if not Assigned(Selections) then
      Selections := TObjectList.Create;
    Selections.Add(Expr);
  except
    on E: Exception do begin
      Result := E.Message;
      Expr.Free;
    end;
  end;
end;

function CheckSelection(var Obs: TObservation; Expr: TExpr):Boolean;
begin
  CurrentObs := @Obs;
  Ju.Value := Obs.JU / 2;
  Jl.Value := Obs.JL / 2;
  Frequency.Value := Obs.Frequency;
  Weight.Value := Obs.Weight;
  StdDev.Value := Obs.Weight;
  SymU.Value := Obs.SymU;
  SymL.Value := Obs.SymL;
  IndexU.Value := Obs.IndexU;
  IndexL.Value := Obs.IndexL;
  Result := Expr.Value <> 0;
end;

function CheckSelections(Obs: TObservation; Selections: TObjectList):Boolean;
var
  i: Integer;
begin
  Result := False;
  if Assigned(Selections) then
    for i := Selections.Count - 1 downto 0 do
      if not CheckSelection(Obs, Selections[i] as TExpr) then
        Exit;
  Result := True;
end;

{ TComputedVariable }

constructor TComputedVariable.Create(AName: String; AObsType: TObsType);
begin
  inherited Create(AName,0);
  ObsType := AObsType;
end;

function TComputedVariable.Value(var a: array of Double): Double;
var
  s: string;
begin
  if not CalculateObs(CurrentObs^, Result, s, False, ObsType) then
    Abort;
end;

initialization
  Parser := TLex.Create;
  Ju := Parser.MakeVariable('J''');
  Jl := Parser.MakeVariable('J"');
  Frequency := Parser.MakeVariable('Frequency');
  Weight := Parser.MakeVariable('Weight');
  StdDev := Parser.MakeVariable('StdDev');
  SymU := Parser.MakeVariable('Sym''');
  SymL := Parser.MakeVariable('Sym"');
  IndexU := Parser.MakeVariable('Index''');
  IndexL := Parser.MakeVariable('Index"');
  EU := Parser.AddFunction(TComputedVariable.Create('E''', obEUpper));
  EL := Parser.AddFunction(TComputedVariable.Create('E"', obELower));
  Intensity := Parser.AddFunction(TComputedVariable.Create('Intensity', obIntensity));
finalization
  Parser.Free;
  Ju.Free; Jl.Free;
  Frequency.Free; Weight.Free; StdDev.Free;
  SymU.Free; SymL.Free;
  IndexU.Free; IndexL.Free;
  EU.Free; EL.Free; Intensity.Free;
end.
