unit PGWork;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Worker, Mixture, Interlock, SpectrumPlot, SysUtils;

type
  TWorkType = (wtSkip, wtCalculateObs, wtAddLine);

  PException = ^Exception;

  TAddIn = record
    Plot: TSpectrumPlot;
    LineList: TLineList;
    Simulation: TSimulation;
    Problem: PException;
    FirstLine: Integer;
    Increment: Integer;
  end;

  TPgopherWorker = class(TInterlockedWorker)
  private
    WorkType: TWorkType;
    {wtCalculateObs}
    APObs: PObservation;
    APValue: PDouble;
    APFailed: PBoolean;
    AI: TAddIn;
  public
    procedure DoWork; override;
    procedure Skip;
    procedure CalculateObs(PObs: PObservation; PValue: PDouble; PFailed: PBoolean);
    procedure AddIn(Plot: TSpectrumPlot; LineList: TLineList; Simulation: TSimulation; var Problem: Exception; FirstLine: Integer; Increment: Integer);
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     MNode, Fitter, Profile, Clock, UCompat;

{ TPgopherWorker }

var
  { One per object might be fastest, but might end up with a lot of unused objects }
  AddInCriticalSection: {$IFDEF FPC}System.{$ENDIF}TRTLCriticalSection;

procedure TPgopherWorker.AddIn(Plot: TSpectrumPlot; LineList: TLineList;
  Simulation: TSimulation; var Problem: Exception; FirstLine,
  Increment: Integer);
begin
  AI.Plot := Plot;
  AI.LineList := LineList;
  AI.Simulation := Simulation;
  AI.Problem := @Problem;
  AI.FirstLine := FirstLine;
  AI.Increment := Increment;
  WorkType := wtAddLine;
  StartWork;
end;

procedure TPgopherWorker.CalculateObs(PObs: PObservation; PValue: PDouble; PFailed: System.PBoolean);
begin
  APObs := PObs;
  APValue := PValue;
  APFailed := PFailed;
  WorkType := wtCalculateObs;
  StartWork;
end;

procedure TPgopherWorker.DoWork;
var
  NotUsed: string;
  LocalProblem: Exception;
begin
  NotUsed := '';
  case WorkType of
    wtSkip: { Make worker free again };
    wtCalculateObs: if not Fitter.CalculateObs(APObs^, APValue^, NotUsed, False, APObs^.ObsType) then
                      APFailed^ := False;
    wtAddLine: with AI do begin
                 LocalProblem := nil;
                 Plot.AddIn(LineList, Simulation, LocalProblem, FirstLine, Increment);
                 if Assigned(LocalProblem) then begin
                   {$IFDEF FPC}System.{$ENDIF}EnterCriticalSection(AddInCriticalSection);
                   try
                     if Assigned(AI.Problem^) then
                       LocalProblem.Free
                     else
                       AI.Problem^:= LocalProblem;
                   finally
                     {$IFDEF FPC}System.{$ENDIF}LeaveCriticalSection(AddInCriticalSection);
                   end;
                 end;
               end
  else
    WorkType := wtSkip;
    raise NotYet;
  end;
  WorkType := wtSkip;
  inherited;
end;

procedure TPgopherWorker.Skip;
begin
  WorkType := wtSkip;
  StartWork;
end;

initialization
  {$IFDEF FPC}System.{$ENDIF}InitCriticalSection(AddInCriticalSection);
finalization
  {$IFDEF FPC}System.{$ENDIF}DoneCriticalSection(AddInCriticalSection);
end.

