unit RenForm;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses
  SysUtils, Variants, Classes, Graphics, Controls, Forms,
  WhatsThis, Dialogs, StdCtrls, Buttons, MNode;

type
  TRenameForm = class(THelpForm)
    OKButton: TButton;
    CancelButton: TButton;
    BraComboBox: TComboBox;
    Label1: TLabel;
    Label2: TLabel;
    KetComboBox: TComboBox;
  private
    MNodePair: TMNodePair;
  public
    function StuffComboBox(MNode: TMNode; ComboBox: TComboBox; Current: TMNode; CurrentName: string): string;
    function Rename(MNode: TMNodePair; ChangingMNode: TMNode = nil; DeleteOnCancel: Boolean = False):Boolean;
    function RenameMNodePairsAsRequired(MNode: TMNode; DeleteOnCancel: Boolean):Boolean;
    function UnStuffComboBox(MNode: TMNode; ComboBox: TComboBox): string;
  end;

var
  RenameForm: TRenameForm;

implementation

{$IFNDEF FPC}
{$R *.dfm}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{ TRenameForm }

function TRenameForm.Rename(MNode: TMNodePair; ChangingMNode: TMNode; DeleteOnCancel: Boolean):Boolean;
var
  E: Exception;
  BraDef, KetDef: string;
begin
  Result := True;
  DeleteOnCancel := DeleteOnCancel and (ChangingMNode = MNode);
  MNodePair := MNode;
  if (MNode.BraOwner = nil) or (MNode.KetOwner = nil) then
    raise Exception.Create('Fix parent node first');
  BraDef := StuffComboBox(MNode.BraOwner, BraComboBox, MNode.Bra, MNode.FBraName);
  KetDef := StuffComboBox(MNode.KetOwner, KetComboBox, MNode.Ket, MNode.FKetName);
  if not(Assigned(ChangingMNode) and (BraDef <> '') and (KetDef <> '')) then begin
    Caption := 'Rename ' + MNode.UniqueName;
    if ShowModal <> mrOK then begin
      Result := DeleteOnCancel;
      if Assigned(ChangingMNode) then
        if DeleteOnCancel then begin
          ChangingMNode.Free;
          Result := False;
          Exit;
        end
      else
        Exit; { No action taken }
    end;
  end;
  E := MNode.SetBraKetNameE(UnStuffComboBox(MNode.BraOwner, BraComboBox),
                            UnStuffComboBox(MNode.KetOwner, KetComboBox), ChangingMNode);
  { If child nodes need renaming as a result of this name change,
    prompt for this }
  if Assigned(E) then begin
    if Assigned(ChangingMNode) then
      RenameMNodePairsAsRequired(ChangingMNode, False)
    else
      RenameMNodePairsAsRequired(MNode, False);
    { Re-issue notifications, as these may have been skipped }
    MNode.Modified := True;
    PerformMNodeNotification(MNode, mnNameChange, -1);
  end;
end;

function TRenameForm.RenameMNodePairsAsRequired(MNode: TMNode; DeleteOnCancel: Boolean):Boolean;
var
  ProblemMNode: TMNodePair;
  E: Exception;
begin
  Result := True;
  { If a paste MNode or other change means that the Bra or Ket of mnode pairs need
    changing, this will prompt as required }
  E := MNode.FixupProperties;
  while Assigned(E) do
    if E is ECantFindState then begin
      ProblemMNode := ECantFindState(E).MNodePair;
      E.Free;
      if Rename(ProblemMNode, MNode, DeleteOnCancel) then
        E := MNode.FixupProperties
      else begin
        Result := False;
        Exit;
      end;
    end else if E is EReportException then
      FreeAndNil(E)
    else if Assigned(E) then
      raise E;
end;

function TRenameForm.StuffComboBox(MNode: TMNode; ComboBox: TComboBox; Current: TMNode; CurrentName: string): string;
var
  I: Integer;
begin
  if Assigned(Current) then
    CurrentName := Current.Name;
  ComboBox.Clear;
  for I := 0 to MNode.ComponentCount-1 do
    if MNode.Components[I] is MNodePair.BraKetClass then begin
      ComboBox.Items.Add(TMNode(MNode.Components[i]).Name);
      if TMNode(MNode.Components[I]).Name = CurrentName then
        ComboBox.ItemIndex := ComboBox.Items.Count-1;
    end;
  if ComboBox.Items.Count = 1 then begin
    Result := ComboBox.Items[0];
    ComboBox.ItemIndex := 0;
  end else
   Result := '';
end;

function TRenameForm.UnStuffComboBox(MNode: TMNode; ComboBox: TComboBox): string;
var
  I, J: Integer;
begin
  Result := '';
  J := ComboBox.ItemIndex;
  for I := 0 to MNode.ComponentCount-1 do
    if MNode.Components[I] is MNodePair.BraKetClass then
      if J = 0 then begin
        Result := TMNode(MNode.Components[I]).Name;
        Exit;
      end else
        Dec(J);
end;

end.
