unit StaticField;
{$I DFS.inc}{Copyright (C) 2009 C M Western; see commandline.pas for license}
interface

uses Types, Math;

type

  TStaticFields = record
    E: Double; { Electric Field, V/m }
    B: Double; { Magnetic Field, Tesla }
    Angle: Double; { Angle between fields, degrees }
  end;

const
  ZeroField: TStaticFields = (E:0; B:0; Angle:0);

function IsNonZero(const StaticField: TStaticFields):Boolean;
function SameFields(const A, B: TStaticFields):Boolean;
function FieldsDiff(const A, B: TStaticFields):TValueRelationship;

implementation

function IsNonZero(const StaticField: TStaticFields):Boolean;
begin
  Result := (StaticField.E <> 0) or (StaticField.B <> 0);
end;

function SameFields(const A, B: TStaticFields):Boolean;
begin
  Result := SameValue(A.E,B.E) and SameValue(A.B,B.B);
end;

function FieldsDiff(const A, B: TStaticFields):TValueRelationship;
begin
  { Return sign of difference for sorting }
  Result := CompareValue(A.E, B.E);
  if Result = EqualsValue then
    Result := CompareValue(A.B, B.E);
  if Result = EqualsValue then
    Result := CompareValue(A.Angle, B.Angle);
end;

end.
