unit Clock;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see UCompat.pas for license}
interface
{$IFDEF CPUX86_64}{$DEFINE CPU86XX}{$ENDIF}
{$IFDEF CPU86}{$DEFINE CPU86XX}{$ENDIF}

type
  TClockFunction = function : Int64;
var
  Clocks: TClockFunction;

//function Clocks: int64; {$IFNDEF MSWINDOWS}{$IFDEF CPU86XX} assembler; {$ENDIF}{$ENDIF}
function ClockFrequency: Double;

procedure SetTimingMethod(Method: Integer);
function GetTimingMethod: Integer;
function TimingMethodName(Method: Integer):string;

implementation

uses {$IFDEF MSWINDOWS} Windows, MMSystem, {$ENDIF}
     {$IFDEF UNIX} Unix, {$ENDIF}
     Sysinfo, SysUtils;

{$IFDEF FPC}
  {$IFDEF CPU86}{$ASMMODE INTEL}{$ENDIF}
{$ENDIF}

var
  CurrentMethod: Integer;
  CurrentCount: Int64;
  { RDTSC values }
  InitialClock: int64;
  InitialTime: TDateTime;

function GetTimingMethod: Integer;
begin
  Result := CurrentMethod;
end;

function ClocksCount:Int64;
begin
  Inc(CurrentCount);
  Result := CurrentCount;
end;

{$IFDEF CPU86XX}
function ClocksRDTSC: int64; {$IFNDEF MSWINDOWS}{$IFDEF CPU86XX} assembler; {$ENDIF}{$ENDIF} forward;

function ClockFrequencyRDTSC: Double;
var
  CurrentClock: int64;
begin
  CurrentClock := ClocksRDTSC;
  if Now = InitialTime then
    Sleep(1);
  Result := (CurrentClock-InitialClock)/(Now-InitialTime)/SecsPerDay;
end;
{$ENDIF}

{$IFDEF OLD_MSWINDOWS}
uses mmSystem;
{$IFDEF PENTIUM}
  type TClock = Comp;

  procedure Clocks(var t: TClock); assembler;
  asm
    {$IFDEF WIN32}
      mov ECX, EDI
      mov EDI, t
      DB 0FH, 031H  { RDTSC }
      mov [EDI], eax
      mov [EDI+4], edx
      mov EDI, ECX
    {$ELSE}
      mov CX, DI
      les DI, t
      DB 0FH, 031H  { RDTSC }
      db 066h
      mov [DI], ax
      db 066h
      mov [DI+4], dx
      mov DI, CX
    {$ENDIF}
  end;
  {$IFDEF WIN32}
    function ClockFrequency: Double;
    var
      s, e: Comp;
      st, et, f: TLargeInteger;
      dt: Double;
    begin
      Clocks(s);
      QueryPerformanceCounter(st);
      QueryPerformanceFrequency(f);
      repeat
        Sleep(100);
        Clocks(e);
        QueryPerformanceCounter(et);
        { dt := (et.QuadPart - st.QuadPart)/f.QuadPart; }
        dt := (et - st)/f;
      until dt > 0.1;
      Result := (e-s)/dt;
    end;
  {$ELSE}
    function ClockFrequency: Double;
    var
      s, e: Comp;
      st, et: Longint;
      dt: Double;
    begin
      Clocks(s);
      st := timeGetTime;
      repeat
        Yield;
        Clocks(e);
        et := timeGetTime;
        dt := (et - st)/1000;
      until dt > 0.1;
      Result := (e-s)/dt;
    end;
  {$ENDIF}
{$ELSE}
  {$IFDEF WIN32}
    type TClock = Comp;
    procedure Clocks(var t: TClock);
    begin
      QueryPerformanceCounter(TLargeInteger(t));
    end;
    function ClockFrequency: Double;
    var
      f: Comp;
    begin
      QueryPerformanceFrequency(TLargeInteger(f));
      Result := f;
    end;
  {$ELSE}
    {$IFDEF timeGetTime}
      type TClock = Longint;
      procedure Clocks(var t: TClock);
      begin
        t := timeGetTime;
      end;
      function ClockFrequency: Double;
      begin
        Result := 1000;
      end;
    {$ELSE}
      var
        lpfnVTD: Longint;
      type TClock = Comp;
      procedure Get_VTD_Address; assembler;
      asm
        { Get address of VTD_Get_Real_Time function }
        push di
        push si
        mov ax, 01684h
        mov bx, 5
        xor di, di
        int 02fh
        mov WORD PTR lpfnVTD, di
        mov WORD PTR lpfnVTD+2, es
        pop si
        pop di
      end;
      procedure Clocks(var t: TClock); assembler;
      asm
        push di
        push si
        { Use VTD_Get_Real_Time function }
        mov ax, 0100h
        call lpfnVTD
        les DI, t
        db 066h
        mov [DI], ax
        db 066h
        mov [DI+4], dx
        pop si
        pop di
      end;
      function ClockFrequency: Double;
      begin
        Get_VTD_Address;
        Result := 1.19318e6;;
      end;
    {$ENDIF}
  {$ENDIF}
{$ENDIF}
{$ENDIF}

{$IFDEF MSWINDOWS}
function ClocksRDTSC: int64; assembler;
asm
  DB 0FH, 031H  { RDTSC }
end;

function ClocksQPC: int64;
begin
  QueryPerformanceCounter({$IFDEF FPC}@{$ENDIF}Result);
end;

function ClockstimeGetTime: int64;
begin
  Result := timeGetTime;
end;

function ClockFrequency: Double;
var
  f: Int64;
begin
  case CurrentMethod of
    0: Result := 1;
    1: begin
         QueryPerformanceFrequency({$IFDEF FPC}@{$ENDIF}f);
         Result := f;
       end;
  2,3: Result := ClockFrequencyRDTSC;
    4: Result := 1000;
  end;
end;

procedure SetTimingMethod(Method: Integer);
var
  PMask, SMask: DWORD;
begin
  if Method < 0 then
    if nCPUs = 1 then
      Method := 2
    else
      Method := 1;
  case Method of
    0: Clocks := {$IFDEF FPC}@{$ENDIF}ClocksCount;
    1: Clocks := {$IFDEF FPC}@{$ENDIF}ClocksQPC;
    2: begin
        try
          GetProcessAffinityMask(GetCurrentProcess(),PMask,SMask);
          SetThreadAffinityMask(GetCurrentThread(), SMask)
        except
          { Ignore exception }
        end;
        Clocks := {$IFDEF FPC}@{$ENDIF}ClocksRDTSC; { Very iffy on multiple CPU system }
       end;
    3: begin
         try
           SetThreadAffinityMask(GetCurrentThread(),1 shl (nCPUs-1));
         except
          { Ignore exception }
         end;
         Clocks := {$IFDEF FPC}@{$ENDIF}ClocksRDTSC;
       end;
    4: Clocks := {$IFDEF FPC}@{$ENDIF}ClockstimeGetTime;
  end;
  CurrentMethod := Method;
end;

function TimingMethodName(Method: Integer):string;
begin
  case Method of
    0: Result := 'Call Count Only';
    1: Result := 'QueryPerformanceCounter';
    2: Result := 'RDTSC';
    3: Result := 'CPU Locked RDTSC';
    4: Result := 'timeGetTime';
  else
    Result := '';
  end;
end;

{$ELSE}

{$IFDEF CPU86XX}
function ClocksRDTSC: int64; assembler;
asm
{$IFDEF FPC}
  {$IFDEF CPU64}
   {Could use $ASMMODE intel here, but gives problem in 2.0.4}
   push   %rdx
   rdtsc
   shl    $32,%rax
   // shrd   $32,%rdx,%rax
   .byte   72
   .byte   15
   .byte   172
   .byte   208
   .byte   32
   pop   %rdx
  {$ELSE}
   rdtsc
  {$ENDIF}
{$ELSE}
  DB 0FH, 031H  { RDTSC }
{$ENDIF}
end;
{$ENDIF}
{$IFDEF DARWIN}
{$IFNDEF CPU86XX}
function mach_absolute_time: Int64; cdecl; external;
function ClocksMAT: int64;
begin
  Result := mach_absolute_time;
end;
type
  mach_timebase_info_data_t = record
    numer: WORD;
    denom: WORD;
  end;
function mach_timebase_info(var mach_timebase_info_data_t):Integer; cdecl; external;
function machClockFrequncy: Double;
var
  m: mach_timebase_info_data_t;
begin
  mach_timebase_info(m);
  Result := 1e9 * m.denom / m.numer;
end;
{$ENDIF}
{$ENDIF}

function Clocksgettimeofday: int64;
var
  tz:timeval;
begin
  fpgettimeofday(@tz,nil);
  Result := int64(tz.tv_sec)*1000000 + tz.tv_usec;
end;

procedure SetTimingMethod(Method: Integer);
begin
  if Method < 0 then
    Method := 2;
  case Method of
    0: Clocks := @ClocksCount;
    1: Clocks := @Clocksgettimeofday;
    {$IFDEF CPU86XX}
    2: begin
//        try
//          GetProcessAffinityMask(GetCurrentProcess(),PMask,SMask);
//          SetThreadAffinityMask(GetCurrentThread(),1)
//        except
//          { Ignore exception }
//        end;
        Clocks := @ClocksRDTSC; { Very iffy on multiple CPU system }
       end;
    {$ELSE}
    2: Clocks := @ClocksMAT;
    {$ENDIF}
  end;
  CurrentMethod := Method;
end;

function TimingMethodName(Method: Integer):string;
begin
  case Method of
    0: Result := 'Call Count Only';
    1: Result := 'gettimeofday';
    {$IFDEF CPU86XX}
    2: Result := 'RDTSC';
    {$ELSE}
    2: Result := 'mach_absolute_time';
    {$ENDIF}
  else
    Result := '';
  end;
end;

function ClockFrequency: Double;
begin
  case CurrentMethod of
    0: Result := 1;
    1: Result := 1000000;
    {$IFDEF CPU86XX}
    2: Result := ClockFrequencyRDTSC;
    {$ELSE}
    2: Result := machClockFrequncy;
    {$ENDIF}
  end;
end;

{$ENDIF}

initialization
  InitialTime := Now;
  {$IFDEF CPU86XX}
  InitialClock := ClocksRDTSC;
  {$ENDIF}
  CurrentCount := 0;
  SetTimingMethod(-1);
end.
