unit PropHelp;

{$I DFS.inc}

interface

uses TypInfo;

function PropertyToStr(AObject: TObject; PropInfo: PPropInfo): string; overload;
function PropertyToStr(AObject: TObject; const PropName: string): string; overload;
procedure SetPropertyFromStr(AObject: TObject; PropInfo: PPropInfo; const s: string);  overload;
procedure SetPropertyFromStr(AObject: TObject; const PropName: string; const s: string); overload;
function TryStrToBoolean(s: String; out Value: Boolean): Boolean;
function StrToBoolean(s: String): Boolean;

implementation

uses SysUtils;

{$IFDEF VER80}
type
  TIntegerSet = set of 0..SizeOf(Integer) * 8 - 1;
{$ENDIF}

{$IFDEF DFS_COMPILER_3_UP}
{$DEFINE PTP}
{$ENDIF}
{$IFDEF FPC}
{$UNDEF PTP}
{$ENDIF}

function GetPropType(PropInfo: PPropInfo): PTypeInfo;
begin
  Result := PropInfo^.PropType{$IFDEF PTP}^{$ENDIF};
end;

function FloatPropertyToStr(AObject: TObject; PropInfo: PPropInfo): string;
const
{$IFNDEF VER80}
  Precisions: array[TFloatType] of Integer = (7, 15, 18, 18, 19);
{$ELSE}
  Precisions: array[TFloatType] of Integer = (7, 15, 18, 18);
{$ENDIF}
begin
  Result := FloatToStrF(GetFloatProp(AObject, PropInfo), ffGeneral,
                        Precisions[GetTypeData(GetPropType(PropInfo))^.FloatType], 0);
end;

{$IFNDEF FPC}
{$IFNDEF DFS_COMPILER_6_UP}
function SetToString(PropInfo: PPropInfo; Value: Integer; Brackets: Boolean): string;
var
  TypeInfo: PTypeInfo;
  I: Integer;
begin
  Result := '';
  TypeInfo := GetTypeData(GetPropType(PropInfo))^.CompType{$IFDEF PTP}^{$ENDIF};
  for I := 0 to SizeOf(TIntegerSet) * 8 - 1 do
    if I in TIntegerSet(Value) then begin
      if Length(Result) <> 0 then Result := Result + ',';
      Result := Result + GetEnumName(TypeInfo, I){$IFDEF VER80}^{$ENDIF};
    end;
  if Brackets then
    Result := '[' + Result + ']';
end;

function GetEnumProp(Instance: TObject; PropInfo: PPropInfo): string;
begin
  Result := GetEnumName(GetPropType(PropInfo), GetOrdProp(Instance, PropInfo)){$IFDEF VER80}^{$ENDIF};
end;
{$ENDIF}
{$ENDIF}

function PropertyToStr(AObject: TObject; const PropName: string): string;
var
  PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(AObject, PropName);
  if not Assigned(PropInfo) then
    raise Exception.Create(PropName + ' not found in ' + AObject.ClassName);
  Result := PropertyToStr(AObject, PropInfo);
end;

function PropertyToStr(AObject: TObject; PropInfo: PPropInfo): string;
begin
  Result := '';
  case PropInfo^.PropType^.Kind of
    tkInteger: Result := IntToStr(GetOrdProp(AObject, PropInfo));
    {$IFDEF FPC}
    tkBool:    if GetOrdProp(AObject, PropInfo) <> 0 then
                 Result := 'True'
               else
                 Result := 'False';
    {$ENDIF}
{$IFNDEF VER80}
    tkWChar,
{$ENDIF}
    tkChar: Result := Char(GetOrdProp(AObject, PropInfo));
    tkEnumeration: Result := GetEnumProp(AObject, PropInfo);
    tkFloat: Result := FloatPropertyToStr(AObject, PropInfo);
{$IFDEF DFS_COMPILER_4_UP}
    tkInt64: Result := IntToStr(GetInt64Prop(AObject, PropInfo));
{$ENDIF}
{$IFNDEF VER80}
    tkWString,
    tkLString,
{$IFNDEF DFS_COMPILER_3_UP}
    tkLWString,
{$ENDIF}
{$ENDIF}
{$IFDEF FPC}
    tkAString,
{$ENDIF}
    tkString: Result := GetStrProp(AObject, PropInfo);
{$IFNDEF VER80}
{$IFNDEF FPC}
    tkVariant: Result := GetVariantProp(AObject, PropInfo);
{$ENDIF}
{$ENDIF}
    tkSet: Result := SetToString(PropInfo, GetOrdProp(AObject, PropInfo), True);
  else
    raise Exception.CreateFmt('Unknown property type for %s', [PropInfo^.Name]);
{   tkClass: Not implemented }
  end;
end;

{$IFNDEF FPC}
{$IFNDEF DFS_COMPILER_6_UP}
procedure SetEnumProp(AObject: TObject; PropInfo: PPropInfo; const S: string);
var
  I: Integer;
  EnumType: PTypeInfo;
begin
  EnumType := GetPropType(PropInfo);
  with GetTypeData(EnumType)^ do
    for I := MinValue to MaxValue do
      if CompareText(GetEnumName(EnumType, I){$IFDEF VER80}^{$ENDIF}, S) = 0 then
      begin
        SetOrdProp(AObject, PropInfo, I);
        Exit;
      end;
end;

function StringToSet(PropInfo: PPropInfo; const Value: string): Integer;
const
  Delims = [' ', ',', '[', ']'];
var
  TypeInfo: PTypeInfo;
  I: Integer;
  s, EnumName: string;
begin
  Result := 0;
  TypeInfo := GetTypeData(GetPropType(PropInfo))^.CompType{$IFDEF DFS_COMPILER_3_UP}^{$ENDIF};
  s := Value;
  while s <> '' do begin
    EnumName := NextName(s);
    if EnumName <> '' then begin
      I := GetEnumValue(TypeInfo, EnumName);
      if I >= 0 then Include(TIntegerSet(Result), I);
    end;
  end;
end;

procedure SetSetProp(AObject: TObject; PropInfo: PPropInfo; const s: string);
begin
  SetOrdProp(AObject, PropInfo, StringToSet(PropInfo, s));
end;
{$ENDIF}
{$ENDIF}

function TryStrToBoolean(s: String; out Value: Boolean): Boolean;
begin
  Result := True;
  if (s = '') or (s[1] in ['n', 'N', 'f', 'F', '0']) then
    Value := False
  else if (s[1] in ['y', 'Y', 't', 'T', '1']) or (s = '-1') then
    Value := True
  else
    Result := False;
end;

function StrToBoolean(s: String): Boolean;
begin
  if not TryStrToBoolean(s,Result) then
    raise Exception.Create('Bad true/false expression: ' + s);
end;

procedure SetPropertyFromStr(AObject: TObject; const PropName: string; const s: string);
var
  PropInfo: PPropInfo;
begin
  PropInfo := GetPropInfo(AObject, PropName);
  if not Assigned(PropInfo) then
    raise Exception.Create(PropName + ' not found in ' + AObject.ClassName);
  SetPropertyFromStr(AObject, PropInfo, s);
end;

procedure SetPropertyFromStr(AObject: TObject; PropInfo: PPropInfo; const s: string);
begin
  case PropInfo^.PropType^.Kind of
    tkInteger:  SetOrdProp(AObject, PropInfo, StrToInt(s));
    {$IFDEF FPC}
    tkBool:     SetOrdProp(AObject, PropInfo, Integer(StrToBoolean(s)));
    {$ENDIF}
{$IFNDEF VER80}
    tkWChar:    SetOrdProp(AObject, PropInfo, Longint(s[1]));
{$ENDIF}
    tkChar:     SetOrdProp(AObject, PropInfo, Integer(s[1]));
    tkEnumeration: SetEnumProp(AObject, PropInfo, s);
    tkFloat:    SetFloatProp(AObject, PropInfo, StrToFloat(s));
{$IFDEF DFS_COMPILER_4_UP}
    tkInt64:    SetInt64Prop(AObject, PropInfo, StrToInt64(s));
{$ENDIF}
{$IFNDEF VER80}
    tkLString,
{$IFNDEF DFS_COMPILER_3_UP}
    tkLWString,
{$ENDIF}
    tkWString,
{$ENDIF}
{$IFDEF FPC}
    tkAString,
{$ENDIF}
    tkString:   SetStrProp(AObject, PropInfo, s);
{$IFNDEF VER80}
{$IFNDEF FPC}
    tkVariant:  SetVariantProp(AObject, PropInfo, s);
{$ENDIF}
{$ENDIF}
    tkSet:      SetSetProp(AObject, PropInfo, s);
  else
    raise Exception.CreateFmt('Unknown property type for %s', [PropInfo^.Name]);
{   tkClass: Not implemented }
  end;
end;

end.
