unit f2c;
{$I DFS.inc}
interface

uses MathSup;

type
  FtnLen = Integer;
  LogBool = LongBool;

{$IFDEF VCL_}
{$DEFINE PROVIDE_G77}
function _sqrt(a: Double): Double; cdecl;
function _log(a: Double): Double; cdecl;
function _exp(a: Double): Double; cdecl;
function _sin(a: Double): Double; cdecl;
function _cos(a: Double): Double; cdecl;
function _tan(a: Double): Double; cdecl;
function _atan(a: Double): Double; cdecl;
function _atan2(a, b: Double): Double; cdecl;
procedure __chkstk; cdecl;
function _pow(a, b: Double): Double; cdecl;
{$DEFINE PROVIDE_GFORTRAN}
{$ENDIF}
{$IFDEF FPC}
  {$DEFINE UND}
  {$IFDEF UNIX}
    {$DEFINE PROVIDE_GFORTRAN}
    {$LINKLIB m}
  {$ELSE}
    {$DEFINE PROVIDE_G77}
  {$ENDIF}
{$ENDIF}

{$IFDEF PROVIDE_GFORTRAN}
function _gfortran_pow_r8_i4(a:Double; p:Integer): Double; cdecl;
function _gfortran_pow_i4_i4(base:Integer; exponent:Integer): Integer; cdecl;
procedure _gfortran_concat_string(destlen: FtnLen; dest: PChar; len1: FtnLen; s1: PChar; len2: FtnLen; s2: PChar); cdecl;
function _gfortran_compare_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar): Integer; cdecl;
procedure _gfortran_copy_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar); cdecl;
function __powidf2(a:Double; p:Integer): Double; cdecl;
function _lroundf(a:Double): Integer; cdecl;
function _memmove(Dest, Source: PChar; length: Integer):Integer; cdecl;
{$DEFINE PROVIDE_G77}
{$ENDIF}

{$IFDEF PROVIDE_G77}
function _pow_di(var a:Double; var p:Integer): Double; cdecl;

function z_abs(var z: Complex):Double; cdecl;
procedure _z_div(var c, a, b: Complex); cdecl;
function _d_sign(var a, b: Double): Double; cdecl;
function _r_sign(var a, b: Single): Single; cdecl;

function _i_dnnt(var a: Double): Integer; cdecl;

function _s_rnge(varn: PChar; offset: Integer; procn: PChar; line: Integer):Integer; cdecl;
function s_cat(lp: PChar; rpp: PPChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl;
function s_cmp(a, b: PChar; la, lb: FtnLen): Integer; cdecl;
procedure s_copy(a, b: PChar; la, lb: FtnLen); cdecl;
{$ENDIF}

function pprint_(var tag, n: Integer; var d: Double):Integer; cdecl;

function errstop_(s: PChar; l: Integer):Integer; cdecl;
function errstopf_(fmt: PChar; args: Pointer { ... } ):Integer; cdecl;
function clog(s: PChar):Integer; cdecl;

function xerror_(MESSG: PChar; NMESSG, NERR, LEVEL: PInteger; len: Integer): Integer; cdecl;
function xerbla_(SRNAME: PChar; var Info: Integer; len: FtnLen): Integer; cdecl;
function dlamc3_(var a: Double; var b: Double): Double; cdecl;
function ieeeinfandnan_:Integer; cdecl;
function disnan_(var a:Double):Integer; cdecl;

function _tend_(var Id:Integer):Integer; cdecl;
function _tstart_(var Id:Integer):Integer; cdecl;

{$IFDEF VCL_}
function __ftol(x:Double):Integer; cdecl;

var
  __turboFloat: Integer;
{$ENDIF}

implementation

uses { Compat, } Math, Sysutils, Profile, Worker;

{$IFDEF FPC}
{$MACRO ON}
{$IFDEF DARWIN}
{$DEFINE alias_:=alias: #95}
{$ELSE}
{$IFDEF MSWINDOWS}
{$DEFINE alias_:=alias: #95}
{$ELSE}
{$DEFINE alias_:=alias:}
{$ENDIF}
{$ENDIF}
{$ENDIF}

{$IFDEF VCL_}
function _sqrt(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'sqrt'; {$ENDIF}
begin
  Result := Sqrt(a);
end;

function _log(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'log'; {$ENDIF}
begin
  Result := Ln(a);
end;

function _exp(a: Double): Double; cdecl;
begin
  Result := Exp(a);
end;

function _sin(a: Double): Double; cdecl;
begin
  Result := Sin(a);
end;

function _cos(a: Double): Double; cdecl; {$IFDEF FPC} alias: 'cos'; {$ENDIF} 
begin
  Result := Cos(a);
end;

function _tan(a: Double): Double; cdecl;
begin
  Result := Tan(a);
end;

function _atan(a: Double): Double; cdecl;
begin
  Result := ArcTan(a);
end;

function _atan2(a, b: Double): Double; cdecl;
begin
  Result := ArcTan2(a, b);
end;

function _pow(a, b: Double): Double; cdecl;
begin
  Result := Power(a,b);
end;

procedure __chkstk; cdecl;
begin
  asm
    pop edx  { Return address }
    { Grow stack in 4k chunks so that automatic stack growth works }
@2: cmp eax, 4096
    jb @1
    sub esp, 4092
    push eax
    sub eax, 4096
    jmp @2
@1: sub esp, eax
    push edx  { Apparently quicker than jmp edx on modern processors }
    ret
  end;
end;

{$ENDIF}

{$IFDEF PROVIDE_G77}
function s_cat(lp: PChar; rpp: PPChar; rnp: PInteger; var np: Integer; ll: FtnLen):Integer; cdecl; {$IFDEF FPC} alias: '_s_cat'; {$ENDIF}
var
  n, i, nc: Integer;
  rp: PChar;
  { Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  n := np;
  for i := 0 to n-1 do begin
    nc := ll;
    if rnp^ < nc then
      nc := rnp^;
    Inc(rnp);
    Dec(ll, nc);
    rp := rpp^;
    Inc(rpp);
    Move(rp^, lp^, nc);
  end;
  FillChar(lp^, ll, ' ');
  Result := 0;
end;

procedure _z_div(var c, a, b: Complex); cdecl;
var
  ratio, den : Double;
  abr, abi, cr: Double;
begin
  abr := Abs(b.re);
  abi := Abs(b.im);
  if abr <= abi then begin
    if abi = 0 then
      raise Exception.Create('complex division by zero');
    ratio := b.re / b.im ;
    den := b.im * (1 + ratio*ratio);
    cr := (a.re*ratio + a.im) / den;
    c.im := (a.im*ratio - a.re) / den;
  end else begin
    ratio := b.im / b.re ;
    den := b.re * (1 + ratio*ratio);
    cr := (a.re + a.im*ratio) / den;
    c.im := (a.im - a.re*ratio) / den;
  end;
  c.re := cr;
end;

function z_abs(var z: Complex):Double; cdecl;
var
  t: Complex;
begin
  t := z;
  if t.re < 0 then
	  t.re := -t.re;
  if t.im < 0 then
	  t.im := -t.im;
  if t.im > t.re then begin
  	Result := t.re;
	  t.re := t.im;
  	t.im := Result;
  end;
  if t.re + t.im = t.re then
    Result := t.re { Avoid Sqrt if one part negligable }
  else
    { recast to avoid overflow }
    Result := t.re*Sqrt(1 + Sqr(t.im/t.re));
end;

function _d_sign(var a, b: Double): Double; cdecl; {$IFDEF FPC} alias: '_d_sign'; {$ENDIF}
begin
  if a >= 0 then
    Result := a
  else
    Result := -a;
  if b < 0 then
    Result := -Result;
end;

function _r_sign(var a, b: Single): Single; cdecl;
begin
  if a >= 0 then
    Result := a
  else
    Result := -a;
  if b < 0 then
    Result := -Result;
end;

function _i_dnnt(var a: Double): Integer; cdecl;
begin
  Result := Round(a);
end;

function _pow_di(var a:Double; var p:Integer): Double; cdecl;  {$IFDEF FPC} alias: '_pow_di'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

function _s_rnge(varn: PChar; offset: Integer; procn: PChar; line: Integer):Integer; cdecl;
var
  msg: string;
begin
  msg := Format('Subscript out of range on file line %d, procedure %s', [line, procn]);
  msg := msg + {$IFDEF FPC}LineEnding{$ELSE}sLineBreak{$ENDIF} + Format('Attempt to access the %d-th element of variable %s', [offset+1, varn]);
  raise Exception.Create(msg);
end;

function s_cmp(a, b: PChar; la, lb: FtnLen): Integer; cdecl; {$IFDEF FPC} alias: '_s_cmp'; {$ENDIF}
var
  aend, bend: PChar;
begin
  aend := a + la;
  bend := b + lb;
  if la <= lb then begin
    while a < aend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while b < bend do
      if b^ <> ' ' then begin
        Result := Ord(' ') - Ord(b^);
        Exit;
      end else
        Inc(b);
  end else begin
    while b < bend do
      if a^ <> b^ then begin
        Result := Ord(a^) - Ord(b^);
        Exit;
      end else begin
        Inc(a);
        Inc(b);
      end;
    while a < aend do
      if a^ <> ' ' then begin
        Result := Ord(a^) - Ord(' ');
        Exit;
      end else
        Inc(a);
  end;
  Result := 0;
end;

procedure s_copy(a, b: PChar; la, lb: FtnLen); cdecl; {$IFDEF UND} alias: '_s_copy'; {$ENDIF}
{ Copy of f2c routine assuming NO_OVERWRITE defined, so lhs must not contain rhs }
begin
  if la <= lb then
    Move(b^, a^, la)
  else begin
    Move(b^, a^, lb);
    FillChar((a+lb)^, la-lb, ' ');
  end;
end;

{$ENDIF}

{$IFDEF PROVIDE_GFORTRAN}
function __powidf2(a:Double; p:Integer): Double; cdecl; {$IFDEF FPC} alias_'__powidf2'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

function _gfortran_pow_r8_i4(a:Double; p:Integer): Double; cdecl;  {$IFDEF FPC} alias_'_gfortran_pow_r8_i4'; {$ENDIF}
begin
  Result := IntPower(a,p);
end;

function _gfortran_pow_i4_i4(base:Integer; exponent:Integer): Integer; cdecl; {$IFDEF FPC} alias_'_gfortran_pow_i4_i4'; {$ENDIF}
begin
  Result := 1;
  if exponent = 0 then
    Exit
  else if exponent < 0 then
    if base = 1 then
      Exit
    else if base = -1 then begin
      if Odd(exponent) then
        Result := -1;
    end else if base = 0 then
      Result := 1 div base { Force divide by zero }
    else
      Result := 0
  else begin
    while exponent > 0 do begin
      while (exponent and 1) = 0 do begin
        exponent := exponent shr 1;
        base := base*base;
      end;
      Dec(exponent);
      Result := Result*base;
    end;
  end;
end;

procedure _gfortran_concat_string(destlen: FtnLen; dest: PChar; len1: FtnLen; s1: PChar; len2: FtnLen; s2: PChar); cdecl; {$IFDEF FPC} alias_'_gfortran_concat_string'; {$ENDIF}
begin
  if len1 <= destlen then
    Move(s1^, dest^, destlen)
  else begin
    Move(s1^, dest^, len1);
    Inc(dest,len1);
    Dec(destlen,len1);
    if len2 <= destlen then
      Move(s2, dest, destlen)
    else begin
      Move(s2^, dest^, len2);
      FillChar(dest^, destlen-len2, ' ');
    end;
  end;
end;

function _gfortran_compare_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar): Integer; cdecl; {$IFDEF FPC} alias_'_gfortran_compare_string'; {$ENDIF}
begin
  Result := s_cmp(a, b, la, lb);
end;

procedure _gfortran_copy_string(la: FtnLen; a: PChar; lb: FtnLen; b: PChar); cdecl;  {$IFDEF FPC} alias_'_gfortran_copy_string'; {$ENDIF}
begin
  s_copy(a, b, la, lb);
end;

function _lroundf(a:Double): Integer; cdecl;
begin
  Result := Round(a);
end;

function _memmove(Dest, Source: PChar; length: Integer):Integer; cdecl;
begin
  Move(Source^, Dest^, length);
end;
{$ENDIF}

{$IFDEF VCL_}
function __ftol(x:Double):Integer; cdecl;
var
  cw: Word;
  r: Int64;
asm
  { Based on Borland C run time library }
        fstcw   cw                // save the control word
        fwait
        mov     ax, cw

        // the last bit in 0c01h disables the invalid fpu op exception
        // if the result overflows 64bits. this is done to remain
        // consistent with casts to 32bit ints which can overflow
        // w/o causing invalid fpu op.
        or      cw, $0C01               // set rounding control to chop
        fldcw   cw
        fistp   r                       // convert to 64-bit integer
        mov     cw, ax
        fldcw   cw                      // restore the control word
        mov     eax, dword ptr r        // return LS 32 bits
        mov     edx, dword ptr [r+4]    //        MS 32 bits
end;
{$ENDIF}

function pprint_(var tag, n: Integer; var d: Double):Integer; cdecl; {$IFDEF FPC} alias_'pprint_'; {$ENDIF}
var
  i: Integer;
  p: PDouble;
begin
  Write(Log,'PPRINT (', tag, ')', n);
  p := @d;
  for i := 0 to n - 1 do begin
    Write(Log, ' ', p^);
    Inc(p);
  end;
  WriteLn(Log);
end;

function errstop_(s: PChar; l: Integer):Integer; cdecl; {$IFDEF FPC} alias_'errstop_'; {$ENDIF}
var
  str: string;
begin
  SetString(str, s, l);
  raise Exception.Create(str);
end;

function errstopf_(fmt: PChar; args: Pointer { ... } ):Integer; cdecl; {$IFDEF FPC} alias_'errstopf_'; {$ENDIF}
var
  w: string;
  p: PChar;
begin
  w := '';
  p := fmt;
  while p^ <> #0 do begin
    case p^ of
     '\': begin Inc(p); w := w + p^; end;
     '#', #0: Break;
     '%': begin
            Inc(p);
            case p^ of
              'd': w := w + IntToStr(PInteger(args)^);
              'g': w := w + FloatToStr(PSingle(args)^);
              'l': begin Inc(p); w := w + FloatToStr(PDouble(args)^); end;
            end;
            Inc(PChar(args),SizeOf(PChar));
          end;
    else
      w := w + p^;
    end;
    Inc(p);
  end;
  raise Exception.Create(w);
  Floor(0.1);
end;

function clog(s: PChar):Integer; cdecl; {$IFDEF FPC} alias_'clog'; {$ENDIF}
begin
  WriteLn(Log,ThreadName,':',s);
end;

function xerror_(MESSG: PChar; NMESSG, NERR, LEVEL: PInteger; len: FtnLen): Integer; cdecl; {$IFDEF FPC} alias_'xerror_'; {$ENDIF}
{ SLATEC error routine }
begin
  raise Exception.CreateFmt('%.*s %d %d', [NMESSG^, MESSG, NERR^, LEVEL^]);
end;

function xerbla_(SRNAME: PChar; var Info: Integer; len: FtnLen): Integer; cdecl; {$IFDEF FPC} alias_'xerbla_'; {$ENDIF}
{ BLAS error routine }
var
  s: string;
begin
  SetString(s, SRNAME, len);
  raise Exception.CreateFmt('Error in %s: Parameter %d had an illegal value', [s, Info]);
end;

function ieeeinfandnan_:Integer; cdecl; {$IFDEF FPC} alias_'ieeeinfandnan_'; {$ENDIF}
begin
  { LAPACK requires exZeroDivide and exInvalidOp to be masked for best operation }
  if
  {$IFDEF CPUPOWERPC}
    True
  {$ENDIF}
  {$IFDEF CPU86}
     ([exZeroDivide,exInvalidOp] * GetExceptionMask) = [exZeroDivide,exInvalidOp]
  {$ENDIF}
  {$IFDEF CPUX86_64}
     ([exZeroDivide,exInvalidOp] * GetExceptionMask) = [exZeroDivide,exInvalidOp]
  {$ENDIF}
  then
    Result := 1
  else
    Result := 0;
end;

function disnan_(var a:Double):Integer; cdecl;  {$IFDEF FPC} alias_'disnan_'; {$ENDIF}
begin
  if IsNan(a) then
    Result := 1
  else
    Result := 0;
end;

function dlamc3_(var a: Double; var b: Double): Double; cdecl; {$IFDEF FPC} alias_'dlamc3_'; {$ENDIF}
begin
  Result := a + b; { This routine must not keep any values in the floating point registers }
end;

var
  TimingStrings: array of AnsiString;

function _tend_(var Id:Integer):Integer; cdecl;
begin
  TEnd(TimingStrings[Id]);
  Result := 0; { Not used }
end;

function _tstart_(var Id:Integer):Integer; cdecl;
begin
  if Id > High(TimingStrings) then
    SetLength(TimingStrings, Id+10);
  if Length(TimingStrings[Id]) = 0 then
    TimingStrings[Id] := Format('Fortran %d', [Id]);
  TStart(TimingStrings[Id]);
  Result := 0; { Not used }
end;

//var
//  i,j: Integer;
//initialization
//  for I := -5 to 10 do
//    for j := -2 to 2 do
//      if (j <> 0) or (i > 0) then
//        WriteLn(j:4, I:4, ' ', gfortran_pow_i4_i4(j,I))
end.
