unit CanCan;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Graphics, {$ELSE} QGraphics, {$ENDIF}
     Types, ExpCan, Classes;

type

  { TCanvasCanvas }

  TCanvasCanvas = class(TExportableCanvas) { A real canvas }
  private
    FCanvas: TCanvas;
  protected
    function GetFont: TFont; override;
  public
    constructor Create(ACanvas: TCanvas);
    function RealCanvas: TCanvas; override;
    function Pen: TPen; override;
    function Brush: TBrush; override;
    function DPI: Integer; override;
    procedure SetPixels(X, Y: Integer; const Value: TColor); override;
    procedure Ellipse(X1, Y1, X2, Y2: Integer); override;
    procedure FillRect(const Rect: TRect); override;
    procedure Polyline(Points: array of TPoint; const nPoints: Integer); override;
    procedure Polygon(Points: array of TPoint; const nPoints: Integer); override;
    procedure RotatedTextOut(x, y: Integer; const str: string; RealTextAngle: Integer; Justify: TAlignment); override;
    procedure StretchDraw(var Rect: TRect; Graphic: TGraphic); override;
    procedure TextOut(X,Y: Integer; const Text: String); override;
    procedure TextRect(ARect: TRect; X, Y: integer; const Text : string); override;
  end;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFDEF LCL_} LCLIntf, LCLType, {$ENDIF}
     Math, SysUtils;

{ TCanvasCanvas }

function TCanvasCanvas.Brush: TBrush;
begin
  Result := FCanvas.Brush;
end;

function TCanvasCanvas.DPI: Integer;
begin
  if FCanvas.HandleAllocated then begin
    Result := (GetDeviceCaps(FCanvas.Handle, LOGPIXELSX)
               +GetDeviceCaps(FCanvas.Handle, LOGPIXELSY)) div 2;
    if Result = 0 then
      Result := 2540; { Assume Metafile }
  end else
    Result := 72;
end;

constructor TCanvasCanvas.Create(ACanvas: TCanvas);
begin
  FCanvas := ACanvas;
end;

procedure TCanvasCanvas.Ellipse(X1, Y1, X2, Y2: Integer);
begin
  FCanvas.Ellipse(X1, Y1, X2, Y2);
end;

procedure TCanvasCanvas.FillRect(const Rect: TRect);
begin
  FCanvas.FillRect(Rect);
end;

function TCanvasCanvas.GetFont: TFont;
begin
  Result := FCanvas.Font;
end;

function TCanvasCanvas.Pen: TPen;
begin
  Result := FCanvas.Pen;
end;

procedure TCanvasCanvas.Polygon(Points: array of TPoint; const nPoints: Integer);
begin
  FCanvas.Polygon({$IFDEF FPC}@Points[0], nPoints, False{$ELSE}Slice(Points,nPoints){$ENDIF});
end;

procedure TCanvasCanvas.Polyline(Points: array of TPoint; const nPoints: Integer);
begin
  FCanvas.Polyline({$IFDEF FPC}Points, 0, nPoints{$ELSE}Slice(Points,nPoints){$ENDIF});
end;

function TCanvasCanvas.RealCanvas: TCanvas;
begin
  Result := FCanvas;
end;

procedure TCanvasCanvas.RotatedTextOut(x, y: Integer; const str: string;
  RealTextAngle: Integer; Justify: TAlignment);
var
  Extent: TSize;
  OldOrient: Integer;
  h: Single;
  c_TextAngle, s_TextAngle: Extended;
begin
    if RealTextAngle <> 0 then
      Exit; { Discard }
  {$IFDEF LCL_}
    OldOrient := FCanvas.Font.Orientation;
    FCanvas.Font.Orientation := 0;
  {$ELSE}
    if RealTextAngle <> 0 then
      Exit; { Discard }
  {$ENDIF}
  Extent := FCanvas.TextExtent(str);
  case Justify of
    taLeftJustify:  h := 0;
    taRightJustify: h := 1;
    taCenter:       h := 0.5;
  end;
  if RealTextAngle = 0 then begin
    if Justify <> taLeftJustify then
      x := x - Round(h*Extent.cx);
    {$IFDEF LCL_}
    Dec(y, Extent.cy); { alignment fix }
    {$ENDIF}
  end else begin
    sincos(RealTextAngle*Pi/1800, s_TextAngle, c_TextAngle);
    x := x - Round(h*Extent.cx*c_TextAngle+Extent.cy*s_TextAngle);
    y := y + Round(h*Extent.cx*s_TextAngle-Extent.cy*c_TextAngle);
  end;
  {$IFDEF LCL_}
    FCanvas.Font.Orientation := RealTextAngle;
  {$ENDIF}
  TextOut(x, y, str);
  {$IFDEF LCL_}
    FCanvas.Font.Orientation := OldOrient;
  {$ENDIF}
end;

procedure TCanvasCanvas.TextOut(X, Y: Integer; const Text: String);
begin
  FCanvas.TextOut(X, Y, Text);
end;

procedure TCanvasCanvas.TextRect(ARect: TRect; X, Y: integer; const Text: string);
begin
  FCanvas.TextRect(ARect, X, Y, Text);
end;

procedure TCanvasCanvas.SetPixels(X, Y: Integer; const Value: TColor);
begin
  FCanvas.Pixels[X,Y] := Value;
end;

procedure TCanvasCanvas.StretchDraw(var Rect: TRect; Graphic: TGraphic);
begin
  FCanvas.StretchDraw(Rect, Graphic);
end;

end.
