unit EditExtr;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} Messages, Controls, Buttons, StdCtrls,
     {$ELSE} QControls, QButtons, QStdCtrls, {$ENDIF}
     {$IFDEF LCL_} LMessages, {$ENDIF}
     Classes, AlEdit;

type
  { Sigle line edit window with an additional control, such as a
    button. The extra control is created with CreateExtraControl.
    If a caption is given, a label is generated for the control }
  TCustomEditExtra = class(TAlignedEdit)
  private
    FCaption: string;
    FEditLabel: TLabel;
    FExtraControl: TControl;
    FLabelSpacing: Integer;
    {$IFDEF VER80}
    SavedParent: TWinControl;
    {$ENDIF}
    {$IFDEF SUBFRAME}
    FEdit: TAlignedEdit;
    function GetAlignment: TAlignment;
    procedure SetAlignment(const Value: TAlignment);
    procedure SetModified(const Value: Boolean);
    function GetModified: Boolean;
    function GetOnChange: TNotifyEvent;
    procedure SetOnChange(const Value: TNotifyEvent);
    function GetOnEnter: TNotifyEvent;
    procedure SetOnEnter(const Value: TNotifyEvent);
    function GetOnExit: TNotifyEvent;
    procedure SetOnExit(const Value: TNotifyEvent);
    function GetText: string;
    function GetAutoSelect: Boolean;
    procedure SetAutoSelect(const Value: Boolean);
    function GetCursor: TCursor;
    procedure SetCursor(const Value: TCursor);
    {$ELSE}
    procedure PositionExtraControl; virtual;
    {$ENDIF}
    procedure SetCaption(const AValue: string);
    procedure SetExtraWidth(const Value: Integer);
    function GetExtraWidth: Integer;
    procedure SetLabelSpacing(const AValue: Integer);
  protected
    {$IFNDEF CLX_}
    procedure CMEnabledchanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CreateHandle; override;
    {$ELSE}
    procedure EnabledChanged; override;
    {$ENDIF}
    {$IFNDEF SUBFRAME}
    procedure SetParent({$IFDEF CLX_} const {$ENDIF} AParent: TWinControl); override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    {$IFNDEF CLX_}
    procedure CMVisiblechanged(var Message: TMessage); message CM_VISIBLECHANGED;
    {$ELSE}
    procedure VisibleChanged; override;
    procedure BoundsChanged; override;
    {$ENDIF}
    {$ENDIF}
    procedure DoEnter; override;
    function CreateExtraControl:TControl; virtual; abstract;
  public
    constructor Create(AOwner: TComponent); override;
    {$IFDEF VER80FUDGE}
    destructor Destroy; override;
    {$ENDIF}
    {$IFNDEF SUBFRAME}
    {$IFNDEF CLX_}
    procedure SetBounds(ALeft: Integer; ATop: Integer; AWidth: Integer; AHeight: Integer); override;
    {$ENDIF}
    {$ENDIF}
    property ExtraControl: TControl read FExtraControl;
    {$IFDEF SUBFRAME}
    property Edit: TAlignedEdit read FEdit;
    property Text: string read GetText;
    property Modified: Boolean read GetModified write SetModified;
    {$ENDIF}
    property EditLabel: TLabel read FEditLabel stored false;
  published
    {$IFDEF SUBFRAME}
    property Align;
    property Anchors;
    property Enabled;
    property TabOrder;
    property Alignment: TAlignment read GetAlignment write SetAlignment default taRightJustify;
    property OnChange: TNotifyEvent read GetOnChange write SetOnChange;
    property OnEnter: TNotifyEvent read GetOnEnter write SetOnEnter;
    property OnExit: TNotifyEvent read GetOnExit write SetOnExit;
    property AutoSelect: Boolean read GetAutoSelect write SetAutoSelect default True;
    property Cursor: TCursor read GetCursor write SetCursor default crDefault;
    {$ENDIF}
    property ExtraWidth: Integer read GetExtraWidth write SetExtraWidth;
    { Set negative to specify the distance of the left edge from the left edge of the edit control;
      Set Positive to specify the distance between the controls }
    property LabelSpacing: Integer read FLabelSpacing write SetLabelSpacing default 3;
    property Caption: string read FCaption write SetCaption; { Hide inherited Caption property }
  end;

  { Edit control with a button }
  TCustomEditButton = class (TCustomEditExtra)
  private
    {$IFDEF VER80}
    FFlat: Boolean; { For compatibility only - no real effect }
    {$ELSE}
    procedure SetFlat(const Value: Boolean);
    function GetFlat: Boolean;
    {$ENDIF}
  protected
    SpeedButton: TSpeedButton;
    function CreateExtraControl:TControl; override;
  published
    property Flat:Boolean {$IFDEF VER80} read FFlat write FFlat
                          {$ELSE} read GetFlat write SetFlat {$ENDIF} default False;
  end;

  { Edit control with a check box }
  TCustomEditCheck = class (TCustomEditExtra)
  private
    FChecked: Boolean;
    function GetChecked: Boolean;
    procedure SetChecked(const AValue: Boolean);
  protected
    CheckBox: TCheckBox;
    function CreateExtraControl:TControl; override;
  published
    property Checked: Boolean read GetChecked write SetChecked default False;
  end;
  
  
implementation

uses SysUtils;

{ TCustomEditExtra }

constructor TCustomEditExtra.Create(AOwner: TComponent);
begin
  FLabelSpacing := 3;
  inherited Create(AOwner);
  {$IFDEF SUBFRAME}
  ControlStyle := ControlStyle - [csFixedHeight, csSetCaption];
  FEdit := TAlignedEdit.Create(Self);
  Edit.Parent := Self;
  Height := Edit.Height;
  Edit.Align := alClient;
  {$ENDIF}
  SetCaption(FCaption); { Force creation of label control if required }
  FExtraControl := CreateExtraControl;
  {$IFDEF VER80FUDGE}
{  if AOwner is TWinControl then
    TWinControl(AOwner).InsertControl(FExtraControl); }
  {$ENDIF}
end;

{$IFDEF VER80FUDGE}
destructor TCustomEditExtra.Destroy;
var
  Control: TControl;
begin
  inherited Destroy;
  { This particularly horrible bit of code works round a bug where a Delphi 1
    can't cope with a component being removed as the form removes another component,
    even if the component is not owned by the form. To fix, just add back another control
    for the parent form to delete instead of the deleted extra control }
  { Not required if TWinControl.Destroy has been patched to cope - fortunately
    Delphi 1 uses patched version }
  if SavedParent <> nil then begin
    Control := TControl.Create(SavedParent);
    Control.ControlState := Control.ControlState + [csReadingState]; { Avoid some processing }
    SavedParent.InsertControl(Control); { Yuck! }
    Control.ControlState := Control.ControlState - [csReadingState];
  end;
end;
{$ENDIF}

{$IFNDEF CLX_}
procedure TCustomEditExtra.CMEnabledchanged(var Message: TMessage);
{$ELSE}
procedure TCustomEditExtra.EnabledChanged;
{$ENDIF}
begin
  inherited;
  if EditLabel <> nil then
    EditLabel.Enabled := Enabled;
  if ExtraControl <> nil then
    ExtraControl.Enabled := {$IFDEF SUBFRAME}Edit.{$ENDIF}Enabled;
end;

{$IFNDEF SUBFRAME}

{$IFDEF CLX_}
procedure TCustomEditExtra.BoundsChanged;
begin
  inherited;
  PositionExtraControl;
end;
{$ENDIF}

{$IFNDEF CLX_}
procedure TCustomEditExtra.SetBounds(ALeft: Integer; ATop: Integer; AWidth: Integer; AHeight: Integer);
begin
  inherited SetBounds(ALeft, ATop, AWidth, AHeight);
  PositionExtraControl;
end;

procedure TCustomEditExtra.CreateHandle;
begin
  inherited;
  PositionExtraControl;
end;
{$ENDIF}

procedure TCustomEditExtra.PositionExtraControl;
var
  ELeft: Integer;
begin
  if HandleAllocated then begin
    if (ExtraControl <> nil) and (ExtraControl.Width > 0) then
      ExtraControl.SetBounds(Left+Width, Top, ExtraControl.Width, Height);
    if EditLabel <> nil then begin
      if LabelSpacing < 0 then
        ELeft := Left + LabelSpacing
      else
        ELeft := Left - EditLabel.Width - LabelSpacing;
      EditLabel.SetBounds(ELeft, Top + (Height - EditLabel.height) div 2, EditLabel.Width, EditLabel.Height);
    end;
  end;
end;

{ May only be necessary to work round LCL bug }
type TOpenLabel = class(TLabel)
  procedure DoAutoSize;
  end;

procedure TOpenLabel.DoAutoSize;
begin
  inherited;
end;

procedure TCustomEditExtra.SetCaption(const AValue: string);
begin
  { Do not test for no change, as we use this procedure to generate the control if necessary }
  FCaption := AValue;
  if (Caption = '') and Assigned(EditLabel) then
    FreeAndNil(FEditLabel);
  if Caption <> '' then begin
    if not Assigned(EditLabel) then begin
      FEditLabel := TLabel.Create(Self);
      EditLabel.FocusControl := Self;
      EditLabel.AutoSize := True;
      if Parent <> nil then begin
        EditLabel.Parent := Parent;
        EditLabel.Visible := Visible;
      end;
    end;
    if Assigned(EditLabel) then begin
      EditLabel.Caption := Caption;
      TOpenLabel(EditLabel).DoAutoSize;
      PositionExtraControl; { Tests for HandleAllocated }
    end;
  end;
end;

procedure TCustomEditExtra.SetParent({$IFDEF CLX_} const {$ENDIF} AParent: TWinControl);
begin
  inherited SetParent(AParent);
  {$IFDEF VER80}
  if AParent <> nil then SavedParent := AParent;
  {$ENDIF}
  if EditLabel <> nil then begin
    EditLabel.Parent := AParent;
    EditLabel.Visible := True;
  end;
  if ExtraControl <> nil then begin
    {.$IFNDEF VER80}
    ExtraControl.Parent := AParent;
    ExtraControl.Visible := True;
    {.$ENDIF}
  end;
  PositionExtraControl; { Just in case }
end;

procedure TCustomEditExtra.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (AComponent = EditLabel) and (Operation = opRemove) then
    FEditLabel := nil;
  if (AComponent = ExtraControl) and (Operation = opRemove) then
    FExtraControl := nil;
end;

{$IFNDEF CLX_}
procedure TCustomEditExtra.CMVisiblechanged(var Message: TMessage);
{$ELSE}
procedure TCustomEditExtra.VisibleChanged;
{$ENDIF}
begin
  inherited;
  if EditLabel <> nil then
    EditLabel.Visible := Visible;
  if ExtraControl <> nil then
    ExtraControl.Visible := Visible;
end;
{$ENDIF}

{$IFDEF SUBFRAME}
function TCustomEditExtra.GetAlignment: TAlignment;
begin
  Result := Edit.Alignment;
end;

procedure TCustomEditExtra.SetAlignment(const Value: TAlignment);
begin
  Edit.Alignment := Value;
end;

function TCustomEditExtra.GetText: string;
begin
  Result := Edit.Text;
end;

procedure TCustomEditExtra.SetModified(const Value: Boolean);
begin
  Edit.Modified := Value;
end;

function TCustomEditExtra.GetModified: Boolean;
begin
  Result := Edit.Modified;
end;

function TCustomEditExtra.GetOnChange: TNotifyEvent;
begin
  Result := Edit.OnChange;
end;

procedure TCustomEditExtra.SetOnChange(const Value: TNotifyEvent);
begin
  Edit.OnChange := Value;
end;

function TCustomEditExtra.GetOnEnter: TNotifyEvent;
begin
  Result := Edit.OnEnter;
end;

procedure TCustomEditExtra.SetOnEnter(const Value: TNotifyEvent);
begin
  Edit.OnEnter := Value;
end;

function TCustomEditExtra.GetOnExit: TNotifyEvent;
begin
  Result := Edit.OnExit;
end;

procedure TCustomEditExtra.SetOnExit(const Value: TNotifyEvent);
begin
  Edit.OnExit := Value;
end;

function TCustomEditExtra.GetAutoSelect: Boolean;
begin
  Result := Edit.AutoSelect;
end;

procedure TCustomEditExtra.SetAutoSelect(const Value: Boolean);
begin
  Edit.AutoSelect := Value;
end;

function TCustomEditExtra.GetCursor: TCursor;
begin
  Result := Edit.Cursor;
end;

procedure TCustomEditExtra.SetCursor(const Value: TCursor);
begin
  Edit.Cursor := Value;
end;
{$ENDIF}

procedure TCustomEditExtra.DoEnter;
begin
  if {$IFNDEF lcl_} AutoSelect and {$ENDIF} not (csLButtonDown in {$IFDEF SUBFRAME}Edit.{$ENDIF}ControlState) then
    {$IFDEF SUBFRAME}Edit.{$ENDIF}SelectAll;
  inherited DoEnter;
end;

procedure TCustomEditExtra.SetExtraWidth(const Value: Integer);
begin
  if ExtraControl <> nil then
    if Value < 1 then
      ExtraControl.Visible := False
      { Note - not honoured at design time }
    else begin
      ExtraControl.Width := Value;
      ExtraControl.Visible := True;
    end;
end;

function TCustomEditExtra.GetExtraWidth: Integer;
begin
  if (ExtraControl <> nil) and ExtraControl.Visible then
    Result := ExtraControl.Width
  else
    Result := 0;
end;

procedure TCustomEditExtra.SetLabelSpacing(const AValue: Integer);
begin
  if FLabelSpacing=AValue then exit;
  FLabelSpacing:=AValue;
  PositionExtraControl;
end;

{ TCustomEditButton }

function TCustomEditButton.CreateExtraControl: TControl;
begin
  SpeedButton := TSpeedButton.Create(Self);
  Result := SpeedButton;
end;

{$IFNDEF VER80}
function TCustomEditButton.GetFlat: Boolean;
begin
  Result := SpeedButton.Flat;
end;

procedure TCustomEditButton.SetFlat(const Value: Boolean);
begin
  SpeedButton.Flat := Value;
end;
{$ENDIF}

{ TCustomEditCheck }

procedure TCustomEditCheck.SetChecked(const AValue: Boolean);
begin
  FChecked := AValue;
  if Assigned(CheckBox) then
    CheckBox.Checked := FChecked;
end;

function TCustomEditCheck.GetChecked: Boolean;
begin
  Result := Assigned(ExtraControl) and CheckBox.Checked
end;

function TCustomEditCheck.CreateExtraControl: TControl;
begin
  CheckBox := TCheckBox.Create(Self);
  CheckBox.Width := CheckBox.Height;
  {$IFDEF LCL_}
  CheckBox.AutoSize := True;
  {$ENDIF}
  CheckBox.Caption := '';
  Result := CheckBox;
end;

end.
