unit FixDlg;

interface
{$I DFS.inc}

uses
  {$IFNDEF CLX_} Messages, Graphics, Controls, Forms, Dialogs, StdCtrls, Buttons,
  {$ELSE} QStdCtrls, QButtons, QControls, QForms, QDialogs, Qt, QTypes, {$ENDIF}
  SysUtils, Classes, FixFnt;

type
  TFixedFontForm = class(TForm)
    Faces: TListBox;
    Label1: TLabel;
    Sizes: TComboBox;
    Label2: TLabel;
    Bold: TCheckBox;
    BitBtn1: TBitBtn;
    BitBtn2: TBitBtn;
    Sample: TMemo;
    Label3: TLabel;
    procedure FormShow(Sender: TObject);
    procedure FacesKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FacesMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure SizesChange(Sender: TObject);
    procedure BoldMouseUp(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure BoldKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure BoldClick(Sender: TObject);
  private
    CurrentIndex: Integer;
    CurrentFace: Integer;
    BestIndex, BestHeight, BestWidth: Integer;
    FFixedFont: TFixedFont;
    procedure SetFixedFont(const Value: TFixedFont); { For Matching previous font }
    procedure DoFaceChange;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Loaded; override;
    procedure CheckSizeChange;
    procedure DoSizeChange;
    procedure CheckFaceChange;
    property FixedFont:TFixedFont read FFixedFont write SetFixedFont;
  end;

var
  FixedFontForm: TFixedFontForm;

implementation

uses {$IFDEF MSWINDOWS} Windows {$ENDIF}
     {$IFDEF LCL_} LResources {$ENDIF}
     {$IFDEF KYLIX} BindHelp {$ENDIF}
     ;

{$IFDEF VCL_}
{$R *.dfm}
{$ENDIF}
{$IFDEF CLX_}
{$R *.xfm}
{$ENDIF}

{$IFDEF MSWINDOWS}
function EnumFacesProc (var lf:TLOGFONT; var tm:TTextMetric;
                        FontType: Cardinal; Form:TFixedFontForm):Integer; {$IFDEF WIN32} stdcall {$ELSE} far {$ENDIF};
var
  Name: string;
begin
  if (tm.tmPitchAndFamily and TMPF_FIXED_PITCH) = 0 then begin
    Name := lf.lfFaceName;
    SetLength(Name, StrLen(lf.lfFaceName));
    if Form.Faces.Items.IndexOf(Name) < 0 then
      Form.Faces.Items.Add(Name);
  end;
  Result := 1;
end;
{$ENDIF}

{$IFDEF CLX_}
{$T+}
var
  FFixedFonts: TStringList;

type
  TFamilyInfo = class(TObject)
  public
    { Styles: TStringList; }
    PointSizes: TIntArray;
    constructor Create({AStyles: TStringList;} APointSizes: TIntArray);
    destructor Destroy; override;
  end;

constructor TFamilyInfo.Create({AStyles: TStringList;} APointSizes: TIntArray);
begin
  inherited Create;
  { Styles := AStyles; }
  PointSizes := APointSizes;
end;

destructor TFamilyInfo.Destroy;
begin
  { FreeAndNil(Styles); }
  inherited;
end;

{$IFDEF DFS_KYLIX_1}
{$DEFINE NOFIXEDPITCH}
{$ENDIF}

{$IFDEF DFS_KYLIX_3}
{$DEFINE NOFIXEDPITCH}
{$ENDIF}

{$IFNDEF NOFIXEDPITCH}

{ TODO: Fix this for Kylix 3 }
{$IFDEF DFS_KYLIX_2_UP}
function QFontDatabase_isFixedPitch(handle: QFontDatabaseH; family: PWideString; style: PWideString; charset: PWideString): Boolean;
 cdecl; external '' name QtNamePrefix + 'QFontDatabase_isFixedPitch';
{$ENDIF}

function FixedFontList: TStringList;
var
  i: Integer;
  FontFamilies: QStringListH;
  FontDatabase: QFontDatabaseH;
  PointSizes: TIntArray;
  FamilyW: WideString;
begin
  if FFixedFonts = nil then begin
    FontFamilies := QStringList_create();
    try
      FontDatabase := QFontDatabase_create();
      try
        QFontDatabase_families(FontDatabase, FontFamilies, True);
        FFixedFonts := QStringListToTStringList(FontFamilies);
        for i := FFixedFonts.Count -1 downto 0 do begin
           QOpenStringList_value(QOpenStringListH(FontFamilies), @FamilyW, i);
           if QFontDatabase_isFixedPitch(FontDatabase, @FamilyW, nil, nil) then begin
             QFontDatabase_pointSizes(FontDatabase, @PointSizes, @FamilyW, nil, nil);
             FFixedFonts.Objects[i] := TFamilyInfo.Create({QStringListToTStringList(StylesQ),} PointSizes)
           end else
             FFixedFonts.Delete(i);
        end;
      finally
        QFontDatabase_destroy(FontDatabase);
      end;
    finally
      QStringList_destroy(FontFamilies);
    end;
  end;
  Result := FFixedFonts;
end;
{$ELSE}
function FixedFontList: TStringList;
var
  i: Integer;
  FontFamilies, StylesQ: QStringListH;
  FontDatabase: QFontDatabaseH;
  Font: QFontH;
  FontInfo: QFontInfoH;
  PointSizes: TIntArray;
  FamilyW, StyleW: WideString;
begin
  if FFixedFonts = nil then begin
    FontFamilies := QStringList_create();
    StylesQ := QStringList_create();
    Font := QFont_create();
    try
      FontDatabase := QFontDatabase_create();
      try
        QFontDatabase_families(FontDatabase, FontFamilies, True);
        FFixedFonts := QStringListToTStringList(FontFamilies);
        for i := FFixedFonts.Count -1 downto 0 do begin
           { Rewrite this when QFontDatabase_isFixedPitch available }
           QOpenStringList_value(QOpenStringListH(FontFamilies), @FamilyW, i);
           QFontDatabase_styles(FontDatabase, StylesQ, @FamilyW, nil);
           if QOpenStringList_count(QOpenStringListH(StylesQ)) > 0 then begin
             QOpenStringList_value(QOpenStringListH(StylesQ), @StyleW, 0);
             QFontDatabase_pointSizes(FontDatabase, @PointSizes, @FamilyW, @StyleW, nil);
             if High(PointSizes) > 0 then begin
               QFontDatabase_font(FontDatabase, Font, @FamilyW, @StyleW, PointSizes[0], nil);
               FontInfo := QFontInfo_create(Font);
               if QFontInfo_fixedPitch(FontInfo) then
                 FFixedFonts.Objects[i] := TFamilyInfo.Create(PointSizes)
               else begin
                 FFixedFonts.Delete(i);
               end;
               QFontInfo_destroy(FontInfo);
             end else
               FFixedFonts.Delete(i);
           end;
        end;
      finally
        QFontDatabase_destroy(FontDatabase);
      end;
    finally
      QStringList_destroy(StylesQ);
      QStringList_destroy(FontFamilies);
      QFont_destroy(Font);
    end;
  end;
  Result := FFixedFonts;
end;
{$ENDIF}
{$ENDIF}

procedure TFixedFontForm.FormShow(Sender: TObject);
{$IFDEF MSWINDOWS}
var
  proc: TFarProc;
{$ENDIF}
begin
  CurrentFace := -1;
  Faces.Clear;
  {$IFDEF MSWINDOWS}
    {$IFDEF FPC}
      EnumFonts(Canvas.Handle, nil, ENUMFONTSPROC(@EnumFacesProc), Pointer(Self));
    {$ELSE}
      proc := MakeProcInstance(@EnumFacesProc, HInstance);
      EnumFonts(Canvas.Handle, nil, proc, Pointer(Self));
      FreeProcInstance(proc);
    {$ENDIF}
  {$ENDIF}
  {$IFDEF CLX_}
  Faces.Items.AddStrings(FixedFontList);
  {$ENDIF}
  CurrentFace := Faces.Items.IndexOf(FixedFont.Name);
  if CurrentFace < 0 then begin
    { Font Not Found }
    CurrentFace := 0;
    FFixedFont.Name := Faces.Items[0];
  end;
  Faces.ItemIndex := CurrentFace;
  DoFaceChange;
end;

procedure AddSize(h, w: Integer; Form:TFixedFontForm);
var
  Name: string;
begin
  if w = 0 then
    Name := IntToStr(h)
  else
    Name := Format('%2d x %2d', [w, h]);
  with Form.Sizes.Items do
    if IndexOf(Name) < 0 then
      Add(Name);
  { Select closest to current font size; look for something smaller if no exact match }
  if h <= Form.FixedFont.Height then
    if (h > Form.BestHeight) or (Form.BestIndex = 0)
        or ((h = Form.BestHeight) and (w <= Form.FixedFont.Width) and (w > Form.BestWidth)) then begin
      Form.BestHeight := h;
      Form.BestWidth := w;
      Form.BestIndex := Form.Sizes.Items.Count - 1;
    end;
end;

procedure AddStandardSizes(Form:TFixedFontForm);
var
  i: Integer;
begin
  { make a size list up }
  for i := 8 to 12 do AddSize(i, 0, Form);
  for i := 7 to 14 do AddSize(2*i, 0, Form);
  AddSize(36, 0, Form);
  AddSize(48, 0, Form);
  AddSize(72, 0, Form);
end;

{$IFDEF MSWINDOWS}
function EnumSizesProc (var lf:TLogFont; var tm:TTextMetric;
                        FontType: Cardinal; Form:TFixedFontForm):Integer; {$IFDEF WIN32} stdcall {$ELSE} far {$ENDIF};
begin
  Result := 1;
  if (lf.lfItalic or lf.lfUnderline or lf.lfStrikeOut) = 0 then
    if (FontType and TRUETYPE_FONTTYPE) <> 0 then begin
      AddStandardSizes(Form);
      Result := 0; { no need to list this one further }
    end else { non-scalable font }
      AddSize (lf.lfHeight, lf.lfWidth, Form);
end;
{$ENDIF}

procedure TFixedFontForm.DoSizeChange;
begin
  CurrentIndex := Sizes.ItemIndex;
  FFixedFont.Bold := Bold.Checked;
  FFixedFont.Size := Sizes.Items[CurrentIndex];
  FFixedFont.CopyTo(Sample.Font);
  Sample.Clear;
  Sample.Lines.Add(FFixedFont.Name
             + ' Sample Aa Bb Cc Dd Ee Ff Gg 0123456789 The Small Brown Fox');
end;

procedure TFixedFontForm.CheckSizeChange;
begin
  if (Sizes.ItemIndex <> CurrentIndex) or (FFixedFont.Bold <> Bold.Checked) then
    DoSizeChange;
end;

procedure TFixedFontForm.CheckFaceChange;
begin
  if CurrentFace <> Faces.ItemIndex then begin
    CurrentFace := Faces.ItemIndex;
    FFixedFont.Name := Faces.Items[CurrentFace];
    DoFaceChange;
  end;
end;

procedure TFixedFontForm.FacesKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  CheckFaceChange;
end;

procedure TFixedFontForm.FacesMouseUp(Sender: TObject;
  Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  CheckFaceChange;
end;

procedure TFixedFontForm.SizesChange(Sender: TObject);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.SetFixedFont(const Value: TFixedFont);
begin
  FFixedFont.Assign(Value);
end;

constructor TFixedFontForm.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFixedFont := TFixedFont.Create;
end;

destructor TFixedFontForm.Destroy;
begin
  FFixedFont.Free;
  inherited Destroy;
end;

procedure TFixedFontForm.Loaded;
begin
  inherited Loaded;
  {$IFNDEF LCL_}
  // Moved from resource file as not present on all systems
  BorderIcons := [biSystemMenu];
  Faces.IntegralHeight := True;
  {$ENDIF}
end;

procedure TFixedFontForm.BoldMouseUp(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.BoldKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
  CheckSizeChange;
end;

procedure TFixedFontForm.DoFaceChange;
var
  {$IFDEF MSWINDOWS}
  proc: TFarProc;
  AName: string;
  {$ENDIF}
  {$IFDEF CLX_}
  FamilyInfo: TFamilyInfo;
  i: Integer;
  {$ENDIF}
begin
  Sizes.Clear;
  CurrentIndex := -1;
  BestIndex := 0;
  BestHeight := 0;
  BestWidth := 0;
  {$IFDEF MSWINDOWS}
    AName := FixedFont.Name + #0;
    {$IFDEF FPC}
      EnumFonts(Canvas.Handle, @AName[1], ENUMFONTSPROC(@EnumSizesProc), Pointer(Self));
    {$ELSE}
      proc := MakeProcInstance(@EnumSizesProc, HInstance);
      EnumFonts(Canvas.Handle, @AName[1], proc, Pointer(Self));
      FreeProcInstance(proc);
    {$ENDIF}
  {$ENDIF}
  {$IFDEF CLX_}
  FamilyInfo := TFamilyInfo(Faces.Items.Objects[CurrentFace]);
  for i := 0 to High(FamilyInfo.PointSizes) do
    AddSize(FamilyInfo.PointSizes[i], 0, self);
  {$ENDIF}
  with Sizes.Items do
    if Count = 0 then { No sizes - ATM can do this }
      AddStandardSizes(Self);
  Sizes.ItemIndex := BestIndex;
  DoSizeChange;
end;

procedure TFixedFontForm.BoldClick(Sender: TObject);
begin
  CheckSizeChange;
end;

{$IFDEF LCL_}
initialization
    {$I FixDlg.lrs}
{$ENDIF}
end.
