unit GMarkDlg;
{$I DFS.inc}{ Copyright (C) 2005,2009 C M Western; see PcontrolsReg.pas for license}
interface

uses {$IFNDEF CLX_} { WinTypes, WinProcs, } Messages, Graphics, Controls, Forms, Dialogs,
     {$ELSE} QGraphics, QControls, QForms, QDialogs, {$ENDIF}
     SysUtils, Classes, GStyle, GraphI, Graphs, GraphMrk, MarkChse;

type
  TGraphMarkDialog = class(TComponent)
  private
    FMarkStyle: TMarkStyle;
    FGraph: TGraphB;
    FTitle: string;
    FMark: Integer;
    FEditAllMarks: Boolean;
    Dialog: TMarkDialog;
    FEnabledItems: TMarkChooseItems;
    procedure ReadFromDialog;
    procedure LoadDialog;
    procedure MarkNumberChange(Sender: TObject);
    procedure SetMarkStyle(Style:TMarkStyle);
    procedure SetEnabledItems(const Value: TMarkChooseItems);
    procedure SetGraph(const Value: TGraphB);
    procedure SetMark(const Value: Integer);
  protected
    { Protected declarations }
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Apply;
    function  Execute:Boolean;
  published
    property Title:string read FTitle write FTitle;
    property Graph:TGraphB read FGraph write SetGraph;
    property Mark:Integer read FMark write SetMark default 0;
    property EditAllMarks: Boolean read FEditAllMarks write FEditAllMarks default False;
    property MarkStyle:TMarkStyle read FMarkStyle write SetMarkStyle;
    property EnabledItems: TMarkChooseItems read FEnabledItems write SetEnabledItems;
  end;

implementation

constructor TGraphMarkDialog.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FMarkStyle := TMarkStyle.Create;
  FEnabledItems := [pcColour, pcSize, pcType];
end;

destructor TGraphMarkDialog.Destroy;
begin
  FMarkStyle.Free;
  FMarkStyle := nil;
  inherited Destroy;
end;

procedure TGraphMarkDialog.SetMarkStyle(Style:TMarkStyle);
begin
  FMarkStyle.Assign(Style);
  if Assigned(Dialog) then
    LoadDialog;
end;

procedure TGraphMarkDialog.LoadDialog;
begin
  Dialog.MarkChoose.MarkStyle := MarkStyle;
  Dialog.MarkChoose.EnabledItems := EnabledItems;
end;

procedure TGraphMarkDialog.ReadFromDialog;
begin
  MarkStyle := Dialog.MarkChoose.MarkStyle;
  if EditAllMarks then
    Mark := Dialog.Mark.Value;
end;

function TGraphMarkDialog.Execute:Boolean;
var
  GEnabledItems: TMarkChooseItems;
begin
  Dialog := TMarkDialog.Create(Application);
  try
    if Title <> '' then Dialog.Caption := Title;
    if not Assigned(FGraph) or (Mark < 0) or (Mark >= MaxPens) then begin
      Dialog.MarkLabel.Visible := False;
      Dialog.Mark.Visible := False;
    end else begin
      MarkStyle := Graph.MarkStyles[Mark];
      GEnabledItems := [pcSize, pcType];
      if grColouredPens in Graph.GOptions then
        GEnabledItems := GEnabledItems + [pcColour];
      EnabledItems := GEnabledItems;
      Dialog.Mark.MaxValue := MaxPens-1;
      Dialog.Mark.Value := Mark;
      if not EditAllMarks then begin
        Dialog.Mark.Visible := False;
        Dialog.MarkNumber.Visible := True;
        Dialog.MarkNumber.Caption := IntToStr(Mark);
      end;
      Dialog.DoApply := {$IFDEF FPC}@{$ENDIF}Apply;
    end;
    LoadDialog;
    Dialog.Mark.OnChange := {$IFDEF FPC}@{$ENDIF}MarkNumberChange;
    Result := Dialog.ShowModal = mrOK;
    if Result then
      ReadFromDialog;
  finally
    Dialog.Free;
    Dialog := nil;
  end;
end;

procedure TGraphMarkDialog.Apply;
begin
  if Assigned(FGraph) and (Mark >= 0) and (Mark < MaxPens) then begin
    if Assigned(Dialog) then
      ReadFromDialog;
    FGraph.Markstyles[Mark] := MarkStyle;
  end;
end;

procedure TGraphMarkDialog.MarkNumberChange(Sender: TObject);
var
  Old: Integer;
begin
  if Mark <> Dialog.Mark.Value then begin
    Old := Mark;
    ReadFromDialog;
    FGraph.Markstyles[Old] := MarkStyle;
    MarkStyle := FGraph.Markstyles[Mark]
  end;
end;

procedure TGraphMarkDialog.SetEnabledItems(const Value: TMarkChooseItems);
begin
  FEnabledItems := Value;
  if Assigned(Dialog) then
    LoadDialog;
end;

procedure TGraphMarkDialog.SetGraph(const Value: TGraphB);
begin
  FGraph := Value;
  if (csDesigning in ComponentState) and (Value <> nil)
     and (Mark >= 0) and (Mark < MaxPens) then
    MarkStyle := FGraph.Markstyles[Mark]; { Update Mark from graph }
end;

procedure TGraphMarkDialog.SetMark(const Value: Integer);
begin
  FMark := Value;
  SetGraph(FGraph); { Update Mark from graph }
end;

end.

