unit GridEx;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface

uses
  Classes, SysUtils, Controls, StdCtrls, Grids;

type

  {$IFNDEF LCL_}
  TSelectEditorEvent = procedure(Sender: TObject; Col,Row: Integer; var Editor: TWinControl) of object;
  {$ENDIF}
  {$IFDEF LCL_}
  TEditStyle =  (esSimple, esEllipsis, esPickList);
  TOnGetPickListItems = procedure (ACol, ARow: Integer; AItems: TStrings) of object;
  {$ENDIF}

  TGetEditStyleEvent = procedure (Sender: TObject; ACol, ARow: Integer; var Style: TEditStyle) of object;


  TStringGridEx = class(TStringGrid)
  private
    FDropDownEditor: TComboBox;
    FOnGetEditStyle: TGetEditStyleEvent;
    FOnGetPickListItems: TOnGetPickListItems;
    FOnEditButtonClick: TNotifyEvent;
  protected
    {$IFDEF LCL_}
    procedure SelectEditor; override;
    function CreateDropDownEditor: TComboBox;
    procedure EditordoSetValue; override;
    {$ENDIF}
    {$IFDEF VCL_}
    function CreateEditor: TInplaceEdit; override;
    function GetEditStyle(ACol, ARow: Longint): TEditStyle; override;
    {$ENDIF}
    procedure DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
  public
    destructor Destroy; override;
    property DropDownEditor: TComboBox read FDropDownEditor;
    procedure OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint);
    procedure SaveAreaToStream(Stream:TStream; Area:TGridRect; WithLabels:Boolean = False);
    procedure AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean; IncludeContents: Boolean = False);
  published
    procedure ClearSelection;
    procedure CopyToClipboard(WithLabels:Boolean = False);
    procedure PasteFromClipboard;
    function CanCopyToClipboard: Boolean;
    function CanPasteFromClipboard: Boolean;

    property OnGetEditStyle: TGetEditStyleEvent read FOnGetEditStyle write FOnGetEditStyle;
    property OnGetPickListItems: TOnGetPickListItems read FOnGetPickListItems write FOnGetPickListItems;
    property OnEditButtonClick: TNotifyEvent read FOnEditButtonClick write FOnEditButtonClick;
  end;

implementation

uses {$IFDEF LCL_} LCLType, {$ENDIF}
     {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     Math, Clipbrd, TextStrm, Huge, WCompat;

{$IFDEF LCL_}

type

  { TStringGridCellComboBox }

  TStringGridCellComboBox = class(TComboBox)
  private
    FGrid: TCustomGrid;
  protected
    procedure Change; override;
    procedure KeyDown(var Key : Word; Shift : TShiftState); override;
    procedure msg_SetValue(var Msg: TGridMessage); message GM_SETVALUE;
    procedure msg_GetValue(var Msg: TGridMessage); message GM_GETVALUE;
    procedure msg_SetGrid(var Msg: TGridMessage); message GM_SETGRID;
  public
    procedure VisibleChanged; override;
  end;

{ TStringGridCellComboBox }

type
  TOpenCustomGrid = class(TCustomGrid)
  end;
  
procedure TStringGridCellComboBox.Change;
begin
  inherited Change;
  if (FGrid <> nil) and FGrid.Visible and (ItemIndex >= 0) then
     TOpenCustomGrid(FGrid).SetEditText(TOpenCustomGrid(FGrid).Col, TOpenCustomGrid(FGrid).Row,
                                        Items[ItemIndex]);
end;

procedure TStringGridCellComboBox.KeyDown(var Key: Word; Shift: TShiftState);
begin
  if (FGrid<>nil) and not (Key in [VK_UP, VK_DOWN]) then
    Fgrid.EditorKeyDown(Self, Key, Shift);
  inherited KeyDown(Key, Shift);
end;

procedure TStringGridCellComboBox.msg_SetValue(var Msg: TGridMessage);
var
  i: Integer;
begin
  for i := 0 to Items.Count-1 do
    if CompareText(Msg.Value, Items[i]) = 0 then begin
      ItemIndex := i;
      Exit;
    end;
  ItemIndex := 0;
end;

procedure TStringGridCellComboBox.msg_GetValue(var Msg: TGridMessage);
begin
  if ItemIndex >= 0 then
    Msg.Value := Items[ItemIndex];
end;

procedure TStringGridCellComboBox.msg_SetGrid(var Msg: TGridMessage);
begin
  FGrid:=Msg.Grid;
  Msg.Options:=EO_AUTOSIZE or EO_SELECTALL;
end;

procedure TStringGridCellComboBox.VisibleChanged;
begin
  inherited VisibleChanged;
  { Ensure drop down arrow is visible }
  if Assigned(FGrid) and (FGrid.Width < Left + Width) then
    Width := FGrid.Width - Left;
end;

{$ENDIF}

{ TStringGridEx }

{$IFDEF LCL_}
procedure TStringGridEx.SelectEditor;
var
  aEditor: TWinControl;
  Style: TEditStyle;
begin
  Style := esSimple;
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Style);
  if Style = esSimple then
    inherited
  else begin
    if not Assigned(DropDownEditor) then begin
      FDropDownEditor := CreateDropDownEditor;
      FDropDownEditor.Visible := False;
    end;
    aEditor:= DropDownEditor;
    if (goEditing in Options) and Assigned(OnSelectEditor) then
      OnSelectEditor(Self, Col, Row, aEditor);
    if aEditor<>Editor then Editor:=aEditor;
 end;
end;

function TStringGridEx.CreateDropDownEditor: TComboBox;
begin
  Result := TStringGridCellComboBox.Create(Self);
  Result.Style := csDropDownList;
  Result.Parent := Self.Parent;
end;

procedure TStringGridEx.EditordoSetValue;
var
  Msg: TGridMessage;
  List: TStringList;
  i: Integer;
begin
  if Assigned(Editor) and (Editor = DropDownEditor) then begin
    Msg.LclMsg.Msg:=GM_SETGRID; { Lockout Changes while list being made up }
    Msg.Grid:=nil;
    Editor.Dispatch(Msg);
    List := TStringList.Create;
    try
      { To avoid crashes, only write to dropdown list if it has changed }
      DoGetPicklistItems(Col, Row, List);
      if DropDownEditor.Items.Count <> List.Count then
        DropDownEditor.Items.Assign(List)
      else
        for i := 0 to List.Count-1 do
          DropDownEditor.Items[i] := List[i]
    finally
      List.Free;
    end;
    Msg.Grid:=Self;
    Editor.Dispatch(Msg);
  end;
  inherited EditordoSetValue;
end;
{$ENDIF}

procedure TStringGridEx.AutoAdjustColumn(aCol: Integer; GrowOnly: Boolean; IncludeContents: Boolean);
var
  W, r: Integer;
  s: string;
begin
  r := 0;
  W := 0;
  { Autosize on column and row headings only }
  while (r < RowCount) and ( IncludeContents or (r < FixedRows) or (aCol < FixedCols) ) do begin
    s := Cells[aCol, r];
    W := Max(Canvas.TextWidth(s), W);
    Inc(r);
  end;
  if W = 0 then
    W := DefaultColWidth
  else
    Inc(W, 6);
  if not GrowOnly or (ColWidths[aCol] < W) then
    ColWidths[aCol] := W;
end;

function TStringGridEx.CanCopyToClipboard: Boolean;
begin
  if EditorMode then
    Result := (InplaceEditor <> nil) and (InplaceEditor is TCustomEdit) and (TCustomEdit(InplaceEditor).SelLength > 0)
  else
    Result := True; { Always one cell selected }
end;

function TStringGridEx.CanPasteFromClipboard: Boolean;
begin
  Result := (goEditing in Options) and ClipboardHasText;
end;

procedure TStringGridEx.ClearSelection;
var
  r,c:Longint;
begin
  if not (goEditing in Options) then Exit;
  { Simulate DEL key }
  if EditorMode and (InplaceEditor is TCustomEdit) then
    with InplaceEditor as TCustomEdit do begin
      if SelLength <= 0 then
        SelLength := 1;
      ClearSelection;
    end
  else
    for r := Selection.Top to Selection.Bottom do
      for c := Selection.Left to Selection.Right do
        SetEditText(c,r,'');
end;

procedure TStringGridEx.CopyToClipboard(WithLabels: Boolean);
var
  s:THugeMemoryStream;
  ch:Char;
begin
  if EditorMode and InplaceEditor.Visible and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).CopyToClipboard
  else begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      SaveAreaToStream(s, Selection, WithLabels);
      ch := #0;
      s.Write(ch, 1);
      s.CopyToClipboard(CF_TEXT);
    finally
      s.Free;
    end;
  end;
end;

destructor TStringGridEx.Destroy;
begin
  if Assigned(DropDownEditor) then
    FreeAndNil(FDropDownEditor);
  inherited Destroy;
end;

{$IFDEF VCL_}
function TStringGridEx.CreateEditor: TInplaceEdit;
begin
  Result := TInplaceEditList.Create(Self);
  TInplaceEditList(Result).OnGetPickListitems := DoGetPicklistItems;
  TInplaceEditList(Result).OnEditButtonClick := FOnEditButtonClick;
end;

function TStringGridEx.GetEditStyle(ACol, ARow: Integer): TEditStyle;
begin
  Result := inherited GetEditStyle(ACol, ARow);
  if Assigned(OnGetEditStyle) then
    OnGetEditStyle(Self, Col, Row, Result);
end;
{$ENDIF}

procedure TStringGridEx.DoGetPicklistItems(ACol, ARow: Integer; AItems: TStrings);
begin
  if Assigned (OnGetPickListItems) then
    OnGetPickListItems(Col, Row, AItems);
end;

procedure TStringGridEx.OffsetLoadFromStream(Stream:TStream; StartRow, StartCol:Longint);
var
  s:TTextStream;
  r,c:Longint;
begin
  s := nil;
  try
    // BeginUpdate;
    s := TTextStream.Create(Stream);
    s.NULLisEOF := True;
    r := StartRow;
    repeat
      for c := StartCol to ColCount-1 do
        if s.GetFreeItem then
          SetEditText(c,r,s.Token);
      Inc(r);
    until s.NextLine;;
  finally
    // EndUpdate;
    s.Free;
  end;
end;

procedure TStringGridEx.PasteFromClipboard;
var
  s: THugeMemoryStream;
  r, c: Longint;
  function MoreThanOneItem: Boolean;
  var
    p: PChar;
  begin
    p := s.Memory;
    Result := (StrPos(p, #9) <> nil) or (StrPos(p, #10) <> nil) or (StrPos(p, #13) <> nil);
  end;
begin
  if ClipboardHasText then begin
    s := nil;
    try
      s := THugeMemoryStream.Create;
      s.SetFromClipboard(CF_TEXT);
      { As a heuristic, assume anything with more than one line, or a tab
        should be pasted into grid as a whole }
      if not EditorMode or MoreThanOneItem then begin
        EditorMode := False;
        r := Min(Selection.Top, Selection.Bottom);
        c := Min(Selection.Left, Selection.Right);
        OffsetLoadFromStream(s, r, c);
        Exit;
      end;
    finally
      s.Free;
    end;
  end;
  if EditorMode and (InplaceEditor is TCustomEdit) then
    TCustomEdit(InplaceEditor).PasteFromClipboard
end;

procedure TStringGridEx.SaveAreaToStream(Stream: TStream; Area: TGridRect;
  WithLabels: Boolean);
var
  s:TTextStream;
  r,c:Longint;
begin
  if (Area.Left <= FixedCols) and WithLabels then
    Area.Left := 0;
  if (Area.Top <= FixedRows) and WithLabels then
    Area.Top := 0;
  s := nil;
  try
    s := TTextStream.Create(Stream);
    for r := Area.Top to Area.Bottom do begin
      for c := Area.Left to Area.Right do begin
        s.PutString(Cells[c,r]);
        if c <> Area.Right then s.Tab;
      end;
      if Area.Top <> Area.Bottom then
        s.Newline;
    end;
  finally
    s.Free;
  end;
end;

end.

