unit SlaveEdit;
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
interface
{$IFDEF XBASED} uses X; {$ENDIF}


type
  TWindowMessage = {$IFDEF XBASED} TAtom {$ELSE} Cardinal {$ENDIF};

procedure LaunchEditor(FileName: string; LineNumber: Integer = -1);
function SaveAllWindows: Boolean;
procedure BroadcastNotification(WindowMessage: TWindowMessage);
function BroadcastName(const Name: string):TWindowMessage;
function RegisterMessage(const Name: string):TWindowMessage;

implementation

uses {$IFDEF MSWINDOWS} ShellAPI, Windows, {$ENDIF}
     {$IFDEF XBASED}
       Xlib, Xatom,
      {$IFDEF LCLgtk} GTKProc, {$ENDIF}
      {$IFDEF LCLgtk2} Gtk2Proc, {$ENDIF}
     {$ENDIF}
     CompatMsg, UCompat, Controls, Dialogs, SysUtils, StrUtils, Forms, LongName, Detach;

function FindSlave:String;
var
  s: string;
begin
  Result := 'slave'{$IFDEF MSWINDOWS} + '.exe'{$ENDIF};
  if not FileExists(Result) then begin
    s := ExtractFilePath(Application.ExeName);
    if FileExists(s + Result) then
      Result := s + Result
    else begin
      s := s + '..' + PathDelim + 'slave' + PathDelim;
      if FileExists(s + Result) then
        Result := s + Result
    end;
  end;
end;

type

  TWindowEnumerator = class
  protected
    ResultHWND: TWnd;
    function CallBack(hWnd: TWnd): Boolean; overload; virtual;
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; overload; virtual; abstract;
  public
    function DoEnumerate: TWnd;
  end;

  TWindowFinder = class(TWindowEnumerator)
  private
    FText: string;
  protected
    function CallBack(hWnd: TWnd; WindowName: string): Boolean; override;
  public
    function FindWindow(Text: string): TWnd;
  end;

  TSaveAllWindows = class(TWindowEnumerator)
  public
    First: Boolean;
  protected
    function CallBack(hWnd: TWnd): Boolean; override;
    {$IFDEF XBASED}
    function StartSave: TWnd;
    function FinishSave: TWnd;
    {$ENDIF}
  end;

procedure BringWindowToFront(hWnd: TWnd);
begin
  {$IFDEF MSWINDOWS}
    {$IFDEF VER80}
      SetActiveWindow(hWnd);
    {$ELSE}
      SetForegroundWindow(hWnd);
    {$ENDIF}
    ShowWindow(hWnd, SW_RESTORE);
  {$ENDIF}
  {$IFDEF XBASED}
    XRaiseWindow(X11Display, hWnd);
    XSetInputFocus(X11Display, hWnd, RevertToNone, CurrentTime);
    XSync(X11Display, False);
  {$ENDIF}
end;

{$IFDEF XBASED}
procedure PostMessage(Window: TWnd; Msg: TWindowMessage; wParam, lParam: Integer);
var
  Event: TXClientMessageEvent;
begin
  Event.display := X11Display;
  Event._type := ClientMessage;
  Event.message_type := Msg;
  Event.format := 32;
  Event.data.l[0] := wParam;
  Event.data.l[1] := lParam;
  Event.window := Window;
  XSendEvent(X11Display, Window, False, NoEventMask, Pointer(@Event));
end;
{$ENDIF}

{$IFDEF LCLcarbon}
procedure PostMessage(Window: TWnd; Msg: TWindowMessage; wParam, lParam: Integer);
begin
  { TODO -cLCLcarbon : Implement }
end;
{$ENDIF}

var
  WM_GOTO_LINE: TWindowMessage = 0;

procedure LaunchEditor(FileName: string; LineNumber: Integer);
var
  Window: TWnd;
  Handle: Integer;
  procedure SendLineNumber;
  begin
    if WM_GOTO_LINE = 0 then
      WM_GOTO_LINE := RegisterMessage('WM_GOTO_LINE');
    PostMessage(Window, WM_GOTO_LINE, 0, LineNumber);
  end;
  function FindWindow:Boolean;
  begin
    with TWindowFinder.Create do begin
      Window := FindWindow(ExtractFileName(FileName));
      Free;
    end;
    Result := Window <> TWnd(0);
  end;
begin
  if FileName = '' then begin
    ShowMessage('Please choose a name for the file before editing it');
    Exit;
  end;
  DoDirSeparators(FileName);
  if FindWindow then begin
    BringWindowToFront(Window);
    if LineNumber > 0 then
      SendLineNumber;
    Exit;
  end;
  if not FileExists(FileName) then
    if MessageDlg(Format('File %s does not exist. Create it?', [FileName]),
                  mtConfirmation, [mbYes, mbNo], 0) = mrYes then begin
      Handle := FileCreate(FileName);
      if Handle >= 0 then
        FileClose(Handle)
      else
        RaiseLastOSError;
    end else
      Exit;
  {$IFDEF MSWINDOWS}
    { Try default application associated with file }
    if ShellExecute(0, nil, PChar(ExtractFileName(FileName)), nil, PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
      { else try slave }
      if ShellExecute(0, nil, PChar(FindSlave), PChar(ExtractFileName(FileName)), PChar(ExtractFilePath(FileName)), SW_SHOW) <= 32 then
        { use notepad as a last resort }
        if WinExec(PChar('notepad ' + FileName), SW_SHOW) < 32 then
          RaiseLastOSError;
  {$ENDIF}
  {$IFDEF UNIX}
  if not Detach1(FindSlave, FileName) then
    if not Detach1('gedit', FileName) then
      if not Detach1('kedit', FileName) then
        if not Detach1('xedit', FileName) then
          RaiseLastOSError;
  {$ENDIF}
  if LineNumber > 0 then begin
    Sleep(0); { Hang about waiting for the window to turn up }
    if not FindWindow then
      Sleep(1000);
    if not FindWindow then
      Sleep(1000);
    if FindWindow then
      SendLineNumber;
  end;
end;

{ TWindowEnumerator }

{$IFDEF MSWINDOWS}
function EnumWindowsProc(hWnd: TWnd; lParam: LPARAM): BOOL; export; stdcall;
begin
  Result := TWindowEnumerator(lParam).CallBack(hWnd);
end;
{$ENDIF}

{$IFDEF XBASED}
{ Duplicate logic of XmuClientWindow }
function ClientWindow(Window: TWindow): TWindow;
var
  WM_STATE: TAtom;
  function GetProperty(AWindow: TWindow): Boolean;
  var
    AType: TAtom;
    Format: Integer;
    nItems, After: DWord;
    data: PChar;
  begin
    AType := None;
    XGetWindowProperty(X11Display, AWindow, WM_STATE, 0, 0, Ord(False), AnyPropertyType,
                       @AType, @Format, @nItems, @After, Pointer(@data));
    Result := Atype <> 0;
  end;
  function SearchChildren(AWindow: TWindow): TWindow;
  var
    Root, Parent: TWindow;
    Children: PWindow;
    nChildren: Cardinal;
    i: Integer;
  begin
    Result := 0;
    if (XQueryTree(X11Display, AWindow, @Root, @Parent, @Children, @nChildren) = 0) or not Assigned(Children) then
      Exit;
    try
      for i := 1 to nChildren do
        if GetProperty(Children[i-1]) then begin
          Result := Children[i-1];
          Exit;
        end;
      for i := 1 to nChildren do begin
        Result := SearchChildren(Children[i-1]);
        if Result <> 0 then
          Exit;
        end;
     finally
       XFree(children);
     end;
  end;
begin
  Result := Window; { Default is current window }
  WM_STATE := XInternAtom(X11Display, 'WM_STATE', True);
  if WM_STATE = 0 then Exit;
  if GetProperty(Window) then Exit; { On current Window }
  Result := SearchChildren(Window);
  if Result = 0 then
    Result := Window;
end;
{$ENDIF}

function TWindowEnumerator.CallBack(hWnd: TWnd): Boolean;
var
  buffer: string;
  {$IFDEF XBASED}
  Name: PChar;
  {$ELSE}
  n: Integer;
  {$ENDIF}
begin
  Result := True; { Continue enumeration by default }
  {$IFDEF MSWINDOWS}
    if not IsWindowVisible(hWnd) then Exit;
    n := GetWindowTextLength(hWnd);
    if n <= 0 then Exit;
    SetLength(buffer, n);
    n := GetWindowText(hWnd, PChar(buffer), Length(buffer)+1);
    if n <= 0 then Exit;
    SetLength(buffer, n);
  {$ENDIF}
  {$IFDEF XBASED}
  if (XFetchName(X11Display, hWnd, @Name) <> 0) and Assigned(Name) then begin
    buffer := Name;
    XFree(Name);
  end else
    Exit; { No name - skip }
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
  Result := CallBack(hWnd, buffer);
end;

function TWindowEnumerator.DoEnumerate: TWnd;
{$IFDEF VCL_}
var
  EnumProc: Pointer;
{$ENDIF}
{$IFDEF XBASED}
var
  Root, Parent: TWindow;
  Children: PWindow;
  nChildren: Cardinal;
  i: Integer;
{$ENDIF}
begin
  ResultHWND := TWnd(0);
  {$IFDEF MSWINDOWS}
    {$IFDEF LCL_}
      EnumWindows(@EnumWindowsProc, LPARAM(self));
    {$ELSE}
      EnumProc := MakeProcInstance(@EnumWindowsProc, Application.Handle);
      try
        EnumWindows(EnumProc, LPARAM(self));
      finally
        FreeProcInstance(EnumProc);
      end;
    {$ENDIF}
  {$ENDIF}
  {$IFDEF XBASED}
    // XSync(X11Display, False);
    if (XQueryTree(X11Display, RootWindow, @Root, @Parent, @Children, @nChildren) <> 0) and Assigned(Children) then
      try
        for i := 1 to nChildren do
          if not CallBack(ClientWindow(Children[i-1])) then
            Break;
      finally
        XFree(Children);
      end;
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
  Result := ResultHWND;
end;

{ TWindowFinder }

function TWindowFinder.CallBack(hWnd: TWnd; WindowName: string): Boolean;
var
  p: Integer;
  Tail: string;
begin
  p := Pos(FText, AnsiLowerCase(WindowName));
  Result := True;
  if p = 0 then Exit;
  if p < Length(WindowName) - Length(FText) + 1 then begin
    { Not at the end of the string - look for "file - application name" }
    Tail := Copy(WindowName, p+Length(FText), 3);
    if Tail <> ' - ' then
      Exit;
  end;
  ResultHWND := hWnd;
  Result := False;
end;

function TWindowFinder.FindWindow(Text: string): TWnd;
begin
  FText := AnsiLowerCase(Text);
  Result := DoEnumerate;
end;

{ TSaveAllWindows }

var
  WM_Save_All: TWindowMessage = 0;

function TSaveAllWindows.CallBack(hWnd: TWnd): Boolean;
var
{$IFDEF MSWINDOWS}
  r: DWORD;
{$ENDIF}
{$IFDEF XBASED}
  Prop: Integer;
  function GetProperty: Integer;
  var
    nItems, BytesLeft: Cardinal;
    ActualType: TAtom;
    ActualFormat: Integer;
    Items: PInteger;
  begin
    Result := 0;
    if (XGetWindowProperty(X11Display, hWnd, WM_Save_All, 0, 1, Ord(False), XA_INTEGER,
                           @ActualType, @ActualFormat, @nItems, @BytesLeft, Pointer(@Items)) = Success)
       and Assigned(Items) then begin
      Result := Items^;
      XFree(Items);
    end;
  end;
{$ENDIF}
{$IFDEF LCLcarbon}
  i: Integer;
{$ENDIF}
begin
  Result := True;
  {$IFDEF MSWINDOWS}
    {$IFDEF OLDER}
    r := SendMessage(hWnd, WM_Save_All, 0, 0);
    {$ELSE}
    if SendMessageTimeout(hWnd, WM_Save_All, 0, 0, SMTO_ABORTIFHUNG or SMTO_BLOCK, 1000, r) = 0 then
      r := 0; { Call failed or timed out }
    {$ENDIF}
    if r = 272727 then begin
      ResultHWND := hWnd;
      Result := False;
    end;
  {$ENDIF}
  {$IFDEF XBASED}
    Prop := GetProperty;
    { Does window have WM_Save_All property set? }
    if Prop <> 0 then begin
      ResultHWND := hWnd;
      if First then
        PostMessage(hWnd, WM_Save_All, MainWindow, 0)
      else
        Result := False;
    end;
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
end;

{$IFDEF XBASED}
function TSaveAllWindows.StartSave: TWnd;
begin
  First := True;
  Result := DoEnumerate;
  XFlush(X11Display);
end;

function TSaveAllWindows.FinishSave: TWnd;
var
  Start: TDateTime;
begin
  First := False;
  Start := Now;
  repeat
    XFlush(X11Display);
    Application.ProcessMessages;
    Result := DoEnumerate;
  until (Result = 0) or (Now - Start > 2/SecsPerDay);
end;
{$ENDIF}

function SaveAllWindows: Boolean;
var
  Window: TWnd;
begin
  Result := True; { Assume OK }
  if WM_Save_All = 0 then
    WM_Save_All := RegisterMessage('WM_Save_All');
  with TSaveAllWindows.Create do
    try
      {$IFDEF XBASED}
      Window := StartSave;
      if Window <> 0 then
        Window := FinishSave;
      {$ENDIF}
      {$IFDEF MSWINDOWS}
      Window := DoEnumerate;
      {$ENDIF}
      { TODO -cLCLcarbon : Implement }
      if Window <> TWnd(0) then begin
        BringWindowToFront(Window);
        Result := False;
      end;
    finally
      Free;
    end;
end;

procedure BroadcastNotification(WindowMessage: TWindowMessage);
begin
  {$IFDEF MSWINDOWS}
  PostMessage(HWND_BROADCAST, WindowMessage, 0, 0);
  {$ENDIF}
  { TODO: UNIX }
  { TODO -cLCLcarbon : Implement }
end;

function BroadcastName(const Name: string):TWindowMessage;
begin
  Result := RegisterMessage(Name);
  BroadcastNotification(Result);
end;

function RegisterMessage(const Name: string):TWindowMessage;
begin
  Result :=
  {$IFDEF MSWINDOWS}
    RegisterWindowMessage(PChar(Name));
  {$ENDIF}
  {$IFDEF XBASED}
    XInternAtom(X11Display, PChar(Name), False);
  {$ENDIF}
  {$IFDEF LCLcarbon}
    0;
  {$ENDIF}
  { TODO -cLCLcarbon : Implement }
end;

end.
