unit SrchStrs;

interface

{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF}
     {$IFNDEF CLX_}  Dialogs, {$ELSE} QDialogs, Types, {$ENDIF}
     Classes;

function SearchStrings(
  Lines: TStrings;      { The Strings to search }
  var Pos: TPoint;      { Character position (char, line) to start at;
                          both zero based;
                          Updated to new position on exit }
  Options:TFindOptions; { Uses: frDown frMatchCase frWholeWord }
  Target: string;       { String to look for - may contain newlines }
  var FoundRect:TRect   { Start and end+1 of matched string }
  ): Boolean;           { True if a match found }

implementation

uses SysUtils, CharType;

type
  TDelta = record
    Jump: Integer; { Next Offset }
    Patlen: Integer;
    Delta: array[0..255] of Integer; { Jump info for each char }
    Patrn: string; { String to scan for }
  end;

function MakeDelta(const Target: string; MatchCase: Boolean): TDelta;
var
  j, JumpBy: Integer;
  ch, ch1: Char;
begin
  with Result do begin
    if MatchCase then
      Patrn := Target
    else
      Patrn := AnsiLowerCase(Target);
    JumpBy := Length(Target);
    for j := 0 to High(Delta) do
      Delta[j] := JumpBy;
    Dec(JumpBy);
    Patlen := JumpBy;

    { Now put in the characters contained in the pattern, duplicating the CASE. }
    for j := 0 to JumpBy do begin
      ch := Patrn[j+1];
      { The last character will have the pattern length, unless there
        are duplicates of it. Get the number to jump from the delta array. }
      if j = JumpBy then
        Jump := JumpBy + Delta[Ord(ch)];
      if not MatchCase and IsCharAlpha(ch) then begin
        ch1 := UpperChar(ch);
        Delta[Ord(ch1)] := JumpBy - j;
      end;
      Delta[Ord(ch)] := JumpBy - j;
    end;
  end;
end;

{ NextCh -- retrieve the next/previous character in the buffer,
   and advance/retreat the point. The order in which this is done
   is significant, and depends upon the direction of the search.
   Forward searches look at the current character and move, reverse
   searches move and look at the character. Fakes newlines off end of buffer }
function NextCh(Lines: TStrings; var Pos: TPoint; Forwards: Boolean): Char;
var
  s: string;
begin
  if Forwards then begin
    s := Lines[Pos.y];
    if (Pos.y >= Lines.Count) or (Pos.x = Length(s)) then begin
      Pos.x := 0;
      Inc(Pos.y);
      Result := #10;
    end else begin
      Result := s[Pos.x+1];
      Inc(Pos.x);
    end
  end else
    if Pos.x = 0 then begin
      Dec(Pos.y);
      if Pos.y >= 0 then
        Pos.x := Length(Lines[Pos.y])
      else
        Pos.x := 0;
      Result := #10;
    end else begin
      Dec(Pos.x);
      Result := Lines[Pos.y][Pos.x+1];
    end;
end;

{ EQ -- Compare two characters.  The "bc" comes from the buffer, "pc"
  from the pattern.  If we are not in MatchCase mode, fold out the case. }
function EQ(bc, pc: Char; MatchCase: Boolean): Boolean;
begin
  if not MatchCase then
    bc := LowerChar(bc);
  Result := bc = pc;
end;

{ LitEQ -- compare the string versus the current characters in the line.
  Returns 0 (no match) or the number of characters matched. }
function LitEQ(const s: string; Lines: TStrings;
               var Pos: TPoint; Options:TFindOptions): Integer;
var
  p: TPoint;
  i: Integer;
begin
  p := Pos;
  Result := 0;
  for i := 1 to Length(s) do
    if not EQ(NextCh(Lines, p, frDown in Options), s[i], frMatchCase in Options) then
      Exit;
  Pos := p;
  Result := Length(s);
end;

{ MoveLocalPoint -- Take the passed Pos and adjust by n characters. If boundry
  is hit, leave it alone and return True. Otherwise all went well, and return False. }
function MoveLocalPoint(n: Integer; Lines: TStrings; var Pos: TPoint): Boolean;
var
  p: TPoint;
  n1: Integer;
begin
  p := Pos;
  Result := True;
  if n > 0 then
    while n > 0 do begin
      if p.y >= Lines.Count then
        Exit; { hit end of buffer }
      n1 := Length(Lines[p.y])-p.x;
      if n1 >= n then begin
        Inc(p.x, n);
        n := 0;
      end else begin
        Inc(p.y);
        p.x := 0;
        Dec(n, n1+1);
      end;
    end
  else
    while n < 0 do begin
      if p.x + n >= 0 then begin
        Inc(p.x, n);
        Break;
      end else begin
        Dec(p.y);
        if p.y < 0 then Exit; { Start of buffer }
        Inc(n, p.x+1);
        p.x := Length(Lines[p.y]);
      end;
    end;
  Pos := p;
  Result := False;
end;


{ FBound -- Move forwards (backwards) at least Jump characters to end of
  next possible match position. Return True if end (start) of buffer hit }
function FBound(Jump: Integer; Lines: TStrings; Delta: TDelta;
                var Pos: TPoint; Options:TFindOptions): Boolean;
var
  p: TPoint;
begin
  p := Pos;
  if not (frDown in Options) then
    Inc(Jump);
  while Jump <> 0 do begin
    if not (frDown in Options) then
      Jump := -Jump;
    Result := MoveLocalPoint(Jump, Lines, p);
    if Result then Exit; { hit end of buffer }
    if (p.y >= Lines.Count) or (p.x = Length(Lines[p.y])) then
      Jump := Delta.Delta[10]
    else
      Jump := Delta.Delta[Ord(Lines[p.y][p.x+1])];
  end;
  Pos := p;
  Result := False;
end;

{ Scanner -- Search for a pattern in either direction.
  Fast version using simplified version of Boyer and Moore
  Software-Practice and Experience, vol 10, 501-506 (1980) }
function Scanner(Lines: TStrings; Delta: TDelta; var Pos: TPoint;
                 Options:TFindOptions; var FoundRect:TRect): Boolean;
var
  p, p1, Match: TPoint;
  n: Integer;
begin
  p := Pos;
  Result := False;
  if (frDown in Options)
     and ((p.y >= Lines.Count) or ((p.y = Lines.Count-1) and (p.x = Length(Lines[p.y])))) then
    Exit; { Speical case to avoid looping  at end of file if replacing newlines }
  if FBound(Delta.Patlen, Lines, Delta, p, Options) then
    Exit; { Too near end of buffer for a match }
  repeat
    { Set up the scanning pointers, and save the current position
      in case we match the search string at this point. }
    if frDown in Options then
      MoveLocalPoint(-Delta.Patlen, Lines, p)
    else
      MoveLocalPoint(Delta.Patlen+1, Lines, p);
    if frWholeWord in Options then begin
      if frDown in Options then
        n := -1
      else
        n := 1;
      if not MoveLocalPoint(n, Lines, p) then
        if IsCharAlphaNumeric(NextCh(Lines, p, frDown in Options)) then begin
          { delimiter - force fail here}
          MoveLocalPoint(-n, Lines, p);
          Continue;
        end;
    end;
    Match := p;
    if LitEQ(Delta.Patrn, Lines, p, Options) <> 0 then begin
       if (frWholeWord in Options) then begin
         p1 := p;
         if IsCharAlphaNumeric(NextCh(Lines, p1, frDown in Options)) then begin
           p := Match;  { Revert to start of pattern }
           NextCh(Lines, p, frDown in Options); { Skip one char }
           Continue;
         end;
       end;
      { Successfull Match }
      Pos := p;
      Result := True;
      if frDown in Options then begin
        FoundRect.TopLeft := Match;
        FoundRect.BottomRight := p;
      end else begin
        FoundRect.TopLeft := p;
        FoundRect.BottomRight := Match;
      end;
      Exit;
    end;
  until FBound(Delta.Jump, Lines, Delta, p, Options);
end;

function SearchStrings(Lines: TStrings; var Pos: TPoint;
                       Options:TFindOptions; Target: string;
                       var FoundRect:TRect): Boolean;
var
  i, j: Integer;
  tegraT: string;
  Delta: TDelta;
begin
  { Discard #13 in Target }
  j := 1;
  for i := 1 to Length(Target) do
    if Target[i] <> #13 then begin
      Target[j] := Target[i];
      Inc(j);
    end;
  SetLength(Target, j-1);
  Result := True;
  if Length(Target) = 0 then Exit;
  if frDown in Options then
    Delta := MakeDelta(Target, frMatchCase in Options)
  else begin
    SetLength(tegraT, Length(Target));
    for i := 1 to Length(Target) do
      tegraT[i] := Target[Length(Target)-i+1];
    Delta := MakeDelta(tegraT, frMatchCase in Options);
  end;
  Result := Scanner(Lines, Delta, Pos, Options, FoundRect);
end;

end.

