unit TextPre;

interface
{$I DFS.inc}{ Copyright (C) 2005 C M Western; see PcontrolsReg.pas for license}
uses
  {$IFDEF MSWINDOWS} Windows, {$ENDIF}
  {$IFNDEF CLX_} Graphics, Forms, Controls, StdCtrls, Buttons, ExtCtrls,
  {$ELSE} QGraphics, QForms, QControls, QStdCtrls, QButtons, QExtCtrls, Types, {$ENDIF}
  SysUtils, Classes, NewSpin, PrintPre, AlEdit, EditExtr, ExpCan;

type     
  TTextSettingsDlg = class(TForm)
    OKBtn: TButton;
    CancelBtn: TButton;
    Wrap: TCheckBox;
    Label1: TLabel;
    Columns: TSpinBox;
    Bevel1: TBevel;
    Label2: TLabel;
    Separation: TSpinBox;
  private
    { Private declarations }
  public
    { Public declarations }
  end;

  type
  TTextPrintPreview = class(TPrintPreview)
  private
    FColumns: Integer;
    FFont: TFont;
    FColumnSeparation: Integer;
    FWrap: Boolean;
    FontHeight: Double; { Line Separation as a fraction of printable page height }
    LinesPerCol, CharsPerCol: Integer;
    procedure SetColumns(const Value: Integer);
    procedure SetFont(const Value: TFont);
    procedure SetColumnSeparation(const Value: Integer);
    procedure SettingsClick(Sender: TObject);
    procedure SetWrap(const Value: Boolean);
  protected
    procedure ReadBack(Sender: TObject); override;
    procedure DoShowPreview(Sender:TObject; Canvas:TExportableCanvas; X, Y, AWidth, AHeight:Longint;
                            APage: Integer; OnPrinter:Boolean); override;
  public
    Lines: TStrings; { Note: must be allocated/freed externally;
                       May be a partially implemented TStrings descendant }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Font: TFont read FFont write SetFont;
    property Columns: Integer read FColumns write SetColumns default 1;
    property ColumnSeparation: Integer read FColumnSeparation write SetColumnSeparation;
    property Wrap: Boolean read FWrap write SetWrap default True;
  end;

{$IFDEF SINGLE}
procedure Register;
{$ENDIF}

implementation

uses {$IFDEF CLX_} Qt, {$ENDIF} {$IFDEF LCL_} LCLType, LCLIntf, {$ENDIF} {$IFNDEF VER80} Math, {$ENDIF} Prevwdlg;

{$IFNDEF LCL_}
{$R *.DFM}
{$ELSE}
{$R *.lfm}
{$ENDIF}

{$IFDEF SINGLE}
procedure Register;
begin
  RegisterComponents('CMW', [TTextPrintPreview]);
end;
{$ENDIF}

{ TTextPrintPreview }

constructor TTextPrintPreview.Create(AOwner: TComponent);
begin
  FColumns := 1;
  FColumnSeparation := 5;
  FFont := TFont.Create;
  Font.OnChange := {$IFDEF FPC}@{$ENDIF}ReadBack;
  FWrap := True;
  inherited Create(AOwner);
  Options := Options + [ppAreaIsFirstPage];
end;

destructor TTextPrintPreview.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TTextPrintPreview.DoShowPreview(Sender: TObject; Canvas: TExportableCanvas;
  X, Y, AWidth, AHeight: Longint; APage: Integer; OnPrinter: Boolean);
var
  l, Count, Column, ColSep, ColWidth, Page: Integer;
  r: TRect;
  dy: Double;
  Line: string;
  procedure DoText(const Text: string);
  begin
    Canvas.TextRect(r, {$IFDEF LCL_} 0, 0, {$ELSE} r.Left, r.Top, {$ENDIF} Text);
  end;
begin
  Canvas.Font := Font;
  {$IFDEF MSWINDOWS}
  Canvas.Font.Height := Round(FontHeight * AHeight);
  {$ENDIF}
  {$IFDEF UNIX}
  Canvas.Font.Size := Round(FontHeight * AHeight); { Use point size }
  {$ENDIF}
  Line := '';
  if Assigned(Lines) then begin
    if Wrap then begin
      Page := 1;
      l := 0;
    end else begin
      l := (APage-1)*Columns*LinesPerCol;
      Page := APage;
    end;
    Count := LinesPerCol;
    dy := AHeight / LinesPerCol;
    Column := 0;
    ColWidth := AWidth;
    ColSep := Trunc(ColumnSeparation/XSize * AWidth);
    if Columns > 1 then
      ColWidth := Trunc((AWidth-(Columns-1)*ColSep)/Columns);
    while l < Lines.Count do begin
      while Count >= LinesPerCol do begin
        r.Left := Column*(ColWidth + ColSep) + X; { TODO: Check win }
        Inc(Column);
        if Column > Columns then begin
          Inc(Page);
          Column := 0;
          Continue;
        end;
        r.Right := r.Left + ColWidth;
        Count := 0;
      end;
      if Page > APage then Break;
      r.Top := Round((Count mod LinesPerCol) * dy) + Y; { TODO: Check win }
      r.Bottom := Round(((Count+1) mod LinesPerCol) * dy) + Y; { TODO: Check win }
      if r.Bottom < r.Top then
        r.Bottom := AHeight + Y; { Check for wrap to next column }
      if Wrap then begin
        if Length(Line) = 0 then
          Line := Lines[l];
        if Length(Line) > CharsPerCol then begin
          if Page = APage then
            DoText(Copy(Line,1,CharsPerCol));
          Line := Copy(Line, 1+CharsPerCol, Length(Line)-CharsPerCol);
          Dec(l); { Some more to come }
        end else begin
          if Page = APage then
            DoText(Line);
          Line := '';
        end;
      end else
        DoText(Lines[l]);
      Inc(l);
      Inc(Count);
    end;
  end;
  inherited DoShowPreview(Sender, Canvas, X, Y, AWidth, AHeight, APage, OnPrinter);
end;

procedure TTextPrintPreview.ReadBack(Sender: TObject);
var
  {$IFNDEF CLX_}
  tm: TTextMetric;
  ScreenDC: HDC;
  Original: THandle;
  {$ENDIF}
  {$IFDEF CLX_}
  tm: QFontMetricsH;
  {$ENDIF}
  tn: TNotifyEvent;
  l, ll, nWrapped: Integer;
  t: Double;
begin
  if not Assigned(PreviewForm) then Exit;
  tn := PreviewForm.OnSettingsClick;
  if not Assigned(tn) then
    PreviewForm.OnSettingsClick := {$IFDEF FPC}@{$ENDIF}SettingsClick;
  {$IFNDEF CLX_}
  ScreenDC := GetDC(0);
  Original := SelectObject(ScreenDC, Font.Handle);
  {$ENDIF}
  {$IFDEF CLX_}
  tm := nil;
  {$ENDIF}
  try
    {$IFNDEF CLX_}
    GetTextMetrics(ScreenDC, tm);
    l := Abs(Font.Size);
    if l = 0 then l := tm.tmHeight; { default LCL font does this }
    FontHeight := l / 72 * 25.4 / PreviewForm.YSize;
    FontHeight := Abs(FontHeight)*(tm.tmHeight + tm.tmExternalLeading)/tm.tmHeight;
    {$ENDIF}
    {$IFDEF CLX_}
    tm := QFontMetrics_create(Font.Handle);
    FontHeight := Abs(Font.Size) / 72 * 25.4 / PreviewForm.YSize;
    // FCharSize.x := Max(1, QFontMetrics_width(tm, 'A'));
    // FCharSize.y := Max(1, QFontMetrics_lineSpacing(tm)); ?? Height/External leading
    // TODO
    {$ENDIF}
    { TODO : LCL_ }
    LinesPerCol := Trunc(1/FontHeight);
    if Assigned(Lines) then begin
      nWrapped := 0;
      { Column width in mm }
      t := PreviewForm.XSize;
      if Columns > 1 then
        t := (t - (Columns-1)*ColumnSeparation) / Columns;
      { in ScreenDC pixels }
      t := t / 25.4 * Screen.PixelsPerInch;
      if Wrap then begin
        { Should test for Font.Pitch = fpFixed, but likely to be very long winded }
        { Column width in characters }
        {$IFNDEF CLX_}
        { TODO: Linux }
        CharsPerCol := Trunc(t / tm.tmAveCharWidth);
        {$ENDIF}
        {$IFDEF CLX_}
        CharsPerCol := Trunc(t / QFontMetrics_width(tm, 'A'));
        {$ENDIF}
        for l := 0 to Lines.Count - 1 do begin
          ll := Length(Lines[l]);
          while ll > CharsPerCol do begin
            Dec(ll, CharsPerCol);
            Inc(nWrapped);
          end;
        end;
      end;
      MaxPages := Ceil((Lines.Count+nWrapped) / LinesPerCol / Columns);
    end;
  finally
    {$IFNDEF CLX_}
    SelectObject(ScreenDC, Original);
    ReleaseDC(0, ScreenDC);
    {$ENDIF}
    {$IFDEF CLX_}
    if tm <> nil then QFontMetrics_destroy(tm);
    {$ENDIF}
  end;
  inherited ReadBack(Sender);
end;

procedure TTextPrintPreview.SetColumns(const Value: Integer);
begin
  if FColumns <> Value then begin
    FColumns := Value;
    if Assigned(PreviewForm) then
      PreviewForm.Changed;
  end;
end;

procedure TTextPrintPreview.SetColumnSeparation(const Value: Integer);
begin
  if FColumnSeparation <> Value then begin
    FColumnSeparation := Value;
    if Assigned(PreviewForm) then
      PreviewForm.Changed;
  end;
end;

procedure TTextPrintPreview.SetWrap(const Value: Boolean);
begin
  if FWrap <> Value then begin
    FWrap := Value;
    if Assigned(PreviewForm) then
      PreviewForm.Changed;
  end;
end;

procedure TTextPrintPreview.SetFont(const Value: TFont);
begin
  FFont.Assign(Value);
end;

procedure TTextPrintPreview.SettingsClick(Sender: TObject);
var
  Dlg: TTextSettingsDlg;
begin
  Dlg := nil;
  try
    Dlg := TTextSettingsDlg.Create(Self);
    Dlg.Columns.Value := Columns;
    Dlg.Wrap.Checked := Wrap;
    Dlg.Separation.Value := ColumnSeparation;
    if Dlg.ShowModal = mrOk then begin
      Columns := Dlg.Columns.Value;
      Wrap := Dlg.Wrap.Checked;
      ColumnSeparation := Dlg.Separation.Value;
    end;
  finally
   Dlg.Release;
  end;
end;

end.
