unit Graphics;
{$I DFS.inc}{Copyright (C) 2005 C M Western; see commandline.pas for license}
interface

uses Classes;

{ TColor definition and implementation; required for console programs when full Graphics unit not required }
type
  TColor =  -$7FFFFFFF-1..$7FFFFFFF;

const
  // The following colors match the predefined Delphi Colors
  clBlack   = TColor($000000);
  clMaroon  = TColor($000080);
  clGreen   = TColor($008000);
  clOlive   = TColor($008080);
  clNavy    = TColor($800000);
  clPurple  = TColor($800080);
  clTeal    = TColor($808000);
  clGray    = TColor($808080);
  clSilver  = TColor($C0C0C0);
  clRed     = TColor($0000FF);
  clLime    = TColor($00FF00);
  clYellow  = TColor($00FFFF);
  clBlue    = TColor($FF0000);
  clFuchsia = TColor($FF00FF);
  clAqua    = TColor($FFFF00);
  clLtGray  = TColor($C0C0C0);
  clDkGray  = TColor($808080);
  clWhite   = TColor($FFFFFF);
  clCream   = TColor($F0FBFF);
  clNone    = TColor($1FFFFFFF);
  clDefault = TColor($20000000);
  clWebSienna = TColor($2D52A0);
  clWebOrange = TColor($00A5FF);

type
   TDPoint = record
     X: Double;
     Y: Double;
   end;
   
   TDRect = record
     case Integer of
       0: (Left, Top, Right, Bottom: Double);
       1: (TopLeft, BottomRight: TDPoint);
   end;

   TCanvas = TObject;
   TGraphic = THandle;

   { TFont }

   TFont = class(TPersistent)
   private
     FHeight: Integer;
     FSize: Integer;
     procedure SetHeight(const AValue: Integer);
     procedure SetSize(const AValue: Integer);
   public
     PixelsPerInch: Integer;
     constructor Create;
     property Size: Integer read FSize write SetSize;
     property Height: Integer read FHeight write SetHeight;
   end;

   TPenStyle = (psSolid, psDash, psDot, psDashDot, psDashDotDot, psinsideFrame, psPattern,psClear);
   TPenMode = (pmBlack, pmWhite, pmNop, pmNot, pmCopy, pmNotCopy,
               pmMergePenNot, pmMaskPenNot, pmMergeNotPen, pmMaskNotPen, pmMerge,
               pmNotMerge, pmMask, pmNotMask, pmXor, pmNotXor);

   { TPen }

   TPen = class(TPersistent)
   private
     FColor: TColor;
     FMode: TPenMode;
     FStyle: TPenStyle;
     FWidth: Integer;
   public
     property Color: TColor read FColor Write FColor;
     property Mode:  TPenMode read FMode write FMode;
     property Style: TPenStyle read FStyle write FStyle;
     property Width: Integer read FWidth write FWidth;
   end;

   TBrushStyle = (bsSolid, bsClear, bsHorizontal, bsVertical, bsFDiagonal,
                   bsBDiagonal, bsCross, bsDiagCross, bsImage, bsPattern);

   { TBrush }

   TBrush = class(TPersistent)
   private
     FColor: TColor;
     FStyle: TBrushStyle;
   public
     property Color : TColor read FColor Write FColor;
     property Style: TBrushStyle read FStyle write FStyle;
   end;

function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
function GetRValue(Colour: TColor):Byte;
function GetGValue(Colour: TColor):Byte;
function GetBValue(Colour: TColor):Byte;
function ColorToRGB(Colour: TColor):TColor;

implementation

uses {$IFDEF MSWINDOWS} Windows, {$ENDIF} UCompat;

function RGB(Red:Byte; Green:Byte; Blue:Byte): TColor;
begin
  Result := Red or (Green shl 8) or (Blue shl 16);
end;

function GetRValue(Colour: TColor):Byte;
begin
  Result := Colour and 255;
end;

function GetGValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 8) and 255;
end;

function GetBValue(Colour: TColor):Byte;
begin
  Result := (Colour shr 16) and 255;
end;

function ColorToRGB(Colour: TColor):TColor;
begin
  if (Result < clBlack) or (Result > clWhite) then
    Result := clBlack
  else
    Result := Colour;
end;

{ TFont }

procedure TFont.SetHeight(const AValue: integer);
begin
  FHeight := AValue;
  FSize := -MulDiv(AValue, 72, PixelsPerInch);
end;

procedure TFont.SetSize(const AValue: integer);
begin
  FSize := AValue;
  FHeight := -MulDiv(AValue, PixelsPerInch, 72);
end;

constructor TFont.Create;
begin
  inherited Create;
  PixelsPerInch := 72;
end;

end.

